package edu.ucla.cs.SMOsTranslator.Pattern;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import net.sf.jsqlparser.expression.Expression;
import net.sf.jsqlparser.schema.Column;
import net.sf.jsqlparser.schema.Table;
import net.sf.jsqlparser.statement.Statement;
import net.sf.jsqlparser.statement.create.table.ColumnDefinition;
import net.sf.jsqlparser.statement.create.table.CreateTable;
import net.sf.jsqlparser.statement.drop.DropTable;
import net.sf.jsqlparser.statement.insert.Insert;
import edu.ucla.cs.SMOsTranslator.Pattern.Exceptions.TooManyTablesException;
import edu.ucla.cs.SMOsTranslator.Pattern.Exceptions.UnrecognizedPatternException;
import edu.ucla.cs.SMOsTranslator.Pattern.visitor.ExtractColumnsVisitor;
import edu.ucla.cs.SMOsTranslator.Pattern.visitor.ReversePartitionConditionVisitor;
import edu.ucla.cs.SMOsTranslator.Pattern.visitor.SourceTableVisitor;
import edu.ucla.cs.SMOsTranslator.Pattern.visitor.WhereConditionVisitor;
import edu.ucla.cs.SMOsTranslator.Pattern.visitor.WherePartitionVisitor;
import edu.ucla.cs.db.prism.backend.schema.Schema;
import edu.ucla.cs.db.prism.backend.schema.exception.IllegalSMOException;
import edu.ucla.cs.db.prism.backend.schema.exception.TableHasNoColumns;
import edu.ucla.cs.velocity.template.AddColumnTemplate;
import edu.ucla.cs.velocity.template.CopyTableTemplate;
import edu.ucla.cs.velocity.template.DecomposeTableTemplate;
import edu.ucla.cs.velocity.template.DropTableTemplate;
import edu.ucla.cs.velocity.template.PartitionTableTemplate;
import edu.ucla.cs.velocity.template.PartitionTemplateSource;
import edu.ucla.cs.velocity.template.RenameColumnTemplate;

public class DecomposePattern extends Pattern{


	private final String[] patternLabels = {Pattern.CREATE_TAB, Pattern.CREATE_TAB, Pattern.INSERT, Pattern.INSERT, Pattern.DROP_TAB};
	private final String patternRecognitionMessage = "Decompose pattern recognized";

	//Schema instance
	private Schema schema;

	//CREATE TABLE variables
	private CreateTable create1;
	private CreateTable create2;
	private List<String> tablesCreated;

	//INSERT variables
	private Insert insert1;
	private Insert insert2;
	private List<String> insertDestinationTables;
	private Table sourceTable1;
	private Table sourceTable2;
	private Expression partitionCondition;
	private List<Column> selectedColumns1;
	private List<Column> selectedColumns2;

	//DROP TABLE variables
	private DropTable drop;
	private Table dropTable;
	private Expression deleteCondition;


	public DecomposePattern(Schema schema){
		this.schema = schema;
	}

	@Override
	public String getPatternRecognitionMessage() {
		return patternRecognitionMessage;
	}

	@Override
	public boolean isPattern(String[] labels, ArrayList<Statement> stmts) throws UnrecognizedPatternException, IllegalSMOException {

		//Pattern VISITOR
		SourceTableVisitor sourceTableVisitor = new SourceTableVisitor();
		ExtractColumnsVisitor columnVisitor = new ExtractColumnsVisitor();

		insertDestinationTables = new ArrayList<String>();
		tablesCreated = new ArrayList<String>();


		ArrayList<Table> sourceTables1;
		ArrayList<Table> sourceTables2;


		//Check #0
		if(checkLabels(labels)){

			//Get the CREATE variables 
			create1 = (CreateTable)stmts.get(0);
			create2 = (CreateTable)stmts.get(1);
			tablesCreated.add(create1.getTable().getName());
			tablesCreated.add(create2.getTable().getName());

			//Get the INSERT variables
			insert1 = (Insert)stmts.get(2);
			insert2 = (Insert)stmts.get(3);
			sourceTables1 = sourceTableVisitor.getTablesFromInsert(insert1);
			sourceTables2 = sourceTableVisitor.getTablesFromInsert(insert2);

			if ((sourceTables1.size() >1) || (sourceTables2.size()>1) )
				throw new TooManyTablesException("The INSERT statement must contain at most one source table");

			sourceTable1 = sourceTables1.get(0);
			sourceTable2 = sourceTables2.get(0);

			insertDestinationTables.add(insert1.getTable().getName());
			insertDestinationTables.add(insert2.getTable().getName());

			selectedColumns1 = columnVisitor.getSelectedColumns(insert1);
			selectedColumns2 = columnVisitor.getSelectedColumns(insert2);

			//Check #4
			//the two set of columns selected in the INSERT can't be null
			//in order to be considered a decompose pattern, at least one column per INSERT 
			if ((selectedColumns1 == null ) || (selectedColumns2 == null))
				throw new TableHasNoColumns("At least one column must be selected from the source tables");


			//Get the DROP variables
			try{
				drop = (DropTable)stmts.get(4);
				if(drop.getTable().size()>1)
					throw new TooManyTablesException("The DROP TABLE statement must refer to at most one table");	
				dropTable = drop.getTable().get(0);
			}
			catch(NullPointerException nEx){}

			//Check #1
			//INSERT destination tables and CREATE tables are the same
			for(String t: insertDestinationTables){
				if(!tablesCreated.contains(t)){
					throw new UnrecognizedPatternException("The INSERT destination tables and the CREATE tables must coincide");
				}
			}

			//Check #2
			if(!sourceTable1.getName().equalsIgnoreCase(sourceTable2.getName()))
				throw new UnrecognizedPatternException("Both the INSERT statement must refer to the same table");

			//Check #3
			//DROP table and INSERT table must coincide
			try{
				if(!dropTable.getName().equalsIgnoreCase(sourceTable1.getName()))
					throw new UnrecognizedPatternException("DROP table and INSERT source table must coincide");
			}
			catch(NullPointerException nEx){}

			return true;
		}
		return false;
	}

	@Override
	public ArrayList<String> translatePattern(ArrayList<Statement> stmts) throws SQLException, IllegalSMOException {

		boolean checkColumn;
		boolean doublePartitioning = false;

		//VISITOR

		//INSERT variables
		List<Column> insertColumns1;
		List<Column> insertColumns2;

		List<String> columnsToAddToTable1 = new ArrayList<String>();
		List<String> columnsToRenameInTable1 = new  ArrayList<String>();
		List<String> columnsToAddToTable2 = new ArrayList<String>();
		List<String> columnsToRenameInTable2 = new  ArrayList<String>();

		ArrayList<String> smos = new ArrayList<String>();

		//RENAME COLUMN
		List<String> columnsToCompareFromSource1 = new ArrayList<String>();
		List<String> columnsToCompareFromSource2 = new ArrayList<String>();
		List<String> columnsToCompareFromDestination1 = new ArrayList<String>();
		List<String> columnsToCompareFromDestination2 = new ArrayList<String>();

		//PARTITION
		PartitionTemplateSource partitionSource;
		final String temporaryPartitionTrashTable = "temp_trash_Table";

		//DOUBLE PARTITIONING
		String table1InDoublePart = null;
		String table2InDoublePart = null;

		insertColumns1 = insert1.getColumns();
		insertColumns2 = insert2.getColumns();

		//selected column can't be null (pattern check )
		//selected columns
		//		if(selectedColumns1 == null)
		//			columnsToCompareFromSource1 = schema.getTable(sourceTable1.getName()).getColumns();
		//		else{
		for(Column col: selectedColumns1)
			columnsToCompareFromSource1.add(col.getColumnName());
		//}
		//		if(selectedColumns2 == null)
		//			columnsToCompareFromSource2 = schema.getTable(sourceTable2.getName()).getColumns();
		//		else{
		for(Column col: selectedColumns2)
			columnsToCompareFromSource2.add(col.getColumnName());
		//}
		//insert destination columns
		if(insertColumns1 == null){
			for(ColumnDefinition colDef: create1.getColumnDefinitions())
				columnsToCompareFromDestination1.add(colDef.getColumnName().getColumnName()); 	
		}
		else{
			for(Column column: insertColumns1)
				columnsToCompareFromDestination1.add(column.getColumnName());
		}
		if(insertColumns2 == null){
			for(ColumnDefinition colDef: create2.getColumnDefinitions())
				columnsToCompareFromDestination2.add(colDef.getColumnName().getColumnName()); 	
		}
		else{
			for(Column column: insertColumns2)
				columnsToCompareFromDestination2.add(column.getColumnName());
		}

		for(int i=0; i<columnsToCompareFromSource1.size(); i++){
			if(!columnsToCompareFromDestination1.get(i).equalsIgnoreCase(columnsToCompareFromSource1.get(i)))
				columnsToRenameInTable1.add(RenameColumnTemplate.getRenameColumnSMO(columnsToCompareFromSource1.get(i), columnsToCompareFromDestination1.get(i), create1.getTable().getName()));
		}
		for(int i=0; i<columnsToCompareFromSource2.size(); i++){
			if(!columnsToCompareFromDestination2.get(i).equalsIgnoreCase(columnsToCompareFromSource2.get(i)))
				columnsToRenameInTable2.add(RenameColumnTemplate.getRenameColumnSMO(columnsToCompareFromSource2.get(i), columnsToCompareFromDestination2.get(i), create2.getTable().getName()));
		}

		//ADD COLUMN
		//add the column not involved in the INSERT clause to the partition
		for(ColumnDefinition colDef: create1.getColumnDefinitions()){
			checkColumn = true;
			for(String col: columnsToCompareFromDestination1){
				if(colDef.getColumnName().getColumnName().equalsIgnoreCase(col))
					checkColumn = false;
			}
			if(checkColumn){
				columnsToAddToTable1.add(AddColumnTemplate.getAddColumnSMO(colDef, tablesCreated.get(0), ""));
			}
		}
		for(ColumnDefinition colDef: create2.getColumnDefinitions()){
			checkColumn = true;
			for(String col: columnsToCompareFromDestination2){
				if(colDef.getColumnName().getColumnName().equalsIgnoreCase(col))
					checkColumn = false;
			}
			if(checkColumn){
				columnsToAddToTable2.add(AddColumnTemplate.getAddColumnSMO(colDef, tablesCreated.get(1), ""));
			}
		}


		//DROP TABLE
		//Check if the source table is dropped in the pattern
		if(dropTable != null){
			smos.add(CopyTableTemplate.getCopyTableSMO(sourceTable1.getName()));
			sourceTable1.setName(CopyTableTemplate.getCopiedTableName(sourceTable1.getName()));
		}

		//PARTITION TABLE
		//P1)Check if a partition condition exist
		try{
			Expression partitionCondition1 = new WherePartitionVisitor().getPartitionCondition(new WhereConditionVisitor().getWhereCondition(insert1));
			Expression partitionCondition2 = new WherePartitionVisitor().getPartitionCondition(new WhereConditionVisitor().getWhereCondition(insert2));

			if(!partitionCondition1.toString().equalsIgnoreCase(partitionCondition2.toString())){

				doublePartitioning = true;

				//The partition condition of the two INSERT statement are different, perform a double partitioning
				smos.add(CopyTableTemplate.getCopyTableSMO(sourceTable1.getName()));
				table1InDoublePart = sourceTable1.getName();
				table2InDoublePart = CopyTableTemplate.getCopiedTableName(sourceTable1.getName());

				//PARTITION TABLE1
				partitionSource = new PartitionTemplateSource();
				partitionSource.setTablePartition1(table1InDoublePart);
				partitionSource.setTablePartition2(table1InDoublePart);
				partitionSource.setTableName(table1InDoublePart);
				partitionSource.setPartitionCondition1(partitionCondition1.toString());
				partitionSource.setPartitionCondition2(new ReversePartitionConditionVisitor().getReversePartitionCondition(partitionCondition1).toString());
				smos.add(replaceNonSmoChar(PartitionTableTemplate.getPartitionTableSMO(partitionSource)));

				//drop useless partition
				smos.add(DropTableTemplate.getDropTableSMO(PartitionTemplateSource.getDroppedPartitionTable(table1InDoublePart)));

				//update the table1 name
				table1InDoublePart = PartitionTemplateSource.getKeptPartitionTable(table1InDoublePart);

				//PARTITION TABLE2
				partitionSource = new PartitionTemplateSource();
				partitionSource.setTablePartition1(table2InDoublePart);
				partitionSource.setTablePartition2(table2InDoublePart);
				partitionSource.setTableName(table2InDoublePart);
				partitionSource.setPartitionCondition1(partitionCondition2.toString());
				partitionSource.setPartitionCondition2(new ReversePartitionConditionVisitor().getReversePartitionCondition(partitionCondition2).toString());
				smos.add(replaceNonSmoChar(PartitionTableTemplate.getPartitionTableSMO(partitionSource)));

				//drop useless partition
				smos.add(DropTableTemplate.getDropTableSMO(PartitionTemplateSource.getDroppedPartitionTable(table2InDoublePart)));

				//update the table2 name
				table2InDoublePart = PartitionTemplateSource.getKeptPartitionTable(table2InDoublePart);
			}
			else{
				//partition the sourceTable
				partitionSource = new PartitionTemplateSource();
				String tableToPartition = sourceTable1.getName();
				partitionSource.setTablePartition1(tableToPartition);
				partitionSource.setTablePartition2(tableToPartition);
				partitionSource.setTableName(tableToPartition);
				partitionSource.setPartitionCondition1(partitionCondition1.toString());
				partitionSource.setPartitionCondition2(new ReversePartitionConditionVisitor().getReversePartitionCondition(partitionCondition1).toString());
				smos.add(replaceNonSmoChar(PartitionTableTemplate.getPartitionTableSMO(partitionSource)));

				//drop useless partition
				smos.add(DropTableTemplate.getDropTableSMO(PartitionTemplateSource.getDroppedPartitionTable(tableToPartition)));

				//update the source table name
				sourceTable1.setName(PartitionTemplateSource.getKeptPartitionTable(tableToPartition));
			}
		}
		catch(NullPointerException nEx){
			//NO PARTITION TO EXECUTE
		}

		//DECOMPOSE THE TABLE
		if(doublePartitioning){

			//DECOMPOSE TABLE1
			smos.add(DecomposeTableTemplate.getDecomposeTableSMO(table1InDoublePart, tablesCreated.get(0),temporaryPartitionTrashTable, selectedColumns1, selectedColumns2));
			//drop temporary trash table
			smos.add(DropTableTemplate.getDropTableSMO(temporaryPartitionTrashTable));

			//DECOMPOSE TABLE2
			smos.add(DecomposeTableTemplate.getDecomposeTableSMO(table2InDoublePart, tablesCreated.get(1), temporaryPartitionTrashTable, selectedColumns2, selectedColumns1));
			//drop temporary trash table
			smos.add(DropTableTemplate.getDropTableSMO(temporaryPartitionTrashTable));

		}
		else{
			smos.add(DecomposeTableTemplate.getDecomposeTableSMO(sourceTable1.getName(), tablesCreated.get(0), tablesCreated.get(1), selectedColumns1, selectedColumns2));
		}



		return smos;
	}

	@Override
	public boolean checkLabels(String[] labels) {

		if ((labels.length > patternLabels.length) || (labels.length < patternLabels.length-1))
			return false;

		for(int i=0; i<patternLabels.length; i++){
			//last label
			if(i==patternLabels.length-1){
				try{
					if(labels[i].equals(patternLabels[i]))
						return true;
					return false;
				}
				catch(IndexOutOfBoundsException iEx){
					//no label at index i 
					return true;
				}
			}
			if(!labels[i].equals(patternLabels[i]))
				return false;
		}
		return true;
	}

}
