package edu.ucla.cs.SMOsTranslator.Pattern;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import net.sf.jsqlparser.expression.Expression;
import net.sf.jsqlparser.schema.Column;
import net.sf.jsqlparser.schema.Table;
import net.sf.jsqlparser.statement.Statement;
import net.sf.jsqlparser.statement.create.table.ColumnDefinition;
import net.sf.jsqlparser.statement.create.table.CreateTable;
import net.sf.jsqlparser.statement.delete.Delete;
import net.sf.jsqlparser.statement.insert.Insert;
import edu.ucla.cs.SMOsTranslator.Pattern.Exceptions.TooManyTablesException;
import edu.ucla.cs.SMOsTranslator.Pattern.Exceptions.UnrecognizedPatternException;
import edu.ucla.cs.SMOsTranslator.Pattern.visitor.ExtractColumnsVisitor;
import edu.ucla.cs.SMOsTranslator.Pattern.visitor.PartitionConditionVisitor;
import edu.ucla.cs.SMOsTranslator.Pattern.visitor.ReversePartitionConditionVisitor;
import edu.ucla.cs.SMOsTranslator.Pattern.visitor.SourceTableVisitor;
import edu.ucla.cs.SMOsTranslator.Pattern.visitor.WhereConditionVisitor;
import edu.ucla.cs.SMOsTranslator.Pattern.visitor.WherePartitionVisitor;
import edu.ucla.cs.db.prism.backend.schema.Schema;
import edu.ucla.cs.db.prism.backend.schema.exception.IllegalSMOException;
import edu.ucla.cs.velocity.template.AddColumnTemplate;
import edu.ucla.cs.velocity.template.DropColumnTemplate;
import edu.ucla.cs.velocity.template.PartitionTableTemplate;
import edu.ucla.cs.velocity.template.RenameColumnTemplate;

public class PartitionPattern1 extends Pattern{


	private final String[] patternLabels = {Pattern.CREATE_TAB, Pattern.INSERT, Pattern.DELETE};
	private final String patternRecognitionMessage = "Partition pattern (type 1) recognized";

	//Schema instance
	private Schema schema;

	//CREATE TABLE variables
	private CreateTable create;
	private Table tableCreated;

	//INSERT variables
	private Insert insert;
	private Table tableFromInsert;
	private Table sourceTable;
	private Expression partitionCondition;
	private Expression whereCondition;
	private List<Column> selectedColumns;
	private List<Column> insertColumns;

	//DELETE variables
	private Delete delete;
	private Table deleteTable;
	private Expression deleteCondition;


	public PartitionPattern1(Schema schema){
		this.schema = schema;
	}

	@Override
	public String getPatternRecognitionMessage() {
		return patternRecognitionMessage;
	}

	@Override
	public boolean isPattern(String[] labels, ArrayList<Statement> stmts) throws UnrecognizedPatternException, IllegalSMOException {

		//Pattern VISITOR
		SourceTableVisitor sourceTableVisitor = new SourceTableVisitor();
		WherePartitionVisitor partitionVisitor = new WherePartitionVisitor();
		WhereConditionVisitor whereVisitor = new WhereConditionVisitor();


		//Check #0
		if(checkLabels(labels)){

			//Get the CREATE variables 
			create = (CreateTable)stmts.get(0);
			tableCreated = create.getTable();

			//Get the DELETE variables
			delete = (Delete)stmts.get(2);
			if(delete.getTables().size()>1)
				throw new TooManyTablesException("The DELETE statement in the pattern must contain at most one table");
			deleteTable = delete.getTables().get(0);
			deleteCondition = delete.getWhere();

			//Get the INSERT variables 
			insert = (Insert)stmts.get(1);
			tableFromInsert = insert.getTable();

			if(sourceTableVisitor.getTablesFromInsert(insert).size()>1)
				throw new TooManyTablesException("The INSERT statement in the pattern must contain at most one table");

			sourceTable = sourceTableVisitor.getTablesFromInsert(insert).get(0);

			//Check #1
			//the table created and the destination table are the same
			if(!tableCreated.getName().equalsIgnoreCase(tableFromInsert.getName()))
				throw new UnrecognizedPatternException("The created table and the source table must coincide");

			//Check #2
			//the source table is the same as the deleted one
			if(!sourceTable.getName().equalsIgnoreCase(deleteTable.getName()))
				throw new UnrecognizedPatternException("The source table and the deleted table must coincide");

			//Check #3
			//the partition is on a column of the source table
			try{
				whereCondition = whereVisitor.getWhereCondition(this.insert);
				partitionCondition = partitionVisitor.getPartitionCondition(whereCondition);
				PartitionConditionVisitor conditionVisitor = new PartitionConditionVisitor();
				Column columnPartitioned = conditionVisitor.getColumnFromPartitionCondition(partitionCondition);
				try{
					if ((columnPartitioned.getTable().getName().equals(sourceTable.getName()) || columnPartitioned.getTable().getName().equals(sourceTable.getAlias())))
						throw new UnrecognizedPatternException("The column in the WHERE condition doesn't belong to the source table");
				}
				catch(NullPointerException nEx){
					//OK, no table name defined for the column, this means it belongs to the source table
				}
			}
			catch(NullPointerException nEx){
				throw new UnrecognizedPatternException("The partition pattern (type 1) must contain at least one partition condition");
			}

			//Check #4
			//the DELETE and INSERT share the same partition condition
			if(!deleteCondition.toString().equalsIgnoreCase(whereCondition.toString()))
				throw new UnrecognizedPatternException("The DELETE and the INSERT condition must coincide");

			//Check #5
			//The column in the select and in the insert statement are the same
			insertColumns = insert.getColumns();
			selectedColumns = new ExtractColumnsVisitor().getSelectedColumns(insert);

			try{
				if(selectedColumns.size() != insertColumns.size()){
					throw new UnrecognizedPatternException("The number of columns in the INSERT clause and in the SELECT clause must be the same");
				}
			}
			catch(NullPointerException nEx){
				//FIXME
				//if one of the to list of columns is different, read the columns from the schema and compare
				return true;
			}

			return true;
		}
		return false;
	}

	@Override
	public ArrayList<String> translatePattern(ArrayList<Statement> stmts) throws SQLException, IllegalSMOException {

		boolean checkColumn;

		List<String> columnsToDropFromPartition = new ArrayList<String>();
		List<String> columnsToAddToPartition = new ArrayList<String>();
		List<String> columnsToRenameInPartition = new ArrayList<String>();

		ArrayList<String> smos = new ArrayList<String>();

		//RENAME COLUMN
		List<String> columnsToCompareFromSource = new ArrayList<String>();
		List<String> columnsToCompareFromDestination = new ArrayList<String>();

		if(selectedColumns == null)
			columnsToCompareFromSource = schema.getTable(sourceTable.getName()).getColumns();
		else{
			for(Column col: selectedColumns)
				columnsToCompareFromSource.add(col.getColumnName());
		}
		if(insertColumns == null){
			for(ColumnDefinition colDef: create.getColumnDefinitions())
				columnsToCompareFromDestination.add(colDef.getColumnName().getColumnName()); 	
		}
		else{
			for(Column column: insertColumns)
				columnsToCompareFromDestination.add(column.getColumnName());
		}

		for(int i=0; i<columnsToCompareFromSource.size(); i++){
			if(!columnsToCompareFromDestination.get(i).equalsIgnoreCase(columnsToCompareFromSource.get(i)))
				//columnsToRenameInPartition.put(columnsToCompareFromDestination.get(i), columnsToCompareFromSource.get(i));
				columnsToRenameInPartition.add(RenameColumnTemplate.getRenameColumnSMO(columnsToCompareFromSource.get(i), columnsToCompareFromDestination.get(i), create.getTable().getName()));
		}


		//ADD COLUMN
		//add the column not involved in the INSERT clause to the partition
		for(ColumnDefinition colDef: create.getColumnDefinitions()){
			checkColumn = true;
			for(String col: columnsToCompareFromDestination){
				if(colDef.getColumnName().getColumnName().equalsIgnoreCase(col))
					checkColumn = false;
			}
			if(checkColumn){
				columnsToAddToPartition.add(AddColumnTemplate.getAddColumnSMO(colDef, tableCreated.getName(), ""));
			}
		}


		//DROP COLUMN
		//drop the column not selected from the partition
		try{
			for(String column: schema.getTable(sourceTable.getName()).getColumns()){
				checkColumn = true;
				for(Column col: selectedColumns){
					if(column.equalsIgnoreCase(col.getColumnName()))
						checkColumn = false;
				}
				if(checkColumn){
					columnsToDropFromPartition.add(DropColumnTemplate.getDropColumnSMO(column,create.getTable().getName()));
				}
			}
		}
		catch(NullPointerException nEx){
			//NO column to be dropped from the partition
		}


		//PARTITION TABLE
		//P1)Get the complementary partition condition
		Expression reversePartitionCondition = new ReversePartitionConditionVisitor().getReversePartitionCondition(partitionCondition);

		//P2) create the PartitionSource object
		String tableToPartion = sourceTable.getName();
		String partition1 = create.getTable().getName();
		String partition2 = sourceTable.getName();
		String condition1 = partitionCondition.toString();
		String condition2 = reversePartitionCondition.toString();

		smos.add(replaceNonSmoChar(PartitionTableTemplate.getPartitionTableSMO(partition1, partition2, condition1, condition2, tableToPartion)));

		smos.addAll(columnsToAddToPartition);
		smos.addAll(columnsToDropFromPartition);
		smos.addAll(columnsToRenameInPartition);

		return smos;
	}

	@Override
	public boolean checkLabels(String[] labels) {
		if (!(labels.length == patternLabels.length))
			return false;

		for(int i=0; i<patternLabels.length; i++){
			//last label
			if(!labels[i].equals(patternLabels[i]))
				return false;
		}
		return true;
	}

}
