/**
 * Copyright (C) 2009 
 * Developer: Fabrizio Moroni fabrizio.moroni@yahoo.it
 * Contributor / Contact Person: Carlo Curino	carlo@curino.us
 * 
 *  * This file is part of SchemEvolutionToolsuite
 * 
 * SchemEvolutionToolsuite is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SchemaEvolutionToolsuite is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with SchemEvolutionToolsuite.  If not, see <http://www.gnu.org/licenses/>.
 */
package edu.ucla.cs.SMOsTranslator;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import net.sf.jsqlparser.statement.Statement;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import edu.ucla.cs.Exceptions.UndefinedApplicationParameterException;
import edu.ucla.cs.SMOsTranslator.Pattern.Pattern;
import edu.ucla.cs.SMOsTranslator.Pattern.PatternFactory;
import edu.ucla.cs.SMOsTranslator.Pattern.StatementLabelGenerator;
import edu.ucla.cs.SMOsTranslator.Pattern.Exceptions.UnrecognizedPatternException;
import edu.ucla.cs.SMOsTranslator.Pattern.visitor.TablesForStatementVisitor;
import edu.ucla.cs.db.prism.backend.schema.Schema;
import edu.ucla.cs.db.prism.backend.schema.exception.IllegalSMOException;
import edu.ucla.cs.db.prism.backend.schema.exception.MergingDifferentTablesException;
import edu.ucla.cs.db.prism.backend.smo.exception.SMOParsingErrorException;
import edu.ucla.cs.db.prism.backend.smo.util.SMOFactory;
import edu.ucla.cs.dbevolution.ApplicationParameters;
import edu.ucla.cs.dbevolution.DbBackend.PrismSchemaLoader;


public class SqlToSmoTranslator {


	private TableDependenciesGraph tableDependenciesGraph;
	private ArrayList<String> translatedStmts;
	private ArrayList<Statement> stmts;
	private ArrayList<TablesPerStatement> tablesPerStatement;

	private static Logger logger = Logger.getLogger(SqlToSmoTranslator.class);

	public SqlToSmoTranslator(){

		tableDependenciesGraph = null;
		translatedStmts = null;
		stmts = null;
	}


	/**
	 * Main class: the following are the step to follow in order to get to the SQL statement translation:
	 * <br>
	 * <br>Read the tables affected
	 * @param args
	 * @throws UnrecognizedPatternException 
	 */
	public static void main(String[] args) throws UnrecognizedPatternException{
	
		try {
			ApplicationParameters.loadApplicationParameters(args);
		}
		catch (UndefinedApplicationParameterException e) {
			logger.error(e);
			System.exit(-1);
		}
		SqlToSmoTranslator translator = new SqlToSmoTranslator();
		translator.translateEvolutionStep();
	}
	
	
	public void translateEvolutionStep(){
	
		ArrayList<Statement> stmts = null;
		StatementTranslator trans = null;

		boolean multiTablesTranslation = false;
		boolean patternMatch = false;

		Schema schema = null;

		ArrayList<String> smos = null;
		Pattern pattern = null;

		//get a log instance
		logger.setLevel(Level.DEBUG);

		//LOAD THE SCHEMA BY THE VERSION NUMBER SET BY THE USER
		logger.info("Loading the schema......");
		schema = PrismSchemaLoader.loadSchemaFromDB(true, ApplicationParameters.getDbbasename());
		logger.info("Schema loaded......");

		//create a new SqlToSmoTranslator object
		//read the input file retrieving the SQL statements
		stmts = (ArrayList<Statement>)EvolutionFileReader.parseStatements(ApplicationParameters.getEvolutionFile());
		logger.info("All the statement have been correctly parsed");


		//if no errors occurred(all the statements have been read without returning exception)
		//create the table-dependencies-graph AND create the IsolationGraph starting from the Table Dependencies Graph		
		List<Set<String>> connectedComponents = createTablesDependenciesGraph(stmts).connectedComponents();
		logger.info("dependencies graph created");

		for(Set<String> component : connectedComponents){
			//retrieve the iterator for the table in the table dep. graph
			//TableDependenciesGraph vertexIsolation = (TableDependenciesGraph)it.next();
			//Iterator<String> itVertexTable = vertexIsolation.iterator();

			logger.info("Analyzing connected component: " + component);

			//check the number of vertex in the subgraph
			//1 ====>  one table, ==> possible 1:1 translation
			//2 ====>  two+ tables ===> check against the patterns
			if(component.size() > 1){
				multiTablesTranslation = true;
			}

			ArrayList<Statement> statementsByTable = new ArrayList<Statement>();
			smos = new ArrayList<String>();

			for(String connectedTable: component){
				statementsByTable.addAll(getStatementsByTable(connectedTable, getTablesPerStatement(), stmts));
				logger.debug("All the statement for table "+ connectedTable + " retrieved.");
			}

			//reorder the statements
			//Statement[] testStmts = new Statement[statementsByTable.size()];
			Statement[] testStmts = new Statement[stmts.size()];
			for(Statement stmt : statementsByTable){
				testStmts[stmts.indexOf(stmt)] = stmt;
			}

			statementsByTable.clear();
			for(Statement stmt: testStmts){
				if(stmt != null)
					statementsByTable.add(stmt);
			}
			logger.info("Statements reordered");

			if(multiTablesTranslation){

				logger.info("Starting the pattern-matching mode");
				logger.info("Statements belonging to the connected subgraph");

				//Create the LABELS for the statements
				String[] labels = generateStatementLabel(statementsByTable);

				logger.info("Statemetn labels created");

				//iterate over the PATTERNS and check if the current batch match a PATTERN
				try{
					for(PatternFactory.PatternType patternType :PatternFactory.PatternType.values()){
						pattern = PatternFactory.createPattern(patternType, schema);
						if(pattern.isPattern(labels,statementsByTable)){
							logger.info(pattern.getPatternRecognitionMessage());
							smos.addAll(pattern.translatePattern(statementsByTable));
							patternMatch = true;
						}
					}
					if(!patternMatch){
						//the current BATCH of statement didn't match any of the patterns
						//try computing the power set 

						logger.info("None of the previous pattern matched");
						logger.info("The system is looking for possible pattern in the current set of statement");

						ArrayList<ArrayList<Statement>> powerSet = statementsPowerSet(statementsByTable);
						int powerSetCount = 0;
						for(ArrayList<Statement> s :powerSet){

							//Create the LABELS for the statements
							if(s.size()==0)
								continue;
							labels = generateStatementLabel(s);

							for(PatternFactory.PatternType patternType :PatternFactory.PatternType.values()){
								pattern = PatternFactory.createPattern(patternType, schema);
								try{
									if(pattern.isPattern(labels,s)){
										logger.info("========================================================");
										logger.info("Analyzing power set element: " + powerSetCount);
										logger.info(pattern.getPatternRecognitionMessage());
										logger.info("The statement composing the pattern are:");
										for(Statement st: s){
											logger.info(st);
										}
										ArrayList<String> smosPowerSet = pattern.translatePattern(s);
										logger.info("********************************************************");
										logger.info("SMO obtained translating the subset of the power set");
										try{
											for(String smoP: smosPowerSet)
												logger.info(smoP);
										}
										catch(NullPointerException nEx){}
										logger.info("========================================================");
									}
								}
								catch(UnrecognizedPatternException uEx){
									//logger.info(uEx);
								}
							}
							powerSetCount++;
						}
					}
				}
				catch(UnrecognizedPatternException uEx){
					logger.error(uEx.getMessage());
				}
				catch(MergingDifferentTablesException mEx){
					logger.error(mEx);
					logger.error("Review the SQL set and try again the translation");
				}
				catch(IllegalSMOException iEx){
					iEx.printStackTrace();
					System.exit(-1);
				}
				catch(SQLException sEx){
					logger.fatal(sEx);
				}
			}
			else{

				//TRANSLATE the single / multiple statements as NON PATTERN statements
				for(Statement stmt: statementsByTable ){
					try{
						try{
							trans = StatementTranslatorDispatcher.getStatementTranslator(stmt,schema);
							smos.addAll(((ArrayList<String>)trans.translateStatement(stmt)));
						}
						catch(UnknowStatementTranslatorException uEx){
							logger.error("Unable to translate statement " + stmt.toString());
						}
						catch(NullPointerException nEx){
							logger.error("No SMO from the current translation: ");
						}
					}
					catch(SQLException e){
						e.printStackTrace();
					}
				}
			}
			//PRINT the generated SMO
			try{
				for(String smo: smos){
					logger.info("SMO: " + smo);
				}
			}
			catch(NullPointerException nEx){
				//NO SMOs generated
			}

			//try to apply the generated SMO to the schema
			try{
				for(String smo: smos){
					schema.applySMO(SMOFactory.createSMOFromString(smo));
				}
			}
			catch (IllegalSMOException e) {
				logger.error(e);
				System.exit(-1);
			} catch (SMOParsingErrorException e) {
				logger.fatal("The following SMO has a wrong syntax, aborting the execution...", e);
				System.exit(-1);
			}
			catch(NullPointerException nEx){
				//NO SMOs generated
			}
			multiTablesTranslation = false;
		}

	}

	private static String[] generateStatementLabel(ArrayList<Statement> stmts){

		String[] labels =  new String[stmts.size()];
		for(int i=0; i<stmts.size();i++){
			labels[i] = StatementLabelGenerator.getStatementLabel(stmts.get(i));
		}
		return labels;
	}


	private TableDependenciesGraph createTablesDependenciesGraph(ArrayList<Statement> stmts/*, SqlToSmoTranslator translator*/){

		TableDependenciesGraph tableDependenciesGraph = new TableDependenciesGraph();
		TablesForStatementVisitor affectedTablesVisitor;

		//given the batch of statements
		//for each statement, extract the affected tables
		ArrayList<TablesPerStatement> tablesPerStatement = new ArrayList<TablesPerStatement>();
		for(Statement stmt: stmts){
			affectedTablesVisitor = new TablesForStatementVisitor();
			ArrayList<String> affectedTables = (ArrayList<String>)affectedTablesVisitor.getAffectedTablesFromStatement(stmt);

			//create the object statements, affected tables..
			TablesPerStatement x = new TablesPerStatement();
			x.setStatement(stmt);
			x.setTables(affectedTables);
			tablesPerStatement.add(x);
		}

		this.tablesPerStatement = tablesPerStatement;

		//create the dependencies graph
		for(TablesPerStatement tbs: tablesPerStatement){
			tableDependenciesGraph.addTablesToDependenciesGraph(tbs.getTables());
		}

		return tableDependenciesGraph;
	}


	/**
	 * Return a list of statements affecting the given table
	 * @param table
	 * @param xxx
	 * @return
	 */
	private static List<Statement> getStatementsByTable(String table, ArrayList<TablesPerStatement> xxx, ArrayList<Statement> stmts){

		ArrayList<Statement> stmtPerTable = new ArrayList<Statement>();

		for(TablesPerStatement tbs: xxx){
			//before retrieving the statement, check the translated FLAG
			if(tbs.getTables().contains((table)) && !tbs.isTranslated()){
				stmtPerTable.add(tbs.getStatement());
				tbs.setTranslated(true);
			}
		}
		return stmtPerTable;
	}


	public TableDependenciesGraph getTableDependenciesGraph() {
		return tableDependenciesGraph;
	}

	public void setTableDependenciesGraph(
			TableDependenciesGraph tableDependenciesGraph) {
		this.tableDependenciesGraph = tableDependenciesGraph;
	}

	public ArrayList<String> getTranslatedStmts() {
		return translatedStmts;
	}

	public void setTranslatedStmts(ArrayList<String> translatedStmts) {
		this.translatedStmts = translatedStmts;
	}

	public ArrayList<Statement> getStmts() {
		return stmts;
	}

	public void setStmts(ArrayList<Statement> stmts) {
		this.stmts = stmts;
	}


	public ArrayList<TablesPerStatement> getTablesPerStatement() {
		return tablesPerStatement;
	}


	public void setTablesPerStatement(
			ArrayList<TablesPerStatement> tablesPerStatement) {
		this.tablesPerStatement = tablesPerStatement;
	}


	/**
	 * Returns the power set from the given set by using a binary counter
	 * Example: S = {a,b,c}
	 * P(S) = {[], [c], [b], [b, c], [a], [a, c], [a, b], [a, b, c]}
	 * @param set String[]
	 * @return LinkedHashSet
	 */
	private static ArrayList<ArrayList<Statement>> statementsPowerSet(ArrayList<Statement> stmts) {

		//create the empty power set
		ArrayList<ArrayList<Statement>> power = new ArrayList<ArrayList<Statement>>();

		//get the number of elements in the set
		int elements = stmts.size();

		//the number of members of a power set is 2^n
		int powerElements = (int) Math.pow(2,elements);

		//run a binary counter for the number of power elements
		for (int i = 0; i < powerElements; i++) {

			//convert the binary number to a string containing n digits
			String binary = intToBinary(i, elements);

			//create a new set
			ArrayList<Statement> innerSet = new ArrayList<Statement>();

			//convert each digit in the current binary number to the corresponding element
			//in the given set
			for (int j = 0; j < binary.length(); j++) {
				if (binary.charAt(j) == '1')
					innerSet.add(stmts.get(j));
			}

			//add the new set to the power set
			power.add(innerSet);

		}
		return power;
	}


	/**
	 * Converts the given integer to a String representing a binary number
	 * with the specified number of digits
	 * For example when using 4 digits the binary 1 is 0001
	 * @param binary int
	 * @param digits int
	 * @return String
	 */
	private static String intToBinary(int binary, int digits) {

		String temp = Integer.toBinaryString(binary);
		int foundDigits = temp.length();
		String returner = temp;
		for (int i = foundDigits; i < digits; i++) {
			returner = "0" + returner;
		}

		return returner;
	} 

}