package edu.ucla.cs.db.prism.backend.ded;
import java.util.Vector;
import java.io.*;
import edu.ucla.cs.db.prism.backend.schema.SchemaModifiedInfo;
import edu.ucla.cs.db.prism.backend.schema.TableModifiedInfo;

//import mapping.Mapping;

public class DEDSet implements Serializable {
	static final long serialVersionUID = 1;
	public int targetVersion; // the schema version on which query is posed
	public int minorTargetVersion; // for major version, this is zero
	public int sourceVersion; // the schema version under which data is stored
	public int minorSourceVersion; // for major version, this is zero
	public Vector<DED> deds;
	
	public DEDSet() {
		deds = new Vector<DED>();
	}

	
	public DEDSet(int sourceVersion, int minorSourceVersion, int targetVersion, int minorTargetVersion) {
		this();
		this.sourceVersion = sourceVersion;		
		this.targetVersion = targetVersion;
		this.minorSourceVersion = minorSourceVersion;
		this.minorTargetVersion = minorTargetVersion;
	}
	
	
	public boolean addAll(DEDSet dedset) {
		return this.deds.addAll(dedset.deds);
	}

	public int size() {
		return deds.size();
	}
	
	public boolean add(DED ded) {
		return deds.add(ded);
	}
	
	public DED get(int index) {
		return deds.get(index);
	}
	public DED remove(int index) {
		return deds.remove(index);
	}
	
	// used by compute (edu.ucla.db.prim.ded used as a view)
	public DED getDEDWithSingleGlobalQuantifier(String relationName) {
		for(int i=0; i<deds.size(); i++) {
			DED ded = deds.get(i);
			if(ded.globalQuantifiers.get(0).relation.equals(relationName) &&
					ded.globalQuantifiers.size() == 1)
				return ded;
		}
		return null;
	}
	
	
	@Override
	public String toString() {
		String str = "";
		str += "Source Version: v" + this.sourceVersion + 
				"m" + this.minorSourceVersion + "\n";
		str += "Target Version: v" + this.targetVersion + 
				"m" + this.minorTargetVersion + "\n";

		for(int i=0; i<this.size(); i++) {
			str += this.get(i).toString() + "\n";
		}
		return str;
	}
	
	public String toHTMLString() {
		String str = "";
		str += "Source Version: v" + this.sourceVersion + 
				"m" + this.minorSourceVersion + "<br>";
		str += "Target Version: v" + this.targetVersion + 
				"m" + this.minorTargetVersion + "<br>";

		for(int i=0; i<this.size(); i++) {
			str += this.get(i).toHTMLString() + "<br>";
		}
		return str;
	}
	
	public String toStringNoMetadata() {
		String str = "";
		for(int i=0; i<this.size(); i++) {
			str += this.get(i).toStringNoMetadata() + "\n";
		}
		return str;
	}
	
	// Unmodified Schema Part Optimization - 2/14/2007
	/*
	Case 1: a table has no change and its columns have no change.
	Case 2: a table has no change and some of its columns have
	changes.
	Case 3: a table has a change and its columns have no change.

	In case 1, we simply generate no DEDs. In case 2, we should include
	a table rewriting rule and its DEDs for query rewriting because its
	columns need the table rewriting to perform column rewriting in our
	rewriting scheme. In case 3, we need to issue table rewriting DEDs
	only and column rewriting is not necessary.
	
	In conclusion,
	Case 1: table no change, column no change -> no MARS-rewriting for both
	Case 2: table no change, some column change -> table MARS-rewriting, changed-column only MARS rewriting 
	Case 3: table change, column no change -> table MARS-rewriting only
	*/
	
	// this method constructs SchemaModifiedInfo from 'this' object
	public SchemaModifiedInfo constructSchemaModifiedInfo() {
		//System.out.println(this.toString());
		//System.out.println("construct...");
		SchemaModifiedInfo schema = new SchemaModifiedInfo(this.targetVersion);

		// step 1. process table DEDs first
		for(int i=0; i<this.size(); i++) {
			DED ded = this.get(i);
			if(!ded.chaseDEDFlag) continue; // process chase DEDs only
			if(!ded.rootDBFlag && ded.tableName != null && ded.columnName == null) { // table DED
				schema.addTable(new TableModifiedInfo(ded.tableName, ded.modifiedFlag));
			}
		}
		// step 2. process column DEDs now
		for(int i=0; i<this.size(); i++) {
			DED ded = this.get(i);
			if(!ded.chaseDEDFlag) continue; // process chase DEDs only
			if(!ded.rootDBFlag && ded.tableName != null && ded.columnName != null) { // column DED
				schema.getTable(ded.tableName).addColumn(ded.columnName, ded.modifiedFlag);
			}
		}
		return schema;
	}
	
	
	public SchemaModifiedInfo getModifiedTargetSchemaPart() {
		// based on the DED information, return the DEDs for the changed part of the target schema
		
		SchemaModifiedInfo schema = this.constructSchemaModifiedInfo();
		// step 1: remove unmodified columns
		for(int i=0; i<schema.size(); i++) {
			int numColumns = schema.getTable(i).getNumColumns();
			for(int j=0; j<numColumns; j++) {
				if(schema.getTable(i).getColumnModifiedFlag(j) == false) {
					//System.out.println("Removing " + schema.getTable(i).tableName + "." +
					//		schema.getTable(i).getColumn(j) + " i:" + i + " j:" + j);
					schema.getTable(i).removeColumn(j);
					j--;
					numColumns--;
				}
			}
		}
		
		// step 2: remove unmodified table with no (modified) columns left
		int numTables = schema.size();
		for(int i=0; i<numTables; i++) {
			if(schema.getTable(i).tableModifiedFlag == true)
				continue;
			if(schema.getTable(i).getNumColumns() > 0)
				continue;
			schema.tableModifiedInfos.remove(i);
			i--;
			numTables--;
		}
		
		// Return value: then we are left with
		//         1. modified tables,
		//		   2. unmodified tables with some modifed columns, and
		//         3. modified columns
		//         select the DEDs (both chase and backchase) for the remaining schema parts
		//         and return
		
		return schema;
	}
	
	
	// find the DEDs matching the target schema part
	public DEDSet getDEDsMacthincTargetSchemaPart(SchemaModifiedInfo schema) {
		//SchemaModifiedInfo schema = this.getModifiedTargetSchemaPart();
		
		DEDSet filteredDEDs = new DEDSet(this.sourceVersion,0, this.targetVersion,0);
		for(int i=0; i<this.size(); i++) {
			DED ded = this.get(i);
			
			if(ded.rootDBFlag) {// root, db mapping DED
				filteredDEDs.add(ded);
			}
			else if(ded.tableName == null && ded.columnName == null) {// text, attr
				// don't include them in the filtered DEDs
			}
			else if(ded.tableName != null && ded.columnName == null) {// table DEDs
				if(schema.containsTable(ded.tableName)) // found in filtered schema
					filteredDEDs.add(ded);
			}
			else if(ded.tableName != null && ded.columnName != null){ // column DEDs
				if(schema.containsColumn(ded.tableName, ded.columnName))
					filteredDEDs.add(ded);
			}
		}
		
		return filteredDEDs;
	}

	// if wantChase is true, return the chase deds only
	// otherwise, return the backchase deds only
	public DEDSet getDEDsforChase(boolean wantChase) {
		DEDSet filteredDEDs = new DEDSet(this.sourceVersion,0, this.targetVersion,0);
		for(int i=0; i<this.size(); i++) {
			DED ded = this.get(i);
			if(ded.chaseDEDFlag == wantChase)
				filteredDEDs.add(ded);
		}
		return filteredDEDs;
	}
	
	public void simplifyDisjunctive() {
		for(int i=0; i<deds.size(); i++) {
			deds.get(i).simplifyDisjunctive();
		}
	}



}
