package edu.ucla.cs.db.prism.backend.marsManage;

import java.io.BufferedWriter;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.PrintStream;
import java.io.StringWriter;
import java.util.Vector;

import edu.ucla.cs.db.prism.backend.MarsMain;
import edu.ucla.cs.db.prism.backend.SQL.SQLUnionSet;
import edu.ucla.cs.db.prism.backend.ded.DEDSet;
import edu.ucla.cs.db.prism.backend.ded.DEDSetHistory;
import edu.ucla.cs.db.prism.backend.marsManage.exception.MarsConstantsSuffixException;
import edu.ucla.cs.db.prism.backend.schema.SchemaHistory;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.ICSchemaHistory;
import edu.ucla.cs.db.prism.backend.util.StreamHandler;


public class MarsRunManage {
	
	static final String codiScriptFileName = "prima.codi";
	static final String marsResultFileName = "out.txt";
	static final boolean runMarsInSeparateJVM = false;
	static final boolean optionOneColumnPerDED = true;
	static final boolean optionMARSTranslate = true; // if false, do chase only
	
	static public SQLUnionSet marsRunManage(SQLUnionSet inputSQLUnionSet, 
											SchemaHistory schemaHistory, DEDSetHistory dedSetHistory,
											int sourceVersion, int sourceMinorVersion, 
											int targetVersion, int targetMinorVersion,
											Vector<String> sourceTablesForTranslate,
											boolean backchaseFlag) {
		
		// step 0.1: convert dedSetHistory into dedSet
		DEDSet dedSet = null;
		if(dedSetHistory != null) {
			dedSet = new DEDSet();
			for(int i=0; i<dedSetHistory.dedSets.size(); i++) {
				dedSet.deds.addAll(dedSetHistory.dedSets.get(i).deds);
			}
		}
		
		
		// step 0.2: MARS-specific fixes
		// attach a suffix "_" to all constants in query, to avoid zero-length constants
		try {
			MarsSpecificFix.addSuffixToConstants(inputSQLUnionSet);
		} catch (MarsConstantsSuffixException e1) {
			e1.printStackTrace();
		}
		// DED suffix are attached at the moment of SMO-to-DED construction now
		// since DED read-back from the metadata DB doesn't parse the DED and
		// it's difficult to attach suffix here
		
		// step 0.3: Relation-Encoded-Function translation
		inputSQLUnionSet.convertFunctionToREF(targetVersion, targetMinorVersion);
		inputSQLUnionSet.convertInequalityToREF(targetVersion, targetMinorVersion);
		

		// step 0.4: among the two-way deds, remove backchase rules, if we do chase only, not mars-translate
		if(!optionMARSTranslate) {
			int size = dedSet.size();
			for(int i=0; i<size; i++) {
				if(!dedSet.get(i).chaseDEDFlag) {
					dedSet.remove(i);
					i--;
					size--;
				}
			}
		}

		
		StringWriter codiScriptStringWriter = new StringWriter(); 
    	BufferedWriter codiScriptBW = new BufferedWriter(codiScriptStringWriter);
		
    	
		// step 1.1: input query, ded, chase/translate command
    	try {
    	
    		codiScriptBW.write(dedSet.toStringNoMetadata());
    	
		} catch(Exception e) {
			e.printStackTrace();
		}
    	CodiGen.writeInputQueries(inputSQLUnionSet, codiScriptBW);
    	
    	if(schemaHistory instanceof ICSchemaHistory)
        	CodiGen.writeCompileCommands(inputSQLUnionSet, codiScriptBW);
		else		
	    	CodiGen.writeCompileCommands(inputSQLUnionSet, codiScriptBW);
		    	
    	int chaseIndex = 0;
    	if(optionMARSTranslate){
    		if(schemaHistory instanceof ICSchemaHistory)
    			ICCodiGen.writeTranslateCommands(inputSQLUnionSet, codiScriptBW,
    					chaseIndex++, 
    					((ICSchemaHistory)schemaHistory), dedSetHistory, 
    					sourceVersion, sourceMinorVersion,
    					sourceTablesForTranslate);

    		else
    			CodiGen.writeTranslateCommands(inputSQLUnionSet, codiScriptBW,
    					chaseIndex++, 
    					schemaHistory, dedSetHistory, 
    					sourceVersion, sourceMinorVersion,
    					sourceTablesForTranslate);

    	}
    	else
    		CodiGen.writeChaseCommands(inputSQLUnionSet, codiScriptBW, 
    				chaseIndex++);

    	if(backchaseFlag) {
    		CodiGen.writeBackchaseCommands(
    				inputSQLUnionSet, codiScriptBW, chaseIndex++);
    	}
    
    	CodiGen.writePrintCommands(inputSQLUnionSet, codiScriptBW, chaseIndex);
    	try {
			codiScriptBW.close();
		} catch(Exception e) {
			e.printStackTrace();
		}
    	String queryDEDCommands = codiScriptStringWriter.toString();
    	
		
		// step 1.2: codi script schema part + the rest parts previously generated
		codiScriptStringWriter = new StringWriter();
		codiScriptBW = new BufferedWriter(codiScriptStringWriter);
	
		if(schemaHistory instanceof ICSchemaHistory)
			ICCodiGen.writeCodiPrelude((ICSchemaHistory)schemaHistory, dedSetHistory, inputSQLUnionSet, codiScriptBW);
		else		
			CodiGen.writeCodiPrelude(schemaHistory, dedSetHistory, inputSQLUnionSet, codiScriptBW);
		
		try{
			codiScriptBW.close();
		} catch(Exception e) {
			e.printStackTrace();
		}
		String schemaPart = codiScriptStringWriter.toString();
		String codiScript = schemaPart + queryDEDCommands;

		try {
			BufferedWriter out = new BufferedWriter(new FileWriter(codiScriptFileName));
			out.write(codiScript);
			out.close();
		} catch(Exception e) {
			e.printStackTrace();
		}
		
		
		// step 2: execute mars code
		if(runMarsInSeparateJVM)
			marsProcessExec();
		else
			marsCall();
		
		
		// step 3: parse the mars result in out.txt
		//System.out.println("result parsing ...");
		Vector<String> queryNames = extractQueryNames(inputSQLUnionSet);
		SQLUnionSet rewrittenXBindQueries =
			MarsResultParser.getRewrittenXBind(marsResultFileName, queryNames, null);

		// step 4: postprocess
		// step 4.1: version untag - now done at rewriteEngine level
		//if(rewrittenXBindQueries!=null)
		//	rewrittenXBindQueries.versionUntag();

		// step 4.2: constant suffix removal
		try {
			MarsSpecificFix.stripSuffixFromConstants(rewrittenXBindQueries);
		} catch (MarsConstantsSuffixException e) {
			e.printStackTrace();
		}
		
		// step 4.3: convert REF back to inequality or function
		rewrittenXBindQueries.convertREFToFunction(targetVersion, targetMinorVersion);
		rewrittenXBindQueries.convertREFToInequality(targetVersion, targetMinorVersion);

		
		return rewrittenXBindQueries;
	}
	
	
	// make a function call to mars entry point
	static private void marsCall() {
		try {
			String args[] = {codiScriptFileName};
	
			//Stopwatch timer = new Stopwatch();
	    	//System.out.println("called mars ... ");
	
	    	PrintStream fileOut = new PrintStream(new FileOutputStream(marsResultFileName));
	    	PrintStream stdout = System.out;

	    	// set System.out to a file output, namely out.txt
	    	// so that mars writes to out.txt
	    	System.setOut(fileOut);
	    	//timer.start();
			MarsMain.main(args);
			//timer.stop();
			
			// recover System.out to the original stdout
			System.setOut(stdout); 
			//System.out.println("MARS call time: " + timer.getElapsedTime() + " ms");
		}
		catch(Exception e) {
			e.printStackTrace();
		}
	}
	
	// starts a new process with a separate JVM
	static private void marsProcessExec() {
		Runtime runtime = Runtime.getRuntime();
		String command = "java -jar -Xms1024M -Xmx1024M -Xss16M -Xoss16M mars-0.4.jar " + codiScriptFileName;
		//Stopwatch timer = new Stopwatch();
		
		//command = "dir > out.txt";
		//System.out.println(command);
		
		try {
			// mars result
        	BufferedWriter marsResultOut = new BufferedWriter(new FileWriter(marsResultFileName));
        	BufferedWriter marsResultErr = new BufferedWriter(new FileWriter("error.txt"));

        	//timer.start();
        	Process p = runtime.exec(command);
			
			
			//BufferedReader inBrStdout = new BufferedReader(new InputStreamReader(p.getInputStream()));
			
			/*
			int bufSize = 1024*10;
			char[] buffer = new char[bufSize];
			while(true) {
				System.out.println("reading");
				int readBytes = inBrStdout.read(buffer, 0, bufSize);
				System.out.println("reading finshed: " + readBytes + " bytes read");
				if(readBytes == -1) break;
				marsResultOut.write(buffer, 0, readBytes);
				for(int i=0; i<readBytes; i++)
					System.out.print(buffer[i]);
			}
			*/
			
            // any output?
			StreamHandler outputHandler = new 
				StreamHandler(p.getInputStream(), marsResultOut, "OUTPUT");

            // any error message?
			StreamHandler errorHandler = new 
				StreamHandler(p.getErrorStream(), marsResultErr, "ERROR");            
            
            // kick them off
			outputHandler.start();
			errorHandler.start();
			
			try {
				p.waitFor();
			} catch(InterruptedException ex) {}
			//timer.stop();
			//System.out.println("MARS Call time: " + timer.getElapsedTime() + " ms");
			//System.out.println("process terminated with code = "+p.exitValue());
			
			//inBrStdout.close();
			marsResultOut.close();
			
		} catch(Exception e) {
			e.printStackTrace();
		}
	}
	static public Vector<String> extractQueryNames(SQLUnionSet sqlUnions) {
		Vector<String> queryNames = new Vector<String>();
		for(int i=0; i<sqlUnions.sqlUnions.size(); i++)
			queryNames.add(sqlUnions.sqlUnions.get(i).name);
		return queryNames;
	}
}
