package edu.ucla.cs.db.prism.backend.rewrite;

import java.util.Vector;

import edu.ucla.cs.db.prism.backend.SQL.RelationDeclaration;
import edu.ucla.cs.db.prism.backend.SQL.SQL;
import edu.ucla.cs.db.prism.backend.SQL.SQLUnion;
import edu.ucla.cs.db.prism.backend.SQL.SQLUnionSet;
import edu.ucla.cs.db.prism.backend.SQL.SelectAttribute;
import edu.ucla.cs.db.prism.backend.ded.DED;
import edu.ucla.cs.db.prism.backend.ded.DEDSet;
import edu.ucla.cs.db.prism.backend.ded.DEDSetHistory;
import edu.ucla.cs.db.prism.backend.exception.ErrorRewritingException;
import edu.ucla.cs.db.prism.backend.exception.RewritingOptimizationException;
import edu.ucla.cs.db.prism.backend.marsManage.MarsRunManage;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.ICSchema;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.ICSchemaHistory;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.ICTable;
import edu.ucla.cs.db.prism.backend.smo.util.Conjunctive;
import edu.ucla.cs.db.prism.backend.smo.util.Predicate;

public class ICPrimRewriteEngine extends PrimRewriteEngine {

	final boolean checkInputQueryCorrectness = true;
	private ICSchemaHistory schemaHistory;
	private DEDSetHistory dedSetHistory;
	
	/**
	 * constructor
	 * @param schemaHistory
	 * @param dedSetHistory
	 * @throws ErrorRewritingException 
	 */
	public ICPrimRewriteEngine(ICSchemaHistory schemaHistory, DEDSetHistory dedSetHistory) throws ErrorRewritingException {
		
		if(schemaHistory == null || schemaHistory.schemas == null || schemaHistory.schemas.isEmpty())
			throw new ErrorRewritingException("The Schema history is null or empty");
		
		if(dedSetHistory == null || dedSetHistory.dedSets == null || dedSetHistory.dedSets.isEmpty())
			throw new ErrorRewritingException("The DED set history is null or empty");

		
		
		this.schemaHistory = schemaHistory;
		this.dedSetHistory = dedSetHistory;
	}
	
	
	
	

	public ICPrimRewriteEngine() {
		// TODO Auto-generated constructor stub
	}





	/**
	 * user interface for query rewriting
	 * @param inputSQLUnion the input (union of) query to rewrite
	 * @param sourceVersion the schema version where the source data is stored
	 * @param targetVersion the schema version where the query is posed
	 * @return
	 * @throws ErrorRewritingException 
	 * @throws RewritingOptimizationException 
	 */
	@Override
	public SQLUnionSet rewrite(SQLUnionSet inputSQLUnionSet, 
			int sourceVersion, int sourceMinorVersion, 
			int targetVersion, int targetMinorVersion) throws ErrorRewritingException, RewritingOptimizationException {
		
		
		// step 1.1.1: check query correctness based on target schema version
		if(checkInputQueryCorrectness) {
			boolean correctQuery = checkInputQueryFootprintCorrectnessAgainstTargetVersionSchema(inputSQLUnionSet, schemaHistory.getSchemaByVersion(targetVersion, targetMinorVersion));
			if(!correctQuery) {
				System.err.println("the input query is ill-posed against the target schema version");
				System.err.println(inputSQLUnionSet);
				System.err.println(schemaHistory.getSchemaByVersion(targetVersion, targetMinorVersion).toStringWithMetadata());
				// changed behavior not to crash the interface at every wrong query... returning null should be ok..
				throw new ErrorRewritingException("The input query is ill-posed against the target schema version");
			}
		}		
		
		// step 1.2: version-tag the input query
		inputSQLUnionSet.versionTag(targetVersion,targetMinorVersion);
    	
		// step 1.3: extract query footprint on the target version
		Vector<String> queryFootprint = getTables(inputSQLUnionSet);

		// step 2: using a loop, find the best next intermediate source version to rewrite query to.
		//         note that this loop finishes when such next intermediate source version is
		//         same as the source version.
		//         otherwise, keep looping, using the chosen intermediate source version as 
		//		   the target version in the next loop
		//
		//         the next intermediate source version is the best next step 
		//		   given:
		//			i) the current target version (Vtgt)
		//          ii) the ultimate source version (Vsrc)
		//
		//         for such intermediate source version Vint, fetch/compute the following:
		//			i) the set of DEDs with src:Vint, tgt:Vtgt (both chase/backchase directions)
		//			ii) the query footprint QF' on Vint
		//		
		//		   among the set of (chase) DEDs chosen above, select only the deds that affect QF on Vtgt.
		//         source tables for such chase DEDs will constitute QF'
		//
		//		   note that we don't use ID-specific optimizations here
		int[] sourceVersionFull = {sourceVersion, sourceMinorVersion};
		int[] tempSourceVersion={-1,-1}; // major, minor version
		int[] tempTargetVersion = {targetVersion, targetMinorVersion};
		Vector<String> tempTargetVersionFootprint = queryFootprint;

		DEDSet tempDEDSet = null;
		DEDSet minTempDEDSet = null;
		DEDSetHistory minDEDSetHistory = new DEDSetHistory();
		ICSchema tempSchema = null;
		ICSchemaHistory queryFootprintSchemaHistory = new ICSchemaHistory();
		while(true) {
			// step 2.1: find the best tempSourceVersion, given sourceVersion and tempTargetVersion
			// composition network design principle: 
			//   greedy algorithm should return the optimal solution at each step

			//System.out.println("DEDSET:" + dedSetHistory.toString());
			
			//System.out.println("before optimization: " + sourceVersionFull[0] + "." + sourceVersionFull[1] + " to " + tempTargetVersion[0] + "." + tempTargetVersion[1]);
			
			tempSourceVersion = getBestTempSourceVersion(dedSetHistory, sourceVersionFull, tempTargetVersion);
			
			//System.out.println("after optimization: " + tempSourceVersion[0] + "." + tempSourceVersion[1] + " to " + tempTargetVersion[0] + "." + tempTargetVersion[1]);

			
			// FIXME in case the above algorithm returns a null tempSourceVersion set to the input
			if(tempSourceVersion == null){
				System.err.println("ERROR ICPrimRewriteEngine.getBestTempSourceVersion: tempSourceVersion was null");
				tempSourceVersion = sourceVersionFull;
			}
			
			// step 2.2: retrieve tempDEDSet from tempTargetVersion to tempSourceVerion
			tempDEDSet = dedSetHistory.get(tempSourceVersion, tempTargetVersion);
			
			if(tempDEDSet == null){
				
				throw new RewritingOptimizationException("NO DED in the dedSetHistory for versions:" + tempSourceVersion[0] + "." + tempSourceVersion[1] + " to " + tempTargetVersion[0] + "."+ tempTargetVersion[1]);
			
			}
				
			
			// step 2.3: minimize tempDEDSet into minTempDEDSet
			//           using tempTargetVersionFootprint and 
			//           tempDEDSet's targetVersion tables
			//           And add it to minDEDSetHistory
			
			minTempDEDSet = extractDEDsUsingTargetFootprint(tempDEDSet, tempTargetVersionFootprint);
			//System.out.println("Step 2.3 minTempDEDSet:" + minTempDEDSet.toString());
			
			// quick fix
			//minTempDEDSet = tempDEDSet;
			
			
			minDEDSetHistory.add(minTempDEDSet);
			
			
			// step 2.4.1: add tempTargetVersionFootprint to queryFootprintSchemaHistory 
			tempSchema = getSchemaForTables(schemaHistory, tempTargetVersion, tempTargetVersionFootprint);
			
			
			//FIXME 
			DEDSet localDEDs = null;
			ICSchema old = null;
			while(true){
				
				// we look for a schema fixpoint, we should also check IC fixpoint
				
				if(tempSchema.equals(old))
					break;
				
				//remember the old schema
				old = tempSchema.clone();
				
				
				
				// step 2.4.2: add local DEDs.
				localDEDs = tempSchema.getICasDEDs();
			
				//Local DEDs might refer to Tables which are "outside" the footprint of the Query, we add them here
				addTablesReferredByLocalDEDs(tempSchema,localDEDs,schemaHistory.getSchemaByVersion(tempTargetVersion));
				
				
			}
			
			
			queryFootprintSchemaHistory.addSchema(tempSchema);
			
			// step 2.5: update tempTargetVersionFootprint into tempDEDSet's source tables
			tempTargetVersionFootprint = getSourceFootprint(minTempDEDSet);
			

			
			// step 2.6: break if we're at last step 
			//           (after adding last footprint on source version 
			//            to queryFootprintSchemaHistory)
			if(compareVersions(tempSourceVersion, sourceVersionFull) == 0) {
				
				tempSchema = getSchemaForTables(schemaHistory, sourceVersionFull,
						tempTargetVersionFootprint);
				
				// same FIXPOINT computation following foreign keys as above
				old = null;
				while(true){	
					if(tempSchema.equals(old))
						break;
					old = tempSchema.clone();
					
					//optimistic break when schema is null
					if(old==null)
						break;
					localDEDs = tempSchema.getICasDEDs();
					addTablesReferredByLocalDEDs(tempSchema,localDEDs,schemaHistory.getSchemaByVersion(sourceVersionFull));
				}
				
				queryFootprintSchemaHistory.addSchema(tempSchema);
				break;
			}
			
			// step 2.7: update tempTargetVersion
			tempTargetVersion = tempSourceVersion;
		}
		
		SQLUnionSet rewrittenSQLUnionSet = null;
		

	//	System.out.println("QUERY:\n"  + inputSQLUnionSet + "\n\n");
	//	System.out.println("query footprint:\n"  + queryFootprintSchemaHistory.toStringWithMetadata() + "\n\n");
	//	System.out.println("minDED:\n"  + minDEDSetHistory + "\n\n");
	//	System.out.println("versions:" + sourceVersion +":"+ sourceMinorVersion+" "+ targetVersion+":"+targetMinorVersion);
		
		// step 3: call mars, using marsManage.MarsRunManage class
		// ATTENTION WE USE BACKCHASE TO EXPLOIT ICs for the OPTIMIZATION
		 rewrittenSQLUnionSet =
			MarsRunManage.marsRunManage(inputSQLUnionSet, 
					queryFootprintSchemaHistory, minDEDSetHistory, 
					sourceVersion, sourceMinorVersion, targetVersion,targetMinorVersion, 
					null, true);
		 
	
		// step 4: version-untag query
		if(rewrittenSQLUnionSet!=null)
			rewrittenSQLUnionSet.versionUntag();
		
		return rewrittenSQLUnionSet;
	}



	private void addTablesReferredByLocalDEDs(ICSchema tempSchema,
			DEDSet minTempDEDSet, ICSchema masterSchema) {

		for(DED d: minTempDEDSet.deds){

			if(d.sourceRelations!=null)
			for(String ict : d.sourceRelations){
				if(!tempSchema.containsTable(ict))
					tempSchema.addTable(masterSchema.getTable(ict));
			}
			if(d.targetRelations!=null)
			for(String ict : d.targetRelations){	
				if(!tempSchema.containsTable(ict))
					tempSchema.addTable(masterSchema.getTable(ict));
			}
		
			
		}
		
		
	}





	@Override
	public SQLUnion rewrite(SQLUnion inputSQLUnion, int sourceVersion, int sourceMinorVersion, int targetVersion, int targetMinorVersion) throws ErrorRewritingException {
		// step 1: give a name to input query, if null, to prevent errors in the later stage
		if(inputSQLUnion.name == null)
			inputSQLUnion.name = "q1";
		// step 2: call rewrite(SQLUnionSet), which does actual rewriting
		SQLUnionSet inputSQLUnionSet = new SQLUnionSet(); // a singleton set
		inputSQLUnionSet.sqlUnions.add(inputSQLUnion);
		//System.out.println(sourceMinorVersion+" "+targetMinorVersion);
		SQLUnionSet rewrittenSQLUnionSet =
			rewrite(inputSQLUnionSet, sourceVersion, sourceMinorVersion,
				targetVersion, targetMinorVersion);
		if(rewrittenSQLUnionSet == null)
			return null;
		if(rewrittenSQLUnionSet.sqlUnions.isEmpty())
			return null;
		else
			return rewrittenSQLUnionSet.sqlUnions.get(0);
	}
	
	/**
	 * user interface for query rewriting
	 * @param inputSQLUnion the input query to rewrite
	 * @param sourceVersion the schema version where the source data is stored
	 * @param targetVersion the schema version where the query is posed
	 * @return
	 * @throws ErrorRewritingException 
	 */
	@Override
	public SQLUnion rewrite(SQL inputSQL, int sourceVersion, int sourceMinorVersion, int targetVersion, int targetMinorVersion) throws ErrorRewritingException {
		SQLUnion inputSQLUnion = new SQLUnion(); // a singleton union
		inputSQLUnion.sqls.add(inputSQL);
		return rewrite(inputSQLUnion, sourceVersion, sourceMinorVersion, targetVersion, targetMinorVersion);
	}


	private Vector<String> getTables(SQLUnionSet sqlUnionSet) {
		if(sqlUnionSet.sqlUnions.size() == 1)
			return getTables(sqlUnionSet.sqlUnions.get(0));

		Vector<String> tables = new Vector<String>();
		for(int i=0; i<sqlUnionSet.sqlUnions.size(); i++) {
			tables.addAll(getTables(sqlUnionSet.sqlUnions.get(i)));
		}
		return tables;
	}
	private Vector<String> getTables(SQLUnion sqlUnion) {
		Vector<String> tables = new Vector<String>();
		for(int i=0; i<sqlUnion.sqls.size(); i++) {
			SQL sql = sqlUnion.sqls.get(i);
			for(int j=0; j<sql.from.relationDeclarationsSize(); j++) {
				RelationDeclaration rd = sql.from.relationDeclarations.get(j);
				if(!tables.contains(rd.relation))
					tables.add(rd.relation);
			}
		}
		return tables;
	}

	protected DEDSet extractDEDsUsingTargetFootprint(DEDSet dedSet, 
			Vector<String> tempTargetVersionFootprint) throws RewritingOptimizationException {
		
		//System.out.println("extractDEDsUsingTargetFootprint DEDSET :" + dedSet.toString());
		if(dedSet == null)
			throw new RewritingOptimizationException("The set of DEDs is empty, no optimization possible");
		DEDSet minDEDSet = new DEDSet();
		minDEDSet.sourceVersion = dedSet.sourceVersion;
		minDEDSet.minorSourceVersion = dedSet.minorSourceVersion;
		minDEDSet.targetVersion = dedSet.targetVersion;
		minDEDSet.minorTargetVersion = dedSet.minorTargetVersion;
		
		String versionPrefix = null;
		
		for(int i=0; i<dedSet.size(); i++) {
			DED ded = dedSet.get(i);
			Vector<String> testTables = null;
			versionPrefix = "v" + dedSet.targetVersion + "m" + dedSet.minorTargetVersion + "_";

			
			if(ded.chaseDEDFlag) { // data-exchange direction, test its target tables
				testTables = ded.targetRelations;
			}
			else { // query-rewriting direction, test its source tables
				testTables = ded.sourceRelations;
			}
			
			// for the DEDs which are "local" to a schema version let's add them all
			if(ded.isLocal()){
				testTables = ded.sourceRelations;
				testTables.addAll(ded.targetRelations);
			}
			
			// if there's an intersection between testTables and tempTargetVersionFootprint, 
			// include this ded into the return value
			for(int j=0; j<testTables.size(); j++) {
				if(tempTargetVersionFootprint.contains(versionPrefix + testTables.get(j))) {
					minDEDSet.add(ded);
					break;
				}
			}
		}
		
		return minDEDSet;
	}
	

	public String rewriteWithNegationAndICMO(String inputSQL) throws ErrorRewritingException{
		ICSchema target = schemaHistory.getOldest();
		ICSchema source = schemaHistory.getMostRecent();
		
		return rewriteWithNegationAndICMO(inputSQL,  source.versionID, source.minorVersionID, target.versionID, target.minorVersionID);
		
	}

	public String rewriteWithNegationAndICMO(String inputSQL, int sourceVersion, int sourceMinorVersion, int targetVersion, int targetMinorVersion) throws ErrorRewritingException {
		
		//System.out.println("ICPrimRewriteEngine.rewriteWithNegationAndICMO REWRITING:" + inputSQL + " from " + targetVersion + "."+ targetMinorVersion + " "+ sourceVersion +"."+sourceMinorVersion );
		
		
		//FIXME limitation, only one negation is dealt with at this point. recursive function call should be used to deal with nested negations.
		
		
		// extract positive and negative parts
		SQL negative = SQL.extractNegative(inputSQL);
		SQL positive = SQL.extractExternalPositive(inputSQL);
		
		int i =0;
		//go over the negative part and add all the tables used from the external positive part
		if(negative!=null && negative.where !=null && negative.where.conjunctives!=null){
			for(Conjunctive con:negative.where.conjunctives){
				for(Predicate p:con.predicates){
					
					// if a predicate uses an alias not in the from close of the negated 
					if(p.term1.tableName!= null && negative.from.getRelation(p.term1.tableName)==null){
						negative.from.addRelationDeclaration(new RelationDeclaration(positive.from.getRelation(p.term1.tableName),p.term1.tableName));
						negative.select.addSelectAttribute(new SelectAttribute("N" + i,p.term1.tableName, p.term1.columnName));
						i++;
					}
					
					if(p.term2.tableName!= null &&  negative.from.getRelation(p.term2.tableName)==null){
						negative.from.addRelationDeclaration(new RelationDeclaration(positive.from.getRelation(p.term2.tableName),p.term2.tableName));
						negative.select.addSelectAttribute(new SelectAttribute("N" + i,p.term2.tableName, p.term2.columnName));
						i++;
					}	
				}
			}
		}
		String output ="";			
		if(negative==null){
			SQLUnion rew= rewrite(positive, sourceVersion, sourceMinorVersion, targetVersion, targetMinorVersion);
			output = rew.toViewString(null);
		}
		else{
			SQLUnion rew1 = rewrite(positive,sourceVersion, sourceMinorVersion, targetVersion, targetMinorVersion);
			SQLUnion rew2 = rewrite(negative,sourceVersion, sourceMinorVersion, targetVersion, targetMinorVersion);		
			
			Vector<SelectAttribute> selectAttToBeRemoved = null;
			Vector<RelationDeclaration> relationDecltoBeRemoved=null;
			
			// removing tables from negative query added because referenced from the positive
			for(int j =0 ; j< rew2.sqls.size() ; j++){
				
				SQL s = rew2.sqls.elementAt(j);

				selectAttToBeRemoved = new Vector<SelectAttribute>();
				relationDecltoBeRemoved = new Vector<RelationDeclaration>();
				
				for(SelectAttribute outputatt : s.select.selectAttributes){
					String reltoremove = "";
					if(outputatt.outputAttributeName.startsWith("N")){
						
						// remove from the select clause
						selectAttToBeRemoved.add(outputatt);
						
						// remove the table from the FROM clause
						for(RelationDeclaration rd:s.from.relationDeclarations){
							if(rd.relationAlias.equals(outputatt.relationAlias)){
								reltoremove = rd.relation;
								relationDecltoBeRemoved.add(rd);
								
							}
						}
						
						// rename the the alias in the WHERE clause
						for(Conjunctive con1:s.where.conjunctives){
							for(Predicate p1:con1.predicates){
								if(outputatt.relationAlias.equals(p1.term1.tableName)){		
									for(SQL possql:rew1.sqls){
										Vector<RelationDeclaration> r = possql.from.getRelationDeclarationsByRelation(reltoremove);
										if(r!=null && !r.isEmpty())
										p1.term1.tableName = r.firstElement().relationAlias;										
									}
								}
							}	
						}	
					}	
				}
		

				for(SelectAttribute sa:selectAttToBeRemoved)
					s.select.selectAttributes.remove(sa);
				for(RelationDeclaration rd:relationDecltoBeRemoved)
					s.from.relationDeclarations.remove(rd);
			}
				
				
				output = rew1.toViewString(null) + " AND NOT EXISTS (\n\t " + rew2.toViewString(null) + "\n\t)";
		}
		
		return output;
	}
	
	
	private Vector<String> getSourceFootprint(DEDSet dedSet) {
		Vector<String> sourceTables = new Vector<String>();
		
		for(int i=0; i<dedSet.size(); i++) {
			DED ded = dedSet.get(i);
			Vector<String> tempSourceTables = null;
			
			// CARLO: modified this to source... check it
			//String versionPrefix = "v" + dedSet.targetVersion + "m" + dedSet.minorTargetVersion + "_";
			
			String versionPrefix = "v" + dedSet.sourceVersion + "m" + dedSet.minorSourceVersion + "_";
			
			if(ded.chaseDEDFlag) { // data-exchange direction, get its source tables
				tempSourceTables = ded.sourceRelations;
			}
			else { // query-rewriting direction, get its target tables
				tempSourceTables = ded.targetRelations;
			}
			
			for(int j=0; j<tempSourceTables.size(); j++) {
				if(!sourceTables.contains(versionPrefix + tempSourceTables.get(j)))
					sourceTables.add(versionPrefix + tempSourceTables.get(j));
			}
		}
		
		return sourceTables;
	}
	private ICSchema getSchemaForTables(ICSchemaHistory schemaHistory, 
			int version[], Vector<String> tables) {
		ICSchema schema = schemaHistory.getSchemaByVersion(version);
		ICSchema schemaSubset = new ICSchema();
		schemaSubset.versionID = schema.versionID;
		schemaSubset.minorVersionID = schema.minorVersionID;

		for(int i=0; i<tables.size(); i++) {
			int index = tables.get(i).indexOf("_");
			int violin = tables.get(i).indexOf("__viol");
			String tableNameWithoutVersionPrefix ="";
			if(violin>0)
				tableNameWithoutVersionPrefix = tables.get(i).substring(index+1,violin);
			else
				tableNameWithoutVersionPrefix = tables.get(i).substring(index+1);
			if(schema.getTable(tableNameWithoutVersionPrefix)!=null)
			schemaSubset.addTable(schema.getTable(tableNameWithoutVersionPrefix));
		}
		return schemaSubset;
	}

	 static public int[] getBestTempSourceVersion(DEDSetHistory dedSetHistory, 
			int sourceVersion[], int[] targetVersion) throws RewritingOptimizationException {
		
		//System.out.println("Looking for Best Versions from " + targetVersion[0] + "." +targetVersion[1]+ " and " + sourceVersion[0] + "." + sourceVersion[1]);
		
		
		int[] sourceVersionFull = sourceVersion;
		int[] bestSourceVersion = targetVersion;
		boolean targetGreaterThanSource = false;
		int test = compareVersions(targetVersion, sourceVersionFull);
		if(test == 0)
			throw new RewritingOptimizationException("The source and target version we are trying to optimize coincide!"); 
		else if(test > 0)
			targetGreaterThanSource = true;
		if(dedSetHistory == null || dedSetHistory.dedSets == null)
			throw new RewritingOptimizationException("The dedSetHistory received in input is null or empty");
		
		
		// find the next tempSourceVersion between targetVersion and sourceVersion
		// among all choices, choose one that is closest to sourceVersion
		if(targetGreaterThanSource){
			for(int i=dedSetHistory.size()-1; i>=0; i--) {
				DEDSet dedSet = dedSetHistory.get(i);
				int[] tempTargetVersion = {dedSet.targetVersion, dedSet.minorTargetVersion};
				int[] tempSourceVersion = {dedSet.sourceVersion, dedSet.minorSourceVersion};
				
				
				if(compareVersions(tempTargetVersion, targetVersion) == 0) {
					if(compareVersions(tempSourceVersion, bestSourceVersion) < 0 &&
						compareVersions(sourceVersionFull, tempSourceVersion) <= 0) {
						bestSourceVersion = tempSourceVersion; // updated
					}
				}
			}
		}
		else
		{
			for(int i=0; i<dedSetHistory.size(); i++) {
				DEDSet dedSet = dedSetHistory.get(i);
				
				int[] tempTargetVersion = {dedSet.targetVersion, dedSet.minorTargetVersion};
				int[] tempSourceVersion = {dedSet.sourceVersion, dedSet.minorSourceVersion};
				if(compareVersions(tempTargetVersion, targetVersion) == 0) {
					
					if( compareVersions(tempSourceVersion, bestSourceVersion) > 0 &&
							compareVersions(sourceVersionFull, tempSourceVersion) >= 0) {
						bestSourceVersion = tempSourceVersion; // updated
					}
				}	
			}
		}
			
		return bestSourceVersion;
	}
	
	/**
	 * return 1 if versionID1 is after versionID2, (versionID1 > versionID2)
	 *        0 if they are equal
	 *        -1 otherwise
	 * @param versionID1
	 * @param versionID2
	 * @return
	 */
	static public int compareVersions(int[] versionID1, int[] versionID2) {
		if(versionID1[0] > versionID2[0]) // compare major version
			return 1;
		else if(versionID1[0] == versionID2[0]) {
			if(versionID1[1] > versionID2[1])
				return 1;
			else if(versionID1[1] == versionID2[1])
				return 0;
		}
		return -1;
	}
	
	static public int compareVersions(int versionID1, int minorVersionID1, 
			int versionID2, int minorVersionID2) {
		int[] versionIDArray1 = {versionID1, minorVersionID1};
		int[] versionIDArray2 = {versionID2, minorVersionID2};
		return compareVersions(versionIDArray1, versionIDArray2);
	}
	
	private boolean checkInputQueryFootprintCorrectnessAgainstTargetVersionSchema(SQLUnionSet sqlUnionSet, ICSchema schema) {
		for(int i=0; i<sqlUnionSet.sqlUnions.size(); i++)
			if(!checkInputQueryFootprintCorrectnessAgainstTargetVersionSchema(
					sqlUnionSet.sqlUnions.get(i), schema))
				return false;
		
		return true;
	}
	private boolean checkInputQueryFootprintCorrectnessAgainstTargetVersionSchema(SQLUnion sqlUnion, ICSchema schema) {
		for(int i=0; i<sqlUnion.sqls.size(); i++)
			if(!checkInputQueryFootprintCorrectnessAgainstTargetVersionSchema(
					sqlUnion.sqls.get(i), schema))
				return false;
		
		return true;
	}
	private boolean checkInputQueryFootprintCorrectnessAgainstTargetVersionSchema(SQL sql, ICSchema schema) {
		// construct a footprint schema
		ICSchema footprint = new ICSchema();
		
		// table name
		for(int i=0; i<sql.from.relationDeclarationsSize(); i++) {
			RelationDeclaration rd = sql.from.relationDeclarations.get(i);
			if(!footprint.containsTable(rd.relation))
				footprint.addTable(new ICTable(rd.relation));
		}
		
		// column name from where clause
		for(int i=0; 
			sql.where.conjunctives.size() > 0 &&
			i<sql.where.conjunctives.get(0).size(); i++) {
			Predicate p = sql.where.conjunctives.get(0).get(i);
			if(p.term1.isRATerm()) {
				String relation = sql.from.getRelation(p.term1.tableName);
				ICTable table = footprint.getTable(relation);
				if(!table.hasColumn(p.term1.columnName))
					table.addColumn(p.term1.columnName);
			}
			if(p.term2.isRATerm()) {
				String relation = sql.from.getRelation(p.term2.tableName);
				ICTable table = footprint.getTable(relation);
				if(!table.hasColumn(p.term2.columnName))
					table.addColumn(p.term2.columnName);
			}
		}

		// column name from select clause
		for(int i=0; i<sql.select.selectAttributesSize(); i++) {
			SelectAttribute sa = sql.select.selectAttributes.get(i);	
			if(sa.attribute != null && sa.relationAlias != null) {
				String relation = sql.from.getRelation(sa.relationAlias);
				ICTable table = footprint.getTable(relation);
				if(!table.hasColumn(sa.attribute))
					table.addColumn(sa.attribute);
			}
		}

		return schema.containsSchema(footprint);
	}
}
