package edu.ucla.cs.db.prism.backend.rewrite;

import edu.ucla.cs.db.prism.backend.SQL.*;
import edu.ucla.cs.db.prism.backend.ded.*;
import edu.ucla.cs.db.prism.backend.exception.ErrorRewritingException;
import edu.ucla.cs.db.prism.backend.exception.RewritingOptimizationException;
import edu.ucla.cs.db.prism.backend.marsManage.MarsRunManage;
import edu.ucla.cs.db.prism.backend.schema.*;
import edu.ucla.cs.db.prism.backend.smo.util.*;

import java.util.*;

public class PrimRewriteEngine {
	final boolean checkInputQueryCorrectness = true;
	private SchemaHistory schemaHistory;
	private DEDSetHistory dedSetHistory;
	
	
	
	public PrimRewriteEngine(){
			
	}
	
	/**
	 * constructor
	 * @param schemaHistory
	 * @param dedSetHistory
	 */
	public PrimRewriteEngine(SchemaHistory schemaHistory, DEDSetHistory dedSetHistory) {
		this.schemaHistory = schemaHistory;
		this.dedSetHistory = dedSetHistory;
	}
	
	
	
	

	/**
	 * user interface for query rewriting
	 * @param inputSQLUnion the input (union of) query to rewrite
	 * @param sourceVersion the schema version where the source data is stored
	 * @param targetVersion the schema version where the query is posed
	 * @return
	 * @throws ErrorRewritingException 
	 */
	public SQLUnionSet rewrite(SQLUnionSet inputSQLUnionSet, 
			int sourceVersion, int sourceMinorVersion, 
			int targetVersion, int targetMinorVersion) throws ErrorRewritingException {
		
		// step 1.1.1: check query correctness based on target schema version
		if(checkInputQueryCorrectness) {
			boolean correctQuery
			= checkInputQueryFootprintCorrectnessAgainstTargetVersionSchema(
					inputSQLUnionSet, schemaHistory.getSchemaByVersion(targetVersion, targetMinorVersion));
			if(!correctQuery) {
				System.err.println("the input query is ill-posed against the target schema version: " + targetVersion +":"+ targetMinorVersion);
				//System.err.println(inputSQLUnionSet);
				//System.err.println(schemaHistory.getSchemaByVersion(targetVersion, targetMinorVersion).toStringWithMetadata());
				// changed behavior not to crash the interface at every wrong query... returning null should be ok..
				return null;
			}
		}		
		
		// step 1.2: version-tag the input query
		inputSQLUnionSet.versionTag(targetVersion,targetMinorVersion);
    	
		// step 1.3: extract query footprint on the target version
		Vector<String> queryFootprint = getTables(inputSQLUnionSet);

		
		// step 2: using a loop, find the best next intermediate source version to rewrite query to.
		//         note that this loop finishes when such next intermediate source version is
		//         same as the source version.
		//         otherwise, keep looping, using the chosen intermediate source version as 
		//		   the target version in the next loop
		//
		//         the next intermediate source version is the best next step 
		//		   given:
		//			i) the current target version (Vtgt)
		//          ii) the ultimate source version (Vsrc)
		//
		//         for such intermediate source version Vint, fetch/compute the following:
		//			i) the set of DEDs with src:Vint, tgt:Vtgt (both chase/backchase directions)
		//			ii) the query footprint QF' on Vint
		//		
		//		   among the set of (chase) DEDs chosen above, select only the deds that affect QF on Vtgt.
		//         source tables for such chase DEDs will constitute QF'
		//
		//		   note that we don't use ID-specific optimizations here
		int[] sourceVersionFull = {sourceVersion, sourceMinorVersion};
		int[] tempSourceVersion={-1,-1}; // major, minor version
		int[] tempTargetVersion = {targetVersion, targetMinorVersion};
		Vector<String> tempTargetVersionFootprint = queryFootprint;

		DEDSet tempDEDSet = null;
		DEDSet minTempDEDSet = null;
		DEDSetHistory minDEDSetHistory = new DEDSetHistory();
		Schema tempSchema = null;
		SchemaHistory queryFootprintSchemaHistory = new SchemaHistory();
		while(true) {
			// step 2.1: find the best tempSourceVersion, given sourceVersion and tempTargetVersion
			// composition network design principle: 
			//   greedy algorithm should return the optimal solution at each step
			tempSourceVersion = getBestTempSourceVersion(dedSetHistory, sourceVersionFull, tempTargetVersion);
			
			// FIXME in case the above algorithm returns a null tempSourceVersion set to the input
			if(tempSourceVersion == null){
				System.err.println("ERROR PrimRewriteEngine.getBestTempSourceVersion: tempSourceVersion was null");
				tempSourceVersion = sourceVersionFull;
			}
			
			// step 2.2: retrieve tempDEDSet from tempTargetVersion to tempSourceVerion
			tempDEDSet = dedSetHistory.get(tempSourceVersion, tempTargetVersion);
			
			if(tempDEDSet == null){
				
				System.err.println("NO DED IN THE dedSetHistory for versions:" + tempSourceVersion[0] + "." + tempSourceVersion[1] + " to " + tempTargetVersion[0] + "."+ tempTargetVersion[1]);
			}
				
			
			// step 2.3: minimize tempDEDSet into minTempDEDSet
			//           using tempTargetVersionFootprint and 
			//           tempDEDSet's targetVersion tables
			//           And add it to minDEDSetHistory
			
			minTempDEDSet = extractDEDsUsingTargetFootprint(tempDEDSet, tempTargetVersionFootprint);
			// quick fix
			//minTempDEDSet = tempDEDSet;
			
			
			minDEDSetHistory.add(minTempDEDSet);
			
			
			// step 2.4: add tempTargetVersionFootprint to queryFootprintSchemaHistory 
			tempSchema = getSchemaForTables(schemaHistory, tempTargetVersion, tempTargetVersionFootprint);
			//tempSchema = getSchemaForTables(schemaHistory, tempSourceVersion, tempTargetVersionFootprint);

			
			queryFootprintSchemaHistory.addSchema(tempSchema);
			
			// step 2.5: update tempTargetVersionFootprint into tempDEDSet's source tables
			tempTargetVersionFootprint = getSourceFootprint(minTempDEDSet);
			
			// step 2.6: break if we're at last step 
			//           (after adding last footprint on source version 
			//            to queryFootprintSchemaHistory)
			if(compareVersions(tempSourceVersion, sourceVersionFull) == 0) {
				tempSchema = getSchemaForTables(schemaHistory, sourceVersionFull,
						tempTargetVersionFootprint);
				queryFootprintSchemaHistory.addSchema(tempSchema);
				break;
			}
			
			// step 2.7: update tempTargetVersion
			tempTargetVersion = tempSourceVersion;
		}
		
		SQLUnionSet rewrittenSQLUnionSet = null;
		

		// step 3: call mars, using marsManage.MarsRunManage class
		 rewrittenSQLUnionSet =
			MarsRunManage.marsRunManage(inputSQLUnionSet, 
					queryFootprintSchemaHistory, minDEDSetHistory, 
					sourceVersion, sourceMinorVersion, targetVersion,targetMinorVersion, 
					null, false);
	
		// step 4: version-untag query
		if(rewrittenSQLUnionSet!=null)
			rewrittenSQLUnionSet.versionUntag();
		
		return rewrittenSQLUnionSet;
	}



	public SQLUnion rewrite(SQLUnion inputSQLUnion, int sourceVersion, int sourceMinorVersion, int targetVersion, int targetMinorVersion) throws ErrorRewritingException {
		// step 1: give a name to input query, if null, to prevent errors in the later stage
		if(inputSQLUnion.name == null)
			inputSQLUnion.name = "q1";
		// step 2: call rewrite(SQLUnionSet), which does actual rewriting
		SQLUnionSet inputSQLUnionSet = new SQLUnionSet(); // a singleton set
		inputSQLUnionSet.sqlUnions.add(inputSQLUnion);
		//System.out.println(sourceMinorVersion+" "+targetMinorVersion);
		SQLUnionSet rewrittenSQLUnionSet =
			rewrite(inputSQLUnionSet, sourceVersion, sourceMinorVersion,
				targetVersion, targetMinorVersion);
		if(rewrittenSQLUnionSet == null)
			return null;
		if(rewrittenSQLUnionSet.sqlUnions.isEmpty())
			return null;
		else
			return rewrittenSQLUnionSet.sqlUnions.get(0);
	}
	
	
	
	

	
	
	/**
	 * user interface for query rewriting
	 * @param inputSQLUnion the input query to rewrite
	 * @param sourceVersion the schema version where the source data is stored
	 * @param targetVersion the schema version where the query is posed
	 * @return
	 * @throws ErrorRewritingException 
	 */
	public SQLUnion rewrite(SQL inputSQL, int sourceVersion, int sourceMinorVersion, int targetVersion, int targetMinorVersion) throws ErrorRewritingException {
		SQLUnion inputSQLUnion = new SQLUnion(); // a singleton union
		inputSQLUnion.sqls.add(inputSQL);
		return rewrite(inputSQLUnion, sourceVersion, sourceMinorVersion, targetVersion, targetMinorVersion);
	}


	private Vector<String> getTables(SQLUnionSet sqlUnionSet) {
		if(sqlUnionSet.sqlUnions.size() == 1)
			return getTables(sqlUnionSet.sqlUnions.get(0));

		Vector<String> tables = new Vector<String>();
		for(int i=0; i<sqlUnionSet.sqlUnions.size(); i++) {
			tables.addAll(getTables(sqlUnionSet.sqlUnions.get(i)));
		}
		return tables;
	}
	private Vector<String> getTables(SQLUnion sqlUnion) {
		Vector<String> tables = new Vector<String>();
		for(int i=0; i<sqlUnion.sqls.size(); i++) {
			SQL sql = sqlUnion.sqls.get(i);
			for(int j=0; j<sql.from.relationDeclarationsSize(); j++) {
				RelationDeclaration rd = sql.from.relationDeclarations.get(j);
				if(!tables.contains(rd.relation))
					tables.add(rd.relation);
			}
		}
		return tables;
	}

	private DEDSet extractDEDsUsingTargetFootprint(DEDSet dedSet, 
			Vector<String> tempTargetVersionFootprint) {
		DEDSet minDEDSet = new DEDSet();
		minDEDSet.sourceVersion = dedSet.sourceVersion;
		minDEDSet.minorSourceVersion = dedSet.minorSourceVersion;
		minDEDSet.targetVersion = dedSet.targetVersion;
		minDEDSet.minorTargetVersion = dedSet.minorTargetVersion;
		
		String versionPrefix = null;
		
		for(int i=0; i<dedSet.size(); i++) {
			DED ded = dedSet.get(i);
			Vector<String> testTables = null;
			versionPrefix = "v" + dedSet.targetVersion + "m" + dedSet.minorTargetVersion + "_";

			if(ded.chaseDEDFlag) { // data-exchange direction, test its target tables
				testTables = ded.targetRelations;
			}
			else { // query-rewriting direction, test its source tables
				testTables = ded.sourceRelations;
			}
			
			// if there's an intersection between testTables and tempTargetVersionFootprint, 
			// include this ded into the return value
			for(int j=0; j<testTables.size(); j++) {
				if(tempTargetVersionFootprint.contains(versionPrefix + testTables.get(j))) {
					minDEDSet.add(ded);
					break;
				}
			}
		}
		
		return minDEDSet;
	}
	

	private Vector<String> getSourceFootprint(DEDSet dedSet) {
		Vector<String> sourceTables = new Vector<String>();
		
		for(int i=0; i<dedSet.size(); i++) {
			DED ded = dedSet.get(i);
			Vector<String> tempSourceTables = null;
			
			// CARLO: modified this to source... check it
			//String versionPrefix = "v" + dedSet.targetVersion + "m" + dedSet.minorTargetVersion + "_";
			
			String versionPrefix = "v" + dedSet.sourceVersion + "m" + dedSet.minorSourceVersion + "_";
			
			if(ded.chaseDEDFlag) { // data-exchange direction, get its source tables
				tempSourceTables = ded.sourceRelations;
			}
			else { // query-rewriting direction, get its target tables
				tempSourceTables = ded.targetRelations;
			}
			
			for(int j=0; j<tempSourceTables.size(); j++) {
				if(!sourceTables.contains(versionPrefix + tempSourceTables.get(j)))
					sourceTables.add(versionPrefix + tempSourceTables.get(j));
			}
		}
		
		return sourceTables;
	}
	private Schema getSchemaForTables(SchemaHistory schemaHistory, 
			int version[], Vector<String> tables) {
		Schema schema = schemaHistory.getSchemaByVersion(version);
		Schema schemaSubset = new Schema();
		schemaSubset.versionID = schema.versionID;
		schemaSubset.minorVersionID = schema.minorVersionID;

		for(int i=0; i<tables.size(); i++) {
			int index = tables.get(i).indexOf("_");
			String tableNameWithoutVersionPrefix = tables.get(i).substring(index+1);
			schemaSubset.addTable(schema.getTable(tableNameWithoutVersionPrefix));
		}
		return schemaSubset;
	}

	static public int[] getBestTempSourceVersion(DEDSetHistory dedSetHistory, 
			int sourceVersion[], int[] targetVersion) throws RewritingOptimizationException {
		
		int[] sourceVersionFull = sourceVersion;
		int[] bestSourceVersion = targetVersion;
		boolean targetGreaterThanSource = false;
		int test = compareVersions(targetVersion, sourceVersionFull);
		if(test == 0)
			return null; // error case
		else if(test > 0)
			targetGreaterThanSource = true;
		
		// find the next tempSourceVersion between targetVersion and sourceVersion
		// among all choices, choose one that is closest to sourceVersion
		
		
		if(targetGreaterThanSource){
			for(int i=dedSetHistory.size()-1; i>=0; i--) {
				DEDSet dedSet = dedSetHistory.get(i);
				int[] tempTargetVersion = {dedSet.targetVersion, dedSet.minorTargetVersion};
				int[] tempSourceVersion = {dedSet.sourceVersion, dedSet.minorSourceVersion};
				
				if(compareVersions(tempTargetVersion, targetVersion) == 0) {
					if(compareVersions(tempSourceVersion, bestSourceVersion) < 0 &&
						compareVersions(sourceVersionFull, tempSourceVersion) <= 0) {
						bestSourceVersion = tempSourceVersion; // updated
					}
				}
			}
		}
		else
		{
			for(int i=0; i<dedSetHistory.size(); i++) {
				DEDSet dedSet = dedSetHistory.get(i);
				int[] tempTargetVersion = {dedSet.targetVersion, dedSet.minorTargetVersion};
				int[] tempSourceVersion = {dedSet.sourceVersion, dedSet.minorSourceVersion};
				if(compareVersions(tempTargetVersion, targetVersion) == 0) {
					
					if( compareVersions(tempSourceVersion, bestSourceVersion) > 0 &&
							compareVersions(sourceVersionFull, tempSourceVersion) >= 0) {
						bestSourceVersion = tempSourceVersion; // updated
					}
				}	
			}
		}
			
		return bestSourceVersion;
	}
	
	/**
	 * return 1 if versionID1 is after versionID2, (versionID1 > versionID2)
	 *        0 if they are equal
	 *        -1 otherwise
	 * @param versionID1
	 * @param versionID2
	 * @return
	 */
	static public int compareVersions(int[] versionID1, int[] versionID2) {
		if(versionID1[0] > versionID2[0]) // compare major version
			return 1;
		else if(versionID1[0] == versionID2[0]) {
			if(versionID1[1] > versionID2[1])
				return 1;
			else if(versionID1[1] == versionID2[1])
				return 0;
		}
		return -1;
	}
	
	static public int compareVersions(int versionID1, int minorVersionID1, 
			int versionID2, int minorVersionID2) {
		int[] versionIDArray1 = {versionID1, minorVersionID1};
		int[] versionIDArray2 = {versionID2, minorVersionID2};
		return compareVersions(versionIDArray1, versionIDArray2);
	}
	
	private boolean checkInputQueryFootprintCorrectnessAgainstTargetVersionSchema(SQLUnionSet sqlUnionSet, Schema schema) {
		for(int i=0; i<sqlUnionSet.sqlUnions.size(); i++)
			if(!checkInputQueryFootprintCorrectnessAgainstTargetVersionSchema(
					sqlUnionSet.sqlUnions.get(i), schema))
				return false;
		
		return true;
	}
	private boolean checkInputQueryFootprintCorrectnessAgainstTargetVersionSchema(SQLUnion sqlUnion, Schema schema) {
		for(int i=0; i<sqlUnion.sqls.size(); i++)
			if(!checkInputQueryFootprintCorrectnessAgainstTargetVersionSchema(
					sqlUnion.sqls.get(i), schema))
				return false;
		
		return true;
	}
	private boolean checkInputQueryFootprintCorrectnessAgainstTargetVersionSchema(SQL sql, Schema schema) {
		// construct a footprint schema
		Schema footprint = new Schema();
		
		// table name
		for(int i=0; i<sql.from.relationDeclarationsSize(); i++) {
			RelationDeclaration rd = sql.from.relationDeclarations.get(i);
			if(!footprint.containsTable(rd.relation))
				footprint.addTable(new Table(rd.relation));
		}
		
		// column name from where clause
		for(int i=0; 
			sql.where.conjunctives.size() > 0 &&
			i<sql.where.conjunctives.get(0).size(); i++) {
			Predicate p = sql.where.conjunctives.get(0).get(i);
			if(p.term1.isRATerm()) {
				String relation = sql.from.getRelation(p.term1.tableName);
				Table table = footprint.getTable(relation);
				if(!table.hasColumn(p.term1.columnName))
					table.addColumn(p.term1.columnName);
			}
			if(p.term2.isRATerm()) {
				String relation = sql.from.getRelation(p.term2.tableName);
				Table table = footprint.getTable(relation);
				if(!table.hasColumn(p.term2.columnName))
					table.addColumn(p.term2.columnName);
			}
		}

		// column name from select clause
		for(int i=0; i<sql.select.selectAttributesSize(); i++) {
			SelectAttribute sa = sql.select.selectAttributes.get(i);	
			if(sa.attribute != null && sa.relationAlias != null) {
				String relation = sql.from.getRelation(sa.relationAlias);
				Table table = footprint.getTable(relation);
				if(!table.hasColumn(sa.attribute))
					table.addColumn(sa.attribute);
			}
		}

		return schema.containsSchema(footprint);
	}

}
