package edu.ucla.cs.db.prism.backend.rewrite.update;

import java.io.StringReader;
import java.util.HashMap;
import java.util.List;
import java.util.Vector;

import edu.ucla.cs.db.prism.backend.SQL.SQL;
import edu.ucla.cs.db.prism.backend.SQL.SelectAttribute;
import edu.ucla.cs.db.prism.backend.ded.DEDSetHistory;
import edu.ucla.cs.db.prism.backend.exception.UnimplementedFeatureException;
import edu.ucla.cs.db.prism.backend.icmo.ICMOSMOSequence;
import edu.ucla.cs.db.prism.backend.rewrite.RewriteEngineAdaptor;
import edu.ucla.cs.db.prism.backend.rewrite.RewritingCacheManager;
import edu.ucla.cs.db.prism.backend.rewrite.update.exception.UnexpectedNegationException;
import edu.ucla.cs.db.prism.backend.schema.exception.TableNotInSchemaException;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.ICSchema;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.ICSchemaHistory;
import edu.ucla.cs.db.prism.backend.smo.util.Predicate;
import edu.ucla.cs.db.prism.backend.smo.util.Term;

import net.sf.jsqlparser.JSQLParserException;
import net.sf.jsqlparser.expression.BinaryExpression;
import net.sf.jsqlparser.expression.Expression;
import net.sf.jsqlparser.expression.operators.relational.ExpressionList;
import net.sf.jsqlparser.expression.operators.relational.ItemsList;
import net.sf.jsqlparser.parser.CCJSqlParserManager;
import net.sf.jsqlparser.schema.Column;
import net.sf.jsqlparser.schema.Table;
import net.sf.jsqlparser.statement.Statement;
import net.sf.jsqlparser.statement.delete.Delete;
import net.sf.jsqlparser.statement.insert.Insert;
import net.sf.jsqlparser.statement.select.SubSelect;
import net.sf.jsqlparser.statement.update.Update;


/**
 * The UpdateRewriteEngine operates by: 1) generating for the input update a query-based representation which is the equivalence of pairs of queries, 2) rewrite independently each of these queries, 3) postprovcess
 * the result, by creating an equivalent outout SQL (might be more than one statement), based on the equivalence of the rewritten queries. Uses for the rewrite the RewriteEngineAdaptor that can rewrite queries 
 * through general  ICMOs and SMOs 
 *    
 * @author Carlo A. Curino (carlo@curino.us)
 *
 */
public class UpdateRewriteEngine {

	ICSchema queriedSchema = null;
	ICSchema sourceSchema = null;
	
	ICMOSMOSequence mosequence = null;
	boolean isInverse = false;
	RewriteEngineAdaptor rea = null;
	
	RewritingCacheManager rcm = null;
	
	public UpdateRewriteEngine(ICSchema queriedSchema, ICSchema sourceSchema, ICMOSMOSequence sequence, boolean isInverse){
		
		this.queriedSchema =queriedSchema;
		this.sourceSchema = sourceSchema;
		this.mosequence = sequence;
		this.isInverse = isInverse;
		
		rea = new RewriteEngineAdaptor(queriedSchema,sourceSchema,sequence,isInverse);
		this.rcm = new RewritingCacheManager();
	}
	
	/**
	 * Rewrite an SQL update through the evolution steps with which the class is instantiated (ICMOSMOSequence sequence)
	 * @param update
	 * @return
	 * @throws TableNotInSchemaException
	 * @throws UnexpectedNegationException
	 * @throws UnimplementedFeatureException
	 */
	public String rewriteUpdate(String update) throws TableNotInSchemaException, UnexpectedNegationException, UnimplementedFeatureException{
		
		//check the cache first 
		
		String rewriting = rcm.lookupQuery(update);
		
		if(rewriting!=null)
			return rewriting;
		
		// transform DML Update statement in a set of queries
		QueryBasedUpdateRep querypairs = update2query(update);
		
		// for each couple of queries rewrite them independently
		for(QueryPair qp:querypairs.querypairs){
			qp.query1 = rea.rewrite(qp.query1);
			qp.query2 = rea.rewrite(qp.query2);
		}
		
		// transform back the query pairs to an SQL DML update statement
		String updateSQLStatement = query2update(querypairs);
		
		//add this rewriting to the cache
		rcm.addRewritingByQuery(update, updateSQLStatement);
		
		return updateSQLStatement;
	}
	
	/**
	 * This methods invoke the postprocessing steps, depending on the type of operation performed. It transforms the query-based representation of updates back to SQL updates.
	 * @param querypairs
	 * @return
	 * @throws UnexpectedNegationException
	 * @throws UnimplementedFeatureException
	 */
	private String query2update(QueryBasedUpdateRep querypairs) throws UnexpectedNegationException, UnimplementedFeatureException {
		
		//System.out.println(querypairs);
		
		if(querypairs.statement instanceof Delete){
			return postprocessDelete(querypairs);			
		}
		if(querypairs.statement instanceof Update){
			return postprocessUpdate(querypairs);			
		}
		if(querypairs.statement instanceof Insert){
			return postprocessInsert(querypairs);			
		}
		return null;
	}

	
	/**
	 * This method invokes appropriate handlers for each type of updates to represent the SQL udpate as a set of equivalent query pairs. 
	 * @param update
	 * @return
	 * @throws TableNotInSchemaException
	 * @throws UnimplementedFeatureException
	 */
	public QueryBasedUpdateRep update2query(String update) throws TableNotInSchemaException, UnimplementedFeatureException{
		
		// parse the update and invoke the appropriate submethod
		CCJSqlParserManager pm = new CCJSqlParserManager();
		try {
			net.sf.jsqlparser.statement.Statement statement = pm.parse(new StringReader(update));
	
			if(statement instanceof Update){		
				return prepareUpdate((Update)statement);
			}

			if(statement instanceof Insert){		
				return prepareInsert((Insert)statement);
			}

			if(statement instanceof Delete){		
				return prepareDelete((Delete)statement);
			}
		} catch (JSQLParserException e) {
			e.printStackTrace();
		}
		
		// return the querybasedupdaterep
		return null;
		
	}


	//---------------------------- PREPROCESSING METHODS --------------------------------
	
	
	private QueryBasedUpdateRep prepareDelete(Delete statement) throws TableNotInSchemaException {
		QueryBasedUpdateRep res = new QueryBasedUpdateRep(statement);
		
		Table table = statement.getTable();
		if(queriedSchema.getTable(table.getName()) == null)
			throw new TableNotInSchemaException("Table " + table.getName() + " is not present in the target schema");
		Vector<String> columns = queriedSchema.getTable(table.getName()).getColumns();
	
		String query1 = "SELECT ";
		
		for(String col:columns){
			query1 += table.getName() + "." + col + ", ";
		}
		
		// add the alias to the WHERE clause, needed for rewriting
		table.setAlias(table.getName());
		Expression p = statement.getWhere();
		ExpressionAliasInjectorVisitor.aliasExpression(p, table, queriedSchema.getTable(table.getName()));
		
		query1 = query1.substring(0, query1.length()-2);
		query1 += " FROM " + table.getName() + " " + table.getName() +" WHERE " + p.toString();
		
		
		String query2 = "SELECT ";
		
		for(String col:columns){
			query2 += table.getName() + "." + col + ", ";
		}
		
		query2 = query2.substring(0, query2.length()-2);
		query2 += " FROM " + table.getName() + " " + table.getName();
		
		
		// add the two queries, indicating that query1 is negated. This avoid rewrite expression and simplify the process back to SQL
		res.add(new QueryPair(query1,query2,true,false));	
		
		return res;
	}


	private QueryBasedUpdateRep prepareInsert(Insert statement) throws UnimplementedFeatureException {
		QueryBasedUpdateRep res = new QueryBasedUpdateRep(statement);
		
		Table table = statement.getTable();
		List<Column> columns = statement.getColumns();
		ItemsList values = statement.getItemsList();
		String query1 = "";
		String query2 = "";
		// create query1 differently based on the type of INSERT we are dealing with
		if(values instanceof ExpressionList){
			query1 = "SELECT";
			for(Object exp:((ExpressionList)values).getExpressions()){
				query1+= " " + exp.toString() + ",";
			} 
			query1 = query1.substring(0,query1.length()-1);
			query1 += " FROM " + table.getName() + " " + table.getName();
		}
		
		if(values instanceof SubSelect){	
			
			
			SelectFromTableVisitor s = new SelectFromTableVisitor();
			((SubSelect) values).getSelectBody().accept(s);
			
			
			for(Table ta:s.tables){
				SelectAliasInjectorVisitor v = new SelectAliasInjectorVisitor(ta,queriedSchema.getTable(ta.getName()));
				((SubSelect) values).getSelectBody().accept(v);
			}
		
			query1 = ((SubSelect) values).getSelectBody().toString().replaceAll(" AS ", " ");
		}
		
		// create the target query2
		query2 = "SELECT ";
		for(Column o:columns){
			query2 += table.getName()+"."+o.getColumnName() + ", ";
		}
		query2 = query2.substring(0,query2.length()-2);
		query2 += " FROM " + table.getName() + " " + table.getName();
		
		res.querypairs.add(new QueryPair(query1,query2,false,false));
		
		return res;
	}



	private QueryBasedUpdateRep prepareUpdate(Update statement) throws UnimplementedFeatureException, TableNotInSchemaException {
		QueryBasedUpdateRep res = new QueryBasedUpdateRep(statement);

		
		Table table = statement.getTable();
		Expression where = statement.getWhere();
		List<Column> updatedcolumns = statement.getColumns();
		List<Expression> expressions = statement.getExpressions();
		
		//fixing aliases for expressions in the SET 
		table.setAlias(table.getName());
		for(Expression p:expressions){
			ExpressionAliasInjectorVisitor.aliasExpression(p, table, queriedSchema.getTable(table.getName()));			
		}
		
			
		if(queriedSchema.getTable(table.getName()) == null)
			throw new TableNotInSchemaException("Table " + table.getName() + " is not present in the target schema");
		Vector<String> columns = queriedSchema.getTable(table.getName()).getColumns();
	
		// build in parallel query1 and query2, simply putting expression in the right place of query1 
		String query1 = "SELECT ";
		String query2 = "SELECT ";
		
		for(String col:columns){
			boolean processed = false;
			for(int i=0;i<updatedcolumns.size();i++){
				if(updatedcolumns.get(i).getColumnName().equals(col)){	
					query1 += expressions.get(i).toString() + ", ";
					processed = true;
					break;
				}
			}
			
			if(!processed)
				query1 += table.getName() + "." + col + ", ";
			
			query2 += table.getName() + "." + col + ", ";
			
		}
		
		
		// add the alias to the WHERE clause, needed for rewriting
		table.setAlias(table.getName());
		Expression p = statement.getWhere();
		ExpressionAliasInjectorVisitor.aliasExpression(p, table, queriedSchema.getTable(table.getName()));
		
		query1 = query1.substring(0, query1.length()-2);
		query1 += " FROM " + table.getName() + " " + table.getName() +" WHERE " + p.toString();
		
		query2 = query2.substring(0, query2.length()-2);
		query2 += " FROM " + table.getName() + " " + table.getName() +" WHERE " + p.toString();
		
		// query3 and query4 are the same of query2, just with a negated condition (handled separately in the QueryPair), it specifies what is not changed by the update
		String query3 = new String(query2);
		String query4 = new String(query2);
	
		// adding the two pairs of queries to the queryset and return
		res.add(new QueryPair(query1,query2,false,false));	
		res.add(new QueryPair(query3,query4,true,true));	
		
		return res;
		
				
	}
	
	
	//---------------------------- POSTPROCESSING METHODS --------------------------------
	
	
	private String postprocessUpdate(QueryBasedUpdateRep querypairs) {
		
		String output = "";
		
		// for the simple cases we do not need the second set of queries, which are implied
		QueryPair qp = querypairs.querypairs.elementAt(0);
	
		SQL q1neg = SQL.extractNegative(qp.query1);
		SQL q2neg = SQL.extractNegative(qp.query2);
	
		SQL q1pos = SQL.extractExternalPositive(qp.query1);
		SQL q2pos = SQL.extractExternalPositive(qp.query2);
		
		output += "UPDATE " + q2pos.from.toTablesOnlyString() + " SET ";
		
		for(int i=0;i<q1pos.select.selectAttributesSize();i++){
			// if the two string are not identical, than 
			if(!q2pos.select.get(i).toViewString().equals(q1pos.select.get(i).toViewString())){
				
				// build a Predicate from the SET portion we are about to insert to check whether is implied by the WHERE clause
				Term t1,t2 = null;
				if(q1pos.select.get(i).relationAlias!=null && q1pos.select.get(i).attribute!=null)
					t1 = new Term(q1pos.select.get(i).relationAlias, q1pos.select.get(i).attribute);
				else
					t1 = new Term(q1pos.select.get(i).toViewString());
				if(q2pos.select.get(i).relationAlias!=null && q2pos.select.get(i).attribute!=null)
					t2 = new Term(q2pos.select.get(i).relationAlias, q2pos.select.get(i).attribute);
				else
					t2 = new Term(q2pos.select.get(i).toViewString());
			
				Predicate test = new Predicate(t2,"=", t1);
				
				SelectAttribute attribute= q2pos.select.get(i);
				Vector<Predicate> vp = q2pos.where.getPredicatesByAttribute(attribute);
				
				
				// avoid setting attribute as we use them in the WHERE 
				if(!vp.contains(test)){
					output += q2pos.select.get(i).toViewString() + "=" + q1pos.select.get(i).toViewString() + ", ";
				
					if(vp != null && !vp.isEmpty()){
						for(Predicate p:vp){
							if( attribute.attribute.equals(p.term2.columnName) && 
							    attribute.relationAlias.equals(p.term2.tableName) && 
							    p.op.equals("=") &&
							    p.term1.columnName != null &&
							    p.term1.tableName != null){
								
								output += p.term1.tableName +"." + p.term1.columnName + "=" + q1pos.select.get(i).toViewString() + ", ";
											
							}
							
							if(attribute.attribute.equals(p.term1.columnName) && 
								attribute.relationAlias.equals(p.term1.tableName) && 
								p.op.equals("=") &&
								p.term2.columnName != null &&
								p.term2.tableName != null){
								output += p.term2.tableName +"." + p.term2.columnName + "=" + q1pos.select.get(i).toViewString() + ", ";
								
							}	
						}		
					}	
				}
			}	
		}
		output = output.substring(0,output.length()-2);
	
		output += " WHERE " + q1pos.where.toString();
		output += " AND NOT EXISTS (" + q1neg.toViewString(null) + ");";

		return output;
	}

	private String postprocessInsert(QueryBasedUpdateRep querypairs) throws UnimplementedFeatureException {
		
		//System.out.println(querypairs);
		
		if(((Insert)querypairs.statement).getItemsList() instanceof ExpressionList){
			
			QueryPair onlypair = querypairs.querypairs.firstElement();
			
			SQL q1 = SQL.parseSQLString(onlypair.query1);
			SQL q2 = SQL.parseSQLString(onlypair.query2);
			

			String output = "";
			
			HashMap hmcolumn = new HashMap();
			HashMap hmvalue = new HashMap();
			

			for(String alias:q2.from.getRelationAliases()){
				
				if(!hmcolumn.containsKey(alias))
					hmcolumn.put(alias, "INSERT INTO " + q2.from.getRelation(alias) + " (");
				if(!hmvalue.containsKey(alias))
					hmvalue.put(alias, " VALUES (");
			
			}
			
			for(int i=0; i<q2.select.selectAttributesSize();i++){
				
				SelectAttribute value = q1.select.get(i);
				SelectAttribute attribute = q2.select.get(i);
				
				String alias2 = null;
				
				Vector<Predicate> vp = q2.where.getPredicatesByAttribute(attribute);
				if(vp != null && !vp.isEmpty()){
					for(Predicate p:vp){
						if( attribute.attribute.equals(p.term2.columnName) && 
						    attribute.relationAlias.equals(p.term2.tableName) && 
						    p.op.equals("=") &&
						    p.term1.columnName != null &&
						    p.term1.tableName != null){
							
							alias2 = p.term1.tableName;
							
						}
						
						if(attribute.attribute.equals(p.term1.columnName) && 
							attribute.relationAlias.equals(p.term1.tableName) && 
							p.op.equals("=") &&
							p.term2.columnName != null &&
							p.term2.tableName != null){
							alias2 = p.term2.tableName;				
						}	
					}		
				}
				
				String alias = q2.select.get(i).relationAlias;
				
				hmcolumn.put(alias, ((String)hmcolumn.get(alias)) + attribute.attribute +",");
				hmvalue.put(alias, ((String)hmvalue.get(alias)) + value.getSingleQuoteConstant() +",");
				if(alias2!=null){	
					hmcolumn.put(alias2, ((String)hmcolumn.get(alias2)) + attribute.attribute +",");
					hmvalue.put(alias2, ((String)hmvalue.get(alias2)) + value.getSingleQuoteConstant() +",");
				}
			}
			
			for(Object o:hmcolumn.keySet()){
				output += ((String)hmcolumn.get(o));
				output = output.subSequence(0,output.length()-1) + ")";
				
				output += ((String)hmvalue.get(o));
				output = output.subSequence(0,output.length()-1) + ");\n";
				
			}
						
			return output;
			
		}else{
			// --------- manage SubSelect case ----------------------
			
			QueryPair onlypair = querypairs.querypairs.firstElement();
			
			SQL q1 = SQL.parseSQLString(onlypair.query1);
			SQL q2 = SQL.parseSQLString(onlypair.query2);
			

			String output = "";
			
			HashMap hmcolumn = new HashMap();
			HashMap hmvalue = new HashMap();
			

			for(String alias:q2.from.getRelationAliases()){
				
				if(!hmcolumn.containsKey(alias))
					hmcolumn.put(alias, "INSERT INTO " + q2.from.getRelation(alias) + " (");
				if(!hmvalue.containsKey(alias))
					hmvalue.put(alias, " SELECT ");
			
			}
			
			for(int i=0; i<q2.select.selectAttributesSize();i++){
				
				SelectAttribute value = q1.select.get(i);
				SelectAttribute attribute = q2.select.get(i);
				
				String alias2 = null;
				
				Vector<Predicate> vp = q2.where.getPredicatesByAttribute(attribute);
				if(vp != null && !vp.isEmpty()){
					for(Predicate p:vp){
						if( attribute.attribute.equals(p.term2.columnName) && 
						    attribute.relationAlias.equals(p.term2.tableName) && 
						    p.op.equals("=") &&
						    p.term1.columnName != null &&
						    p.term1.tableName != null){
							
							alias2 = p.term1.tableName;
							
						}
						
						if(attribute.attribute.equals(p.term1.columnName) && 
							attribute.relationAlias.equals(p.term1.tableName) && 
							p.op.equals("=") &&
							p.term2.columnName != null &&
							p.term2.tableName != null){
							alias2 = p.term2.tableName;				
						}	
					}		
				}
				
				String alias = q2.select.get(i).relationAlias;
				
				hmcolumn.put(alias, ((String)hmcolumn.get(alias)) + attribute.attribute +",");
				if(value.attribute != null)
					hmvalue.put(alias, ((String)hmvalue.get(alias)) + value.attribute +",");
				else
					hmvalue.put(alias, ((String)hmvalue.get(alias)) + value.getSingleQuoteConstant() +",");
					
				if(alias2!=null){	
					hmcolumn.put(alias2, ((String)hmcolumn.get(alias2)) + attribute.attribute +",");
					
					if(value.attribute != null)
						hmvalue.put(alias2, ((String)hmvalue.get(alias2)) + value.attribute +",");
					else
						hmvalue.put(alias2, ((String)hmvalue.get(alias2)) + value.getSingleQuoteConstant() +",");
					}
			}
			
			for(Object o:hmcolumn.keySet()){
				output += ((String)hmcolumn.get(o));
				output = output.subSequence(0,output.length()-1) + ")";
				
				output += ((String)hmvalue.get(o));
				output = output.subSequence(0,output.length()-1) + " "+q1.from + " WHERE " + q1.where + ";\n";
				
			}
						
			return output;
		}
		
	}

	private String postprocessDelete(QueryBasedUpdateRep querypairs) throws UnexpectedNegationException, UnimplementedFeatureException {
		
		QueryPair onlypair = querypairs.querypairs.firstElement();
		String out = "";
		if(onlypair.negated2)
			throw new UnexpectedNegationException("The second query of a QueryPair shouldn't be negated");
		
		if(onlypair.negated1){
			SQL neg = SQL.extractNegative(onlypair.query1);		
			SQL sql = SQL.extractExternalPositive(onlypair.query1);
			
			// check first the simple case in which there is still one single table
			if(sql.from.relationDeclarationsSize()==1){
				out = "DELETE " + sql.from.toString() + " WHERE " + sql.where;
				
			}else{		
				//deal with the more complex case in which more than one table is involved
				out = "DELETE " + sql.from.toAliasString() +  " FROM "+sql.from.toTablesOnlyString()+ " WHERE " + sql.where;
			}	
			if(neg!=null)
				out+= " AND NOT EXISTS (" + neg.toViewString(null) + ");";
		}
		
		return out;
	}


	
	
	
	private String removequotes(String string) {

		if(string.charAt(0) == '"'){
		
			if(string.charAt(string.length()-1) == '"'){
				//System.out.println(string + " -> " + string.substring(1,string.length()-1));

				return string.substring(1,string.length()-1);
		
			}
		}	
		return string;
	}

	public void printStatus() {

		System.out.println(rea.toString());
	} 
}
