package edu.ucla.cs.db.prism.backend.schema;

import java.io.Serializable;
import java.sql.Timestamp;
import java.util.Iterator;
import java.util.Vector;

import edu.ucla.cs.db.prism.backend.exception.SchemaErrorException;
import edu.ucla.cs.db.prism.backend.icmo.ICMOImpl;
import edu.ucla.cs.db.prism.backend.icmo.MOImpl;
import edu.ucla.cs.db.prism.backend.icmo.MOInterface;
import edu.ucla.cs.db.prism.backend.schema.exception.ColumnExistsException;
import edu.ucla.cs.db.prism.backend.schema.exception.ColumnNotInTableException;
import edu.ucla.cs.db.prism.backend.schema.exception.IllegalAddColumnArgumentException;
import edu.ucla.cs.db.prism.backend.schema.exception.IllegalSMOException;
import edu.ucla.cs.db.prism.backend.schema.exception.InputTableReplicatedException;
import edu.ucla.cs.db.prism.backend.schema.exception.MergingDifferentTablesException;
import edu.ucla.cs.db.prism.backend.schema.exception.SameColumnNameException;
import edu.ucla.cs.db.prism.backend.schema.exception.TableConstrainedException;
import edu.ucla.cs.db.prism.backend.schema.exception.TableExistsException;
import edu.ucla.cs.db.prism.backend.schema.exception.TableNotInSchemaException;
import edu.ucla.cs.db.prism.backend.schema.exception.TableHasNoColumns;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.ICSchema;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.ICTable;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.exceptions.ColumnIsConstrainedException;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.exceptions.ColumnsAndTypesMismatch;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.exceptions.ExpressionParsignErrorException;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.exceptions.InboundForeignKeyOnMergeException;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.exceptions.InboundForeignKeyOnPartitionException;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.exceptions.IntegrityConstraintsExistsException;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.exceptions.OutboundForeignKeyOnMergeException;
import edu.ucla.cs.db.prism.backend.smo.JoinTableSMO;
import edu.ucla.cs.db.prism.backend.smo.SMOImpl;
import edu.ucla.cs.db.prism.backend.smo.exception.NoTableInSchemaException;
import edu.ucla.cs.db.prism.backend.smo.exception.noColumnsMatchInfPres;
import edu.ucla.cs.db.prism.backend.smo.exception.noCommonColumnInfPres;
import edu.ucla.cs.db.prism.backend.smo.exception.notInfPresException;
import edu.ucla.cs.db.prism.backend.smo.util.SMOSequenceInterface;
import edu.ucla.cs.db.prism.exception.NotImplementedException;

/**
 * This class represent the internal reresentation of the Schema, which is evolved by SMOs
 * 
 * @author Carlo A. Curino (carlo@curino.us)
 */
public class Schema implements Serializable, Comparable {
	static final long serialVersionUID = 0;
	private Vector<Table> tables;	
	public int versionID;
	public int minorVersionID; //created by Hyun 1/23/2008
	protected String schemaName = null;
	public boolean isHTableSchema = false; // used for PRIMA HTable management (Hyun 3/24/2008)
	
	public String DEFAULT_CHARACTER_SET_NAME;
	public String DEFAULT_COLLATION_NAME;
	public String SQL_PATH;
	public String comment;
	public String REVISION_NUMBER;
	public Timestamp tstart;
	public Timestamp tend;
	
	
	public Schema(int versionID, int minorVersionID, String schemaName,
			String default_character_set_name, String default_collation_name,
			String sql_path, String comment, String revision_number,
			Timestamp tstart, Timestamp tend) {
		super();
		this.versionID = versionID;
		this.minorVersionID = minorVersionID;
		this.setSchemaName(schemaName);
		DEFAULT_CHARACTER_SET_NAME = default_character_set_name;
		DEFAULT_COLLATION_NAME = default_collation_name;
		SQL_PATH = sql_path;
		this.comment = comment;
		REVISION_NUMBER = revision_number;
		this.tstart = tstart;		
		this.tend = tend;		

		if(SQL_PATH == null)
			SQL_PATH=" ";
		if(comment == null)
			comment = "";
		setTables(new Vector<Table>());
	}

	public Schema () {
		setTables(new Vector<Table>());
		versionID = -1;
	}

	public Schema (int versionID) {
		setTables(new Vector<Table>());
		this.versionID = versionID;
	}

	public Schema (Schema srcSchema) {
		this.setSchemaName(srcSchema.getSchemaName());
		this.setTables(new Vector<Table>());
		for(int i=0; i<srcSchema.size(); i++)
			this.addTable(new Table(srcSchema.getTable(i)));
		this.versionID = srcSchema.versionID;
		this.minorVersionID = srcSchema.minorVersionID;
		this.tstart = srcSchema.tstart;		
		this.tend = srcSchema.tend;	
		
	}

	public Schema clone() {
		return new Schema(this);
	}
	
	public void addTable(Table table) {
		getTables().add(table);
	}
	
	public Table getTable(int i) {
		return getTables().get(i);
	}
	
	public Table getTable(String tablename) {
		for(int i=0; i<getTables().size(); i++)
			if(getTables().get(i).getTableName().equalsIgnoreCase(tablename)) {
				//System.out.println("tables.get("+i+") : "+tables.get(i));
				return getTables().get(i);
			}
		//	System.out.println(tablename + "NOT FOUND");
		//	System.out.println(this.tables);
		return null;
	}
	
	/*
	 * @return all tables and columns in this schema
	 */
	public Vector<String> getAllSchemaParts() {
		Vector<String> allSchemaParts = new Vector<String>(getTables().size()*20);
		for(int i=0; i<getTables().size(); i++) {
			allSchemaParts.add(getTables().get(i).getTableName());
			for(int j=0; j<getTables().get(i).getNumColumns(); j++) {
				allSchemaParts.add(getTables().get(i).getTableName() + "." 
						+ getTables().get(i).getColumn(j));
			}
		}
		return allSchemaParts;
	}

	/*
	 * @return all tables in this schema
	 */
	public Vector<String> getAllTables() {
		Vector<String> allTables = new Vector<String>(getTables().size());
		for(int i=0; i<getTables().size(); i++) {
			allTables.add(getTables().get(i).getTableName());
		}
		return allTables;
	}

	public int indexOf(String tableName) {
		for(int i=0; i<size(); i++)
			if(getTables().get(i).getTableName().contentEquals(tableName))
				return i;
		return -1;
	}

	// check to see if this schema contains a matching table as table2
	// table name should match, column name should match, and column order should match
	public int indexOf(Table table2) {
		for(int i=0; i<getTables().size(); i++) {
			Table table = getTables().get(i);

			// test 1: table name match
			if(!table.getTableName().equals(table2.getTableName()))
				continue;
			
			// test 2: table cardinality match
			if(table.getColumns().size() != table2.getColumns().size())
				continue;
			
			// test 3: individual columns
			boolean allColumnsMatch = true;
			for(int j=0; j<table.getColumns().size(); j++) {
				if(!table.getColumns().get(j).equals(table2.getColumns().get(j))) {
					allColumnsMatch = false;
					break;
				}
			}
			if(allColumnsMatch)
				return i;
		}
		return -1;
	}
	
	public boolean containsTable(Table table2) {
		int index = indexOf(table2);
		if(index<0) 
			return false;
		else 
			return true;
	}
	
	public boolean containsTable(String tableName) {
		int index = indexOf(tableName);
		if(index<0) 
			return false;
		else 
			return true;
	}
	
	public boolean containsColumn(String tablename, String columnname) {
		boolean found = false;
		Table table = null;
				
		for(int i=0; i<getTables().size(); i++) {
			if(getTables().get(i).getTableName().contentEquals(tablename)) {
				table = getTables().get(i);
				break;
			}
		}
		
		if(table == null)			
			return false;
		else if(table.indexOf(columnname) > -1)
			return true;
		else 
			return false;
	}

	
	public int size() {
		return getTables().size();
	}
	
	public String toString() {
		return getTables().toString();
	}

	public String toStringWithMetadata() {
		return "Schema version: v" + versionID + "m" + minorVersionID + "\n" + 
			getTables().toString();
	}
	
	
	
	public Schema cloneApplySMO(SMOImpl smo) throws IllegalSMOException{
		Schema newSchema = this.clone();
		newSchema.applySMO(smo);
		return newSchema;
	} 
	
	/**
	 * Override SMO target Major and Minor
	 * @param smo
	 * @param major
	 * @param minor
	 * @return
	 * @throws IllegalSMOException 
	 */
	public boolean applySMO(SMOImpl smo, int major, int minor) throws IllegalSMOException {

		boolean app = applySMO(smo);
		this.minorVersionID = minor;
		this.versionID = major;
		return app;
	}	
	
	
	
	public boolean applyOperator(MOInterface operator) throws SchemaErrorException, IllegalSMOException, ExpressionParsignErrorException{
		
		if(operator instanceof SMOImpl)
			return applySMO((SMOImpl) operator);
		if(operator instanceof ICMOImpl){
			if(this instanceof ICSchema){
				return ((ICSchema)this).applyICMO((ICMOImpl) operator);
			}
			
			throw new SchemaIsNotICAwareException("This schema is not IC-aware and you are trying to issues an ICMO on it.");
		}		
		
		return false;
	}
	
	
	/**
	 * Apply the SMO to the current Schema
	 * @param impl
	 * @return
	 * @throws ColumnsAndTypesMismatch 
	 * @throws  
	 * @throws  
	 */
	public boolean applySMO(SMOImpl impl) throws IllegalSMOException{
		
		String operatorType = impl.getOperatorType();
		//int size = this.size();
		
		if(operatorType.contentEquals("NOP")) {
			return applyNopSMO(impl);
		}
		if(operatorType.contentEquals("CREATE TABLE")) {
			return applyCreateTableSMO(impl);
		}
		else if(operatorType.contentEquals("DROP TABLE")) {
			return applyDropTableSMO(impl);
		}
		else if(operatorType.contentEquals("RENAME TABLE")) {
			return applyRenameTableSMO(impl);		
		}
		else if(operatorType.contentEquals("COPY TABLE")) {
			return applyCopyTableSMO(impl);		
		}
		else if(operatorType.contentEquals("PARTITION TABLE")) {
			return applyPartitionTableSMO(impl);
		}
		else if(operatorType.contentEquals("MERGE TABLE")) {
			return applyMergeTableSMO(impl);			
		}
		else if(operatorType.contentEquals("ADD COLUMN")) {
			return applyAddColumnSMO(impl);	
		}
		else if(operatorType.contentEquals("DROP COLUMN")) {
			return applyDropColumnSMO(impl);
		}
		else if(operatorType.contentEquals("RENAME COLUMN")) {
			return applyRenameColumnSMO(impl);
		}
		else if(operatorType.contentEquals("JOIN TABLE")) {
			return applyJoinTableSMO(impl);
		}
		else if(operatorType.contentEquals("DECOMPOSE TABLE")) {
			return applyDecomposeTableSMO(impl);
		}

		
		return false;
	}

	/**
	 * @param impl
	 * @return
	 */
	public boolean applyNopSMO(MOImpl impl) { 
		this.versionID = impl.getTargetMajor();
		this.minorVersionID = impl.getTargetMinor();
		return true;
	}

	/**
	 * @param smo
	 * @return
	 * @throws TableNotInSchemaException 
	 * @throws TableExistsException 
	 * @throws ColumnNotInTableException 
	 * @throws NoTableInSchemaException 
	 * @throws noCommonColumnInfPres 
	 * @throws noColumnsMatchInfPres 
	 * @throws IllegalSMOException 
	 */
	public boolean applyDecomposeTableSMO(SMOImpl smo) throws TableNotInSchemaException, TableExistsException, ColumnNotInTableException, noColumnsMatchInfPres, noCommonColumnInfPres, NoTableInSchemaException, IllegalSMOException {
		if(indexOf(smo.getTableName()) < 0) {
			//System.err.println("Decomposed table is not in the schema\n" +
			//		this.toString() + "\n" + smo);
			throw new TableNotInSchemaException("Source table" +smo.getTableName() +" is not in the schema\n"); 
		}
		else if(indexOf(smo.getTableName2()) >= 0) {
			// if new table is the same as source table name, then ok
			// since the source table will be deleted
			if(!smo.getTableName().equals(smo.getTableName2())) {
				//System.err.println("Table with new name is already in the schema\n" +
				//		this.toString() + "\n" + smo);
				throw new TableExistsException("Target table" +smo.getTableName2() +" is already in the schema\n"); 
			}
		}
		else if(indexOf(smo.getTableName3()) >= 0) {
			// if new table is the same as source table name, then ok
			// since the source table will be deleted
			if(!smo.getTableName().equals(smo.getTableName3())) {
				//System.err.println("Table with new name is already in the schema\n" +
				//		this.toString() + "\n" + smo);
				throw new TableExistsException("Target table" +smo.getTableName3() +" is already in the schema\n"); 
			}
		}
		
		// check attributes - they should exist in the source schema
		for(int i=0; smo.getAttributeList2() != null &&
						i<smo.getAttributeList2().size(); i++) {
			String attr = smo.getAttributeList2().get(i);
			if(!containsColumn(smo.getTableName(), attr)) {
				//System.err.println("Used attribute is not in the schema\n" +
				//		this.toString() + "\n" + smo);
				throw new ColumnNotInTableException("The column " + attr +" is not present in Table "+ smo.getTableName()+ " in the schema\n"); 
			}
		}
		for(int i=0; smo.getAttributeList3() != null &&
						i<smo.getAttributeList3().size(); i++) {
			String attr = smo.getAttributeList3().get(i);
			if(!containsColumn(smo.getTableName(), attr)) {
				//System.err.println("Used attribute is not in the schema\n" +
				//		this.toString() + "\n" + smo);
				throw new ColumnNotInTableException("The column " + attr +" is not present in Table "+ smo.getTableName()+ " in the schema\n"); 
			}
		}

		
		
		int index = this.indexOf(smo.getTableName());
		if(index<0) 
			throw new TableNotInSchemaException("Source table" +smo.getTableName() +" is not in the schema\n"); 
		
		Table intab = this.getTable(smo.getTableName());
		
		
		Table table2 = new Table(smo.getTableName2());
		Table table3 = new Table(smo.getTableName3());
		table2.setColumns((Vector<String>)smo.getAttributeList2().clone());
		
		Iterator<String> col = table2.getColumns().iterator();
		while(col.hasNext()){
			table2.getColTypes().add(intab.getColType(col.next()));
		}
		
		table3.setColumns((Vector<String>)smo.getAttributeList3().clone());

		col = table3.getColumns().iterator();
		while(col.hasNext()){
			table3.getColTypes().add(intab.getColType(col.next()));
		}
		
		getTables().remove(index);
		getTables().add(index, table2);
		getTables().add(index+1, table3);
		this.versionID = smo.getTargetMajor();
		this.minorVersionID = smo.getTargetMinor();
		return true;
	}

	/**
	 * @param smo
	 * @return
	 * @throws ColumnNotInTableException 
	 * @throws ColumnExistsException 
	 * @throws TableNotInSchemaException 
	 */
	public boolean applyRenameColumnSMO(SMOImpl smo) throws ColumnNotInTableException, ColumnExistsException, TableNotInSchemaException {
		int index = this.indexOf(smo.getTableName());
		
		
		if(index < 0) {
			//System.err.println("Renaming column in non-existent table\n" +
			//		this.toString() + "\n" + smo);
			throw new TableNotInSchemaException("Source table" +smo.getTableName() +" is not in the schema\n"); 
		}
		else if(getTable(index).indexOf(smo.getColumnName()) < 0) {
			//System.err.println("Column to rename is not in the table\n" +
			//		this.toString() + "\n" + smo);
			throw new ColumnNotInTableException("The source column "+ smo.getColumnName() +" to rename is not in the table\n");
		}
		else if(getTable(index).indexOf(smo.getColumnName2()) >= 0) {
			//System.err.println("Column with new name is already in the table\n" +
			//		this.toString() + "\n" + smo);
			throw new ColumnExistsException("The target column "+ smo.getColumnName2() +" exists\n");
		}
		int tableIndex = this.indexOf(smo.getTableName());
		if(tableIndex<0) 
			throw new TableNotInSchemaException("Source table" +smo.getTableName() +" is not in the schema\n"); 
		int columnIndex = getTables().get(tableIndex).indexOf(smo.getColumnName());
		if(columnIndex<0) 
			throw new ColumnNotInTableException("The source column "+ smo.getColumnName() +" to rename is not in the table\n");

		getTables().get(tableIndex).getColumns().set(columnIndex, smo.getColumnName2());
		//tables.get(tableIndex).columns.remove(columnIndex);
		//tables.get(tableIndex).columns.add(columnIndex, smo.getColumnName2());
		
		this.versionID = smo.getTargetMajor();
		this.minorVersionID = smo.getTargetMinor();
		return true;
	}

	/**
	 * @param smo
	 * @return
	 * @throws TableNotInSchemaException 
	 * @throws ColumnNotInTableException 
	 * @throws ColumnIsConstrainedException 
	 */
	public boolean applyDropColumnSMO(SMOImpl smo) throws TableNotInSchemaException, ColumnNotInTableException, ColumnIsConstrainedException {
		int index = this.indexOf(smo.getTableName());
		if(index < 0) {
			//System.err.println("Dropping column from non-existent table\n" +
			//		this.toString() + "\n"+ smo.getSourceMajor() + "." + smo.getSourceMinor() + " "+  smo);
			throw new TableNotInSchemaException("Source table" +smo.getTableName() +" is not in the schema\n"); 
		}
		else if(getTable(index).indexOf(smo.getColumnName()) < 0) {
			//System.err.println("Column to drop is not in the table\n" +
			//	this.toString() + "\n" + smo.getSourceMajor() + "." + smo.getSourceMinor() + " "+  smo);
			throw new ColumnNotInTableException("The column "+ smo.getColumnName() +" to drop is not in the table\n");
		}
		
		int tableIndex = this.indexOf(smo.getTableName());
		if(tableIndex<0) 
			throw new TableNotInSchemaException("Source table" +smo.getTableName() +" to be merged is not in the schema\n"); 		int columnIndex = getTables().get(tableIndex).indexOf(smo.getColumnName());
		if(columnIndex<0) 
			throw new ColumnNotInTableException("The column "+ smo.getColumnName() +" to drop is not in the table\n");
		
		
		getTables().get(tableIndex).getColumns().remove(columnIndex);
		getTables().get(tableIndex).getColTypes().remove(columnIndex);
		
		this.versionID = smo.getTargetMajor();
		this.minorVersionID = smo.getTargetMinor();
		return true;
	}

	/**
	 * @param smo
	 * @return
	 * @throws TableNotInSchemaException 
	 * @throws ColumnExistsException 
	 * @throws IllegalAddColumnArgumentException 
	 */
	public boolean applyAddColumnSMO(SMOImpl smo) throws TableNotInSchemaException, ColumnExistsException, IllegalAddColumnArgumentException {
		int index = this.indexOf(smo.getTableName());
		if(index < 0) {
			//System.err.println("Adding column to non-existent table\n" +
			//		this.toString() + "\n" + smo);
			throw new TableNotInSchemaException("Source table" +smo.getTableName() +" is not in the schema\n"); 

		}
		else if(getTable(index).indexOf(smo.getColumnName()) >= 0) {
			//System.err.println("The column to add is already in the table\n" +
			//		this.toString() + "\n" + smo);
			throw new ColumnExistsException("The column " +smo.getColumnName() +" already exists in table " +smo.getTableName() +"\n"); 

		}
		if(smo.getFunctionArgs() != null) {
			for(int i=0; i<smo.getFunctionArgs().size(); i++) {
				String functionArg = smo.getFunctionArgs().get(i);
				if(getTable(index).indexOf(functionArg) < 0) {
					System.err.println("Add column's function argument is not in the table\n" +
							this.toString() + "\n" + smo);
			throw new IllegalAddColumnArgumentException("Add column's function argument " + functionArg + " is not in the table"+smo.getTableName()+"\n");
				}
			}
		}
		//remove redundant check
		//if(index<0) return false; // error - adding column to non-existent table				
		int colIndex = getTables().get(index).indexOf(smo.getColumnName());
		if(colIndex>=0) 
			throw new ColumnExistsException("The column " +smo.getColumnName() +" already exists in table " +smo.getTableName() +"\n"); 
		getTables().get(index).addColumn(smo.getColumnName(),smo.getColumnType());
		this.versionID = smo.getTargetMajor();
		this.minorVersionID = smo.getTargetMinor();
		return true;
	}

	/**
	 * @param smo
	 * @return
	 * @throws MergingDifferentTablesException 
	 * @throws InputTableReplicatedException 
	 * @throws TableNotInSchemaException 
	 * @throws TableExistsException 
	 * @throws notInfPresException 
	 * @throws InboundForeignKeyOnMergeException 
	 * @throws OutboundForeignKeyOnMergeException 
	 * @throws IntegrityConstraintsExistsException 
	 */
	public boolean applyMergeTableSMO(SMOImpl smo) throws MergingDifferentTablesException, InputTableReplicatedException, TableNotInSchemaException, TableExistsException, notInfPresException, InboundForeignKeyOnMergeException, OutboundForeignKeyOnMergeException, IntegrityConstraintsExistsException {
		//boolean sameName2 = false; // to manage merge into an existing table
		if(indexOf(smo.getTableName()) < 0) {
			System.err.println("Input table is not in the schema\n" +
					this.toStringWithMetadata() + "\n" + smo.toMetadataString());
			throw new TableNotInSchemaException("Source table" +smo.getTableName() +" to be merged is not in the schema\n"); 
		}
		else if(indexOf(smo.getTableName2()) < 0) {
			System.err.println("Input table is not in the schema\n" +
					this.toStringWithMetadata() + "\n" + smo.toMetadataString());
			throw new TableNotInSchemaException("Source table" +smo.getTableName2() +" to be merged is not in the schema\n"); 
		}
		else if(smo.getTableName2().equals(smo.getTableName2()) &&
					smo.getTableName2().equals(smo.getTableName())) {
			System.err.println("Two input tables have same name! \n" +
					this.toString() + "\n" + smo);
			throw new InputTableReplicatedException("The two source table have the same name\n");
		}
		
		if(indexOf(smo.getTableName3()) >= 0 &&
				!smo.getTableName3().equals(smo.getTableName()) &&
				!smo.getTableName3().equals(smo.getTableName2())) {
			System.err.println("New tablename is already used in the schema (reusing one of two input table name is allowed)\n" +
					this.toStringWithMetadata() + "\n" + smo.toMetadataString());
			throw new TableExistsException("Target partition table" +smo.getTableName3() +" is already in the schema\n"); 		

		}
		
		int ind = this.indexOf(smo.getTableName());
		int ind2 = this.indexOf(smo.getTableName2());
		//redundant check
		//if(ind<0 || ind2<0) return false; // error - merging non-existent table(s)
		
		// check to see if two tables have same columns
		Table tab1 = getTables().get(ind);
		Table tab2 = getTables().get(ind2);
		if(tab1.getNumColumns() != tab2.getNumColumns()) {
			System.err.println("Merging two tables with different signatures 1\n" +
					this.toString() + "\n" + smo);
			throw new MergingDifferentTablesException("The table being merged have different signatures\n");
		}
		
		for(int i=0; i<tab1.getColumns().size(); i++) {
			if(!tab2.getColumns().contains(tab1.getColumns().get(i))) {
				System.err.println("Merging two tables with different signatures 2\n" +
						this.toString() + "\n" + smo);
				throw new MergingDifferentTablesException("The table being merged have different signatures\n");
			}
		}
		for(int i=0; i<tab2.getColumns().size(); i++) {
			if(!tab1.getColumns().contains(tab2.getColumns().get(i))) {
				System.err.println("Merging two tables with different signatures 3\n" +
						this.toString() + "\n" + smo);
				throw new MergingDifferentTablesException("The table being merged have different signatures\n");
			}
		}
		

		//if(!sameName){
			getTables().get(ind).setTableName(smo.getTableName3());
			getTables().remove(ind2);
		/*}
		else{
			if(smo.getTableName2().equals(smo.getTableName3())){
				tables.remove(this.indexOf(smo.getTableName()));
			}
			else{
				tables.remove(this.indexOf(smo.getTableName2()));
			}
		}*/
			this.versionID = smo.getTargetMajor();
			this.minorVersionID = smo.getTargetMinor();
			return true;
	}

	/**
	 * @param smo
	 * @return
	 * @throws TableExistsException 
	 * @throws TableNotInSchemaException 
	 * @throws ColumnsAndTypesMismatch 
	 * @throws InboundForeignKeyOnPartitionException 
	 * @throws NotImplementedException 
	 * @throws IntegrityConstraintsExistsException 
	 * @throws notInfPresExceptionI
	 * @throws InboundForeignKeyOnMergeException nboundForeignKeyOnMergeException 
	 */
	public boolean applyPartitionTableSMO(SMOImpl smo) throws TableExistsException, TableNotInSchemaException, ColumnsAndTypesMismatch, notInfPresException,InboundForeignKeyOnPartitionException, NotImplementedException, IntegrityConstraintsExistsException {
		if(indexOf(smo.getTableName()) < 0) {
			System.err.println("Partitioned table is not in the schema\n" +
					this.toString() + "\n" + smo);
			throw new TableNotInSchemaException("Source table" +smo.getTableName() +" to be partitioned is not in the schema\n"); 
		}
		else if(indexOf(smo.getTableName2()) >= 0) {
			// if new table is the same as source table name, then ok
			// since the source table will be deleted
			if(!smo.getTableName().equals(smo.getTableName2())) {
				System.err.println("Table with new name is already in the schema\n" +
						this.toString() + "\n" + smo);
				throw new TableExistsException("Target partition table" +smo.getTableName2() +" is already in the schema\n"); 		
			}
		}
		else if(indexOf(smo.getTableName3()) >= 0) {
			// if new table is the same as source table name, then ok
			// since the source table will be deleted
			if(!smo.getTableName().equals(smo.getTableName3())) {
				System.err.println("Table with new name is already in the schema\n" +
						this.toString() + "\n" + smo);
				throw new TableExistsException("Target partition table" +smo.getTableName3() +" is already in the schema\n"); 		
			}
		}
		int index = this.indexOf(smo.getTableName());
		//comment redundant check
		//if(index<0) return false; // error - partitioning non-existent table
		Table table2 = getTables().get(index);
		Table table3 = new Table(table2);
		table2.setTableName(smo.getTableName2());
		table3.setTableName(smo.getTableName3());
		getTables().add(index+1, table3);
		this.versionID = smo.getTargetMajor();
		this.minorVersionID = smo.getTargetMinor();
		return true;
	}

	/**
	 * @param smo
	 * @return
	 * @throws TableExistsException 
	 * @throws TableNotInSchemaException 
	 */
	public boolean applyCopyTableSMO(SMOImpl smo) throws TableExistsException, TableNotInSchemaException {
		if(indexOf(smo.getTableName()) < 0) {
			System.err.println("Source of copy table is not in the schema\n" +
					this.toString() + "\n" + smo);
			throw new TableNotInSchemaException("Source table" +smo.getTableName() +" to be copied is not in the schema\n"); 
		}
		else if(indexOf(smo.getTableName2()) >= 0) {
			System.err.println("Target of copy table is already in the schema\n" +
					this.toString() + "\n" + smo.toMetadataString());
			throw new TableExistsException("Target table" +smo.getTableName2() +" location is already in the schema\n"); 		
		}
		int index = this.indexOf(smo.getTableName());
		//comment redundant check
		//if(index<0) return false; // error - deleting non-existent table
		Table copiedTable = getTables().get(index).clone();
		copiedTable.setTableName(smo.getTableName2());
		getTables().add(copiedTable);
		this.versionID = smo.getTargetMajor();
		this.minorVersionID = smo.getTargetMinor();
		return true;
	}

	/**
	 * @param smo
	 * @return
	 * @throws TableExistsException 
	 * @throws TableNotInSchemaException 
	 */
	public boolean applyRenameTableSMO(SMOImpl smo) throws TableExistsException, TableNotInSchemaException {
		if(indexOf(smo.getTableName()) < 0) {
			System.err.println("The table to rename is not in the schema\n" +
					this.toString() + "\n" + smo);
			
			throw new TableNotInSchemaException("Renamed table" +smo.getTableName() +" is not in the schema\n"); 
		}
		else if(indexOf(smo.getTableName2()) >= 0) {
			System.err.println("Table with the new name is already in the schema\n" +
					this.toString() + "\n" + smo);
			throw new TableExistsException("Target rename table" +smo.getTableName2() +" is already in the schema\n"); 		}
		int index = this.indexOf(smo.getTableName());
		//comment redundant check
		//if(index<0) return false; // error - deleting non-existent table
		getTables().get(index).setTableName(smo.getTableName2());
		this.versionID = smo.getTargetMajor();
		this.minorVersionID = smo.getTargetMinor();
		return true;
	}

	/**
	 * @param smo
	 * @return
	 * @throws TableNotInSchemaException 
	 * @throws TableConstrainedException 
	 */
	public boolean applyDropTableSMO(SMOImpl smo) throws TableNotInSchemaException, TableConstrainedException {
		if(indexOf(smo.getTableName()) < 0) {
			System.err.println("The table to drop is not in the schema\n" +
					this.toString() + "\n" + smo);
			
			throw new TableNotInSchemaException("The table " +smo.getTableName() +" to be dropped is not in  schema " + this.versionID + ":" + this.minorVersionID +"\n"); 
		}
		int index = this.indexOf(smo.getTableName());
		
		//remove redundant check
		//if(index<0) return false; // error - deleting non-existent table
		getTables().remove(index);
		this.versionID = smo.getTargetMajor();
		this.minorVersionID = smo.getTargetMinor();
		return true;
	}

	/**
	 * @param impl
	 * @return
	 * @throws TableExistsException 
	 * @throws ColumnExistsException 
	 */
	public boolean applyCreateTableSMO(SMOImpl impl) throws TableExistsException, ColumnExistsException, TableHasNoColumns {
		if(indexOf(impl.getTableName()) >= 0) {
			System.err.println("The table to create is already in the schema\n" +
					this.toString() + "\n" + impl);
			
			throw new TableExistsException("The table " + impl.getTableName() + " is already in the schema\n");
			
		}
		//Added by Swami - check for table having no columns and throw exception
		if(impl.getColumnNameList() == null || impl.getColumnNameList().size() <= 0)
		{
			throw new TableHasNoColumns("There are no Columns defined for Table:" +impl.getTableName() +"\n");
		}
		
		//Added by Swami - check if the columns added to the table are unique
		//need to check if the number of columns is greater than 1 - else its fine
		if(impl.getColumnNameList() != null && impl.getColumnNameList().size() > 1)
		{
			int i = 0, j = 0, listsize = 0;
			listsize = impl.getColumnNameList().size();
			for(i=0; i<listsize; i++)
			{
				if(i+1 < listsize)
				{
					for(j=i+1; j<listsize; j++)
					{
						if(impl.getColumnNameList().get(i)==impl.getColumnNameList().get(j))
						{
							//same column name
							//throw same column name exception
							//throw new ColumnExistsException("The column " +smo.getColumnName() +" already exists in table " +smo.getTableName() +"\n");
							throw new ColumnExistsException("Same Column name in the table not allowed in " +impl.getTableName() +"\n");
						
						}
					}
				}
			}
		}
		
		Table table = new Table(impl.getTableName());
		addTable(table);
		if(impl.getColumnNameList() != null && impl.getColumnNameList().size() > 0) {
			for(int i=0; i<impl.getColumnNameList().size(); i++){
				table.addColumn(impl.getColumnNameList().get(i),impl.getColumnTypeList().get(i));
			}	
		}
		this.versionID = impl.getTargetMajor();
		this.minorVersionID = impl.getTargetMinor();
		return true;
	}

	/**
	 * implements the schema modification given a JOIN smo
	 * @param smo
	 * @return
	 * @throws TableNotInSchemaException 
	 * @throws TableExistsException 
	 * @throws SameColumnNameException 
	 * @throws IllegalSMOException 
	 * @throws NoTableInSchemaException 
	 * @throws noCommonColumnInfPres 
	 * @throws noColumnsMatchInfPres 
	 */
	public boolean applyJoinTableSMO(SMOImpl smo) throws TableNotInSchemaException, TableExistsException, SameColumnNameException, noColumnsMatchInfPres, noCommonColumnInfPres, NoTableInSchemaException, IllegalSMOException {
		

		
		if(indexOf(smo.getTableName()) < 0) {
			//System.err.println("Joined table is not in the schema\n" +
			//		this.toString() + "\n" + smo);
			throw new TableNotInSchemaException("Joined table" +smo.getTableName() +" is not in the schema\n");
			
		}else if(indexOf(smo.getTableName2()) < 0) {
			//System.err.println("Joined table is not in the schema\n" +
			//		this.toString() + "\n" + smo);
			throw new TableNotInSchemaException("Joined table" +smo.getTableName2() +" is not in the schema\n");
		}
		else if(indexOf(smo.getTableName3()) >= 0) {
			// if new table is the same as source table name, then ok
			// since the source table will be deleted
			if(!smo.getTableName3().equals(smo.getTableName()) &&
					!smo.getTableName3().equals(smo.getTableName2())) {
				//System.err.println("Table with new name is already in the schema\n" +
				//		this.toString() + "\n" + smo);
				throw new TableExistsException("The table " + smo.getTableName3() + " is already in the schema\n");
			}
		}
		
		int index = this.indexOf(smo.getTableName());
		int index2 = this.indexOf(smo.getTableName2());
		
		// COMMENTING REDUNDANT CHECK
		//if(index<0 || index2<0) return false; // error - merging non-existent table(s)
		
		// create a union of columns from two source tables (table2, table3)
		Table table = this.getTable(index);
		Table table2 = this.getTable(index2);
		Table newTable = new Table(smo.getTableName3());
		newTable.setColumns((Vector<String>)table.getColumns().clone());
		newTable.setColTypes((Vector<String>)table.getColTypes().clone());
		
		for(int i=0; i<table2.getColumns().size(); i++) {
			if(!newTable.getColumns().contains(table2.getColumns().get(i))){
				newTable.getColumns().add(table2.getColumns().get(i));
				newTable.getColTypes().add(table2.getColTypes().get(i));	
			}			
			// check whether the column is part of the join condition if not complains, otherwise add only one of the two columns
			else if(!((JoinTableSMO)smo).joinOnColumn(table2.getColumns().get(i))){			
				//System.err.println("Joined tables have two columns with a same name\n" + this.toString() + "\n" + smo);
				//System.err.println(table2.getColumns().get(i));
				
				throw new SameColumnNameException("Joined tables have two columns with a same name\n");
			}
				
		}
		// drop source tables, create target table
		
		index = this.indexOf(smo.getTableName());
		index2 = this.indexOf(smo.getTableName2());
		
		getTables().set(index, newTable);			
		getTables().remove(index2);
		
		this.versionID = smo.getTargetMajor();
		this.minorVersionID = smo.getTargetMinor();
		return true;
	}
	
	
	
	
	public Schema applySMOSequence(SMOSequenceInterface smos) throws IllegalSMOException {
		Schema newSchema = new Schema(this);
		newSchema.versionID = -1; // nullify the new version's ID - we don't know it
		for(int i=0; i<smos.size(); i++)
			newSchema.applySMO((SMOImpl)smos.get(i));
		return newSchema;
	}
	
	public Schema applySMOSequenceFirstN(SMOSequenceInterface smos, int n) throws IllegalSMOException {
		Schema newSchema = new Schema(this);
		newSchema.versionID = -1; // nullify the new version's ID - we don't know it
		
		if(n<0) return null;
		if(n>smos.size()) n=smos.size();
		
		for(int i=0; i<n; i++)
			newSchema.applySMO((SMOImpl)smos.get(i));
		return newSchema;
	}


	// Given schema S' and smo, find and return schema S, where
	// S' = S.applySMO(smo)
	// return true if S != S'
	// 10-19-2007
	
	
	//FIXME check Exception managment here
	public boolean revertSMOForQueriedSchemaFootprint(SMOImpl smo, Schema fullSchemaBeforeSMO) {
		String operatorType = smo.getOperatorType();
		int size = this.size();

		if(operatorType.contentEquals("CREATE TABLE")) {
			int index = this.indexOf(smo.getTableName());
			if(index<0) return false; // no change - created table is not in footprint
			getTables().remove(index);
			return true;
		}
		else if(operatorType.contentEquals("DROP TABLE")) {
			return false; // special treatment for queried-schema-footprint
			// we should add smo.tableName and its columns for general schema-smo-reverting
		}
		else if(operatorType.contentEquals("RENAME TABLE")) {
			int index = this.indexOf(smo.getTableName2());
			if(index<0) return false; // renamed table is not in footprint
			getTables().get(index).setTableName(smo.getTableName());
			return true;
		}
		else if(operatorType.contentEquals("COPY TABLE")) {
			int index = this.indexOf(smo.getTableName2());
			if(index<0) return false; // copied table is not in footprint
			getTables().get(index).setTableName(smo.getTableName());
			return true;
		}
		else if(operatorType.contentEquals("PARTITION TABLE")) {
			int index2 = this.indexOf(smo.getTableName2());
			int index3 = this.indexOf(smo.getTableName3());
			if(index2<0 && index3<0) 
				return false; // partitioned table is not in footprint
			else if(index2 >= 0 && index3 >= 0) {
				getTables().get(index2).setTableName(smo.getTableName());
				getTables().remove(index3);
			}
			else if(index2 >= 0 && index3 < 0) {
				getTables().get(index2).setTableName(smo.getTableName());
			}
			else { // if(index2 < 0 && index3 >=0)
				getTables().get(index3).setTableName(smo.getTableName());
			}
			return true;
		}
		else if(operatorType.contentEquals("MERGE TABLE")) {
			int index = this.indexOf(smo.getTableName());
			if(index<0) return false; // merged table is not in footprint
			Table table2 = getTables().get(index);
			Table table3 = new Table(table2);
			table2.setTableName(smo.getTableName2());
			table3.setTableName(smo.getTableName3());
			getTables().add(index+1, table3);
			return true;
		}
		else if(operatorType.contentEquals("ADD COLUMN")) {
			int tableIndex = this.indexOf(smo.getTableName());
			if(tableIndex<0) return false; // added column's table is not in footprint	

			int columnIndex = getTables().get(tableIndex).indexOf(smo.getColumnName());
			if(columnIndex<0) return false; // added column is not in footprint
			getTables().get(tableIndex).getColumns().remove(columnIndex);
			return true;
		}
		else if(operatorType.contentEquals("DROP COLUMN")) {
			int index = this.indexOf(smo.getTableName());
			if(index<0) return false; // dropped column's table is not in footprint				
			getTables().get(index).addColumn(smo.getColumnName());
			return true;
		}
		else if(operatorType.contentEquals("RENAME COLUMN")) {
			int tableIndex = this.indexOf(smo.getTableName());
			if(tableIndex<0) return false; // renamed column's table is not in footprint
			int columnIndex = getTables().get(tableIndex).indexOf(smo.getColumnName2());
			if(columnIndex<0) return false; // renamed column is not in footprint
			getTables().get(tableIndex).getColumns().remove(columnIndex);
			getTables().get(tableIndex).getColumns().add(columnIndex, smo.getColumnName());
			return true;
		}
		else if(operatorType.contentEquals("COPY COLUMN")) {
			int tableIndex2 = this.indexOf(smo.getTableName2());
			if(tableIndex2<0) return false; // copy-column's destination table is not in footprint
			int columnIndex2 = getTables().get(tableIndex2).indexOf(smo.getColumnName());
			if(columnIndex2<0) return false; // copied column is not in footprint
			getTables().get(tableIndex2).getColumns().remove(columnIndex2);
			
			// if the source table is not in the footprint, expand it by adding the source table
			int tableIndex = this.indexOf(smo.getTableName());
			if(tableIndex<0 && fullSchemaBeforeSMO != null) {// copy-column's source table is not in footprint
				getTables().add(fullSchemaBeforeSMO.getTable(smo.getTableName()).clone()); 
			}
			else if(tableIndex<0) { // when fullSchemaBeforeSMO is not available
				getTables().add(new Table(smo.getTableName())); 
			}
			return true;
		}
		else if(operatorType.contentEquals("COPYBACK COLUMN")) {
			//same as drop column when considering only schema, and not considering data
			//int tableIndex = this.indexOf(smo.tableName);
			//if(tableIndex<0) return false; // error - deleting column to non-existent table
			//int columnIndex = tables.get(tableIndex).indexOf(smo.columnName);
			//if(columnIndex<0) return false; // error - deleting non-existent column
			//tables.get(tableIndex).columns.remove(columnIndex);
			//return true;
			return false;
		}
		else if(operatorType.contentEquals("MOVE COLUMN")) {
			int tableIndex = this.indexOf(smo.getTableName());
			int tableIndex2 = this.indexOf(smo.getTableName2());
			if(tableIndex<0 && tableIndex2<0) return false; // move-column's source or destination table is not in footprint
			
			else if(tableIndex<0 && tableIndex2 >=0) {
				int columnIndex = getTables().get(tableIndex2).indexOf(smo.getColumnName());
				if(columnIndex<0) return false; // moved column is not in the footprint
				getTables().get(tableIndex2).getColumns().remove(smo.getColumnName());
				if(fullSchemaBeforeSMO != null)	
					getTables().add(fullSchemaBeforeSMO.getTable(smo.getTableName()).clone()); 
				else
					getTables().add(new Table(smo.getTableName()));
			}
			else if(tableIndex>=0 && tableIndex2 <0)  {
				getTables().get(tableIndex).getColumns().add(smo.getColumnName());
				if(fullSchemaBeforeSMO != null)	
					getTables().add(fullSchemaBeforeSMO.getTable(smo.getTableName2()).clone()); 
				else
					getTables().add(new Table(smo.getTableName2()));
			}
			else {
				if(fullSchemaBeforeSMO != null)	
					getTables().add(fullSchemaBeforeSMO.getTable(smo.getTableName()).clone()); 
				else
					getTables().add(new Table(smo.getTableName()));
				if(fullSchemaBeforeSMO != null)	
					getTables().add(fullSchemaBeforeSMO.getTable(smo.getTableName2()).clone()); 
				else
					getTables().add(new Table(smo.getTableName2()));
			}
			//int columnIndex = tables.get(tableIndex).indexOf(smo.columnName);
			//if(columnIndex>=0) return false; // error - source table still has the column after copy-column
			//tables.get(tableIndex2).columns.remove(smo.columnName);
			return true;
		}	
		
		return false;
	}
	

	/*
	public boolean applySMOForQueriedSchemaFootprint(SMO smo) {
		String operatorType = smo.operatorType;
		int size = this.size();

		if(operatorType.contentEquals("CREATE TABLE")) {
			return false;
		}
		else if(operatorType.contentEquals("DROP TABLE")) {
			int index = this.indexOf(smo.tableName);
			if(index<0) return false; // error - deleting non-existent table
			tables.remove(index);
			return true;
		}
		else if(operatorType.contentEquals("RENAME TABLE")) {
			int index = this.indexOf(smo.tableName);
			if(index<0) return false; // error - deleting non-existent table
			tables.get(index).tableName = smo.tableName2;
			return true;
		}
		else if(operatorType.contentEquals("COPY TABLE")) {
			int index = this.indexOf(smo.tableName);
			if(index<0) return false; // error - deleting non-existent table
			Table copiedTable = tables.get(index).clone();
			copiedTable.tableName =smo.tableName2;
			tables.add(copiedTable);
			return true;
		}
		else if(operatorType.contentEquals("PARTITION TABLE")) {
			int index = this.indexOf(smo.tableName);
			if(index<0) return false; // error - partitioning non-existent table
			Table table2 = tables.get(index);
			Table table3 = new Table(table2);
			table2.tableName = smo.tableName2;
			table3.tableName = smo.tableName3;
			tables.add(index+1, table3);
			return true;
		}
		else if(operatorType.contentEquals("MERGE TABLE")) {
			int index2 = this.indexOf(smo.tableName2);
			int index3 = this.indexOf(smo.tableName3);
			if(index2<0 || index3<0) return false; // error - merging non-existent table(s)
			tables.get(index2).tableName = smo.tableName;
			tables.remove(index3);
			return true;
		}
		else if(operatorType.contentEquals("ADD COLUMN")) {
			int index = this.indexOf(smo.tableName);
			if(index<0) return false; // error - adding column to non-existent table				
			tables.get(index).addColumn(smo.columnName);
			return true;
		}
		else if(operatorType.contentEquals("DROP COLUMN")) {
			int index = this.indexOf(smo.tableName);
			if(index<0) return false; // error - dropping column from non-existent table				
			tables.get(index).columns.remove(smo.columnName);
			return true;
		}
		else if(operatorType.contentEquals("RENAME COLUMN")) {
			int tableIndex = this.indexOf(smo.tableName);
			if(tableIndex<0) return false; // error - renaming column to non-existent table
			int columnIndex = tables.get(tableIndex).indexOf(smo.columnName);
			if(columnIndex<0) return false; // error - renaming non-existent column
			tables.get(tableIndex).columns.remove(columnIndex);
			tables.get(tableIndex).columns.add(columnIndex, smo.columnName2);
			return true;
		}
		else if(operatorType.contentEquals("COPY COLUMN")) {
			int tableIndex = this.indexOf(smo.tableName);
			if(tableIndex<0) return false; // error - copying column from non-existent table
			int columnIndex = tables.get(tableIndex).indexOf(smo.columnName);
			if(columnIndex<0) return false; // error - copying non-existent column
			int tableIndex2 = this.indexOf(smo.tableName2);
			if(tableIndex2<0) return false; // error - copying column to non-existent table
			int columnIndex2 = tables.get(tableIndex2).indexOf(smo.columnName);
			if(columnIndex2>=0) return false; // error - target table already has a column with the same name
			
			tables.get(tableIndex2).addColumn(smo.columnName);
			return true;
		}
		else if(operatorType.contentEquals("COPYBACK COLUMN")) {
			//same as drop column when considering only schema, and not considering data
			int tableIndex = this.indexOf(smo.tableName);
			if(tableIndex<0) return false; // error - deleting column to non-existent table
			int columnIndex = tables.get(tableIndex).indexOf(smo.columnName);
			if(columnIndex<0) return false; // error - deleting non-existent column
			tables.get(tableIndex).columns.remove(columnIndex);
			return true;
		}
		else if(operatorType.contentEquals("MOVE COLUMN")) {
			int tableIndex = this.indexOf(smo.tableName);
			if(tableIndex<0) return false; // error - copying column from non-existent table
			int columnIndex = tables.get(tableIndex).indexOf(smo.columnName);
			if(columnIndex<0) return false; // error - copying non-existent column
			int tableIndex2 = this.indexOf(smo.tableName2);
			if(tableIndex2<0) return false; // error - copying column to non-existent table
			int columnIndex2 = tables.get(tableIndex2).indexOf(smo.columnName);
			if(columnIndex2>=0) return false; // error - target table already has a column with the same name
			
			tables.get(tableIndex).columns.remove(columnIndex);
			tables.get(tableIndex2).addColumn(smo.columnName);
			return true;
		}	
		
		return false;
	}
	*/
	
	/*
	public MappingSet getIdentityMappingSet() {
		MappingSet mappings = new MappingSet();
		for(int i=0; i<this.size(); i++)
			this.getTable(i).getIdentityMappingSet(mappings);	
		return mappings;
	}
	*/

	/*
	public MappingSet getMappingSet(SMOSequence smos) {
		// this (schema): source schema
		// smo: source-to-target
		// return: target-to-source schema
		MappingSet mappings = this.getIdentityMappingSet();
		mappings.updateWithSMOSequence(this, smos);
		return mappings;
	}
	*/
	
	public static Schema createSampleSchema2() {
		Schema schema = new Schema(1); // versionID = 1

		Table table = new Table("emp-acct");
		table.addColumn("eid");
		table.addColumn("name");
		table.addColumn("ethnicity");
		table.addColumn("dept");
		table.addColumn("job");
		table.addColumn("carplate");
		schema.addTable(table);

		table = new Table("emp-med");
		table.addColumn("eid");
		table.addColumn("weight");
		table.addColumn("bp");
		schema.addTable(table);
		
		table = new Table("salary");
		table.addColumn("job");
		table.addColumn("amount");
		schema.addTable(table);

		return schema;
	}

	public static Schema createSampleSchema() {
		Schema schema = new Schema(1); // versionID = 1

		Table table = new Table("emp");
		table.addColumn("eid");
		table.addColumn("name");
		table.addColumn("ethnicity");
		table.addColumn("dept");
		table.addColumn("job");
		table.addColumn("carplate");
		table.addColumn("weight");
		table.addColumn("bp");
		schema.addTable(table);
		
		table = new Table("salary");
		table.addColumn("job");
		table.addColumn("amount");
		schema.addTable(table);

		return schema;
		
	}
	

	public static Schema createSmallSampleSchema() {
		Schema schema = new Schema(1); // versionID = 1
		Table table;
		
		table = new Table("emp");
		table.addColumn("eid");
		table.addColumn("job");
		schema.addTable(table);

		table = new Table("salary");
		table.addColumn("job");
		table.addColumn("amount");
		schema.addTable(table);

		return schema;
		
	}

	public void setSchemaName(String schemaName) {
		this.schemaName = schemaName;
	}

	public String getSchemaName() {
		return schemaName;
	}
	
	public boolean equals(Object object) {
		return equals((Schema)object);
	}
	
	public boolean equals(Schema schema2) {

		if(schema2 == null) 
			return false; // this object is not null
		else if(this.getTables().size() != schema2.getTables().size())
			return false;
		else {
			for(int i=0; i<this.getTables().size(); i++) {
				Table table = this.getTables().get(i);
				Table table2 = schema2.getTable(table.getTableName());
				if(table2 == null) // this.table is not found in schema2
					return false;
				else if(!table.equals(table2))
					return false;
			}
		}
		
		// passed all test
		return true;
	}

	// return {diff1, diff2} - diff1 for "this" and diff2 for "schema2" 
	public Schema[] getDiffs(Schema schema2) {
		Schema diff1 = new Schema();
		Schema diff2 = new Schema();
		
		if(schema2 == null) 
			return null; // this object is not null
		else {
			for(int i=0; i<this.getTables().size(); i++) {
				Table table = this.getTables().get(i);
				Table table2 = schema2.getTable(table.getTableName());
				if(table2 == null) {// this.table is not found in schema2
					diff1.addTable(table);
				}
				else if(!table.equals(table2)) {
					diff1.addTable(table);
					diff2.addTable(table2);
				}
			}
			
			for(int i=0; i<schema2.getTables().size(); i++) {
				Table table2 = schema2.getTables().get(i);
				Table table = this.getTable(table2.getTableName());
				if(table == null) {// this.table is not found in schema2
					diff2.addTable(table2);
				}
			}
		}
		
		// passed all test
		Schema[] diffs = {diff1, diff2};
		return diffs;
	}
	
	/**
	 * 
	 * @param schema2
	 * @return true if this contains schema2
	 */
	public boolean containsSchema(Schema schema2) {
		if(schema2 == null) 
			return true;
		
		for(int i=0; i<schema2.size(); i++) {
			Table table2 = schema2.getTable(i);
			int index = this.indexOf(table2.getTableName());
			if(index < 0)
				return false;
			Table table = this.getTables().get(index);
			if(!table.getColumns().containsAll(table2.getColumns()))
				return false;
		}
		
		return true;
	}

	public int compareTo(Object o) {
		Schema s = (Schema) o;
		if(s.versionID<this.versionID)
			return 1;
		if(s.versionID==this.versionID && s.minorVersionID < this.minorVersionID)
			return 1;
		if(s.versionID==this.versionID && s.minorVersionID == this.minorVersionID)
			return 0;
		
		return -1;
	}

	/**
	 * @param tables the tables to set
	 */
	public void setTables(Vector<Table> tables) {
		this.tables = tables;
	}

	/**
	 * @return the tables
	 */
	public Vector<Table> getTables() {
		Vector<Table> vect = new Vector<Table>();
		if(this instanceof ICSchema){
			for(ICTable ict : ((ICSchema)this).ictables)
				vect.add(ict);
			return vect;
		}else
			return tables;
	}

	
	
}
