package edu.ucla.cs.db.prism.backend.schema.integrityconstraints;

import java.sql.Timestamp;
import java.util.Iterator;
import java.util.Set;
import java.util.Vector;

import org.jgrapht.DirectedGraph;
import org.jgrapht.alg.CycleDetector;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;

import edu.ucla.cs.db.prism.backend.ded.DEDSet;
import edu.ucla.cs.db.prism.backend.exception.SchemaErrorException;
import edu.ucla.cs.db.prism.backend.icmo.AddForeignKey;
import edu.ucla.cs.db.prism.backend.icmo.AddPrimaryKey;
import edu.ucla.cs.db.prism.backend.icmo.AddValueConstraint;
import edu.ucla.cs.db.prism.backend.icmo.DropForeignKey;
import edu.ucla.cs.db.prism.backend.icmo.DropPrimaryKey;
import edu.ucla.cs.db.prism.backend.icmo.DropValueConstraint;
import edu.ucla.cs.db.prism.backend.icmo.ICMOImpl;
import edu.ucla.cs.db.prism.backend.icmo.MOImpl;
import edu.ucla.cs.db.prism.backend.schema.Schema;
import edu.ucla.cs.db.prism.backend.schema.exception.ColumnExistsException;
import edu.ucla.cs.db.prism.backend.schema.exception.ColumnNotInTableException;
import edu.ucla.cs.db.prism.backend.schema.exception.IllegalAddColumnArgumentException;
import edu.ucla.cs.db.prism.backend.schema.exception.IllegalSMOException;
import edu.ucla.cs.db.prism.backend.schema.exception.InputTableReplicatedException;
import edu.ucla.cs.db.prism.backend.schema.exception.MergingDifferentTablesException;
import edu.ucla.cs.db.prism.backend.schema.exception.SameColumnNameException;
import edu.ucla.cs.db.prism.backend.schema.exception.TableConstrainedException;
import edu.ucla.cs.db.prism.backend.schema.exception.TableExistsException;
import edu.ucla.cs.db.prism.backend.schema.exception.TableHasNoColumns;
import edu.ucla.cs.db.prism.backend.schema.exception.TableNotInSchemaException;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.exceptions.ColumnIsConstrainedException;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.exceptions.ColumnsAndTypesMismatch;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.exceptions.ExpressionParsignErrorException;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.exceptions.InboundForeignKeyOnMergeException;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.exceptions.InboundForeignKeyOnPartitionException;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.exceptions.InboundForeignKeySplitException;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.exceptions.IntegrityConstraintsExistsException;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.exceptions.OutboundForeignKeyOnMergeException;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.exceptions.OutboundForeignKeySplitException;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.exceptions.PrimaryKeySplitException;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.exceptions.SchemaMismatchException;
import edu.ucla.cs.db.prism.backend.smo.DecomposeTableSMO;
import edu.ucla.cs.db.prism.backend.smo.JoinTableSMO;
import edu.ucla.cs.db.prism.backend.smo.MergeTableSMO;
import edu.ucla.cs.db.prism.backend.smo.PartitionTableSMO;
import edu.ucla.cs.db.prism.backend.smo.SMOImpl;
import edu.ucla.cs.db.prism.backend.smo.exception.NoTableInSchemaException;
import edu.ucla.cs.db.prism.backend.smo.exception.noColumnsMatchInfPres;
import edu.ucla.cs.db.prism.backend.smo.exception.noCommonColumnInfPres;
import edu.ucla.cs.db.prism.backend.smo.exception.notInfPresException;
import edu.ucla.cs.db.prism.backend.smo.util.Predicate;
import edu.ucla.cs.db.prism.backend.smo.util.SMOSequenceInterface;
import edu.ucla.cs.db.prism.exception.NotImplementedException;

/**
 * This class represent the Schema WITH integrity constraints. Introduces checks
 * for IC violation to applySMO functions, and implements applyICMO.
 * 
 * @author Carlo A. Curino (carlo@curino.us)
 * 
 */
public class ICSchema extends Schema {

	private static final long serialVersionUID = 1L;
	public Vector<ICTable> ictables;

	public ICSchema(int versionID, int minorVersionID, String schemaName,
			String default_character_set_name, String default_collation_name,
			String sql_path, String comment, String revision_number,
			Timestamp tstart, Timestamp tend) {

		super(versionID, minorVersionID, schemaName,
				default_character_set_name, default_collation_name, sql_path,
				comment, revision_number, tstart, tend);

		ictables = new Vector<ICTable>();
	}

	
	
	
	public boolean equals(ICSchema schema2) {

		if(schema2 == null) 
			return false; // this object is not null
		else if(this.ictables.size() != schema2.ictables.size())
			return false;
		else {
			for(int i=0; i<this.ictables.size(); i++) {
				ICTable table = this.ictables.get(i);
				ICTable table2 = schema2.getTable(table.getTableName());
				if(table2 == null) // this.table is not found in schema2
					return false;
				else if(!table.equals(table2))
					return false;
			}
		}
		
		// passed all test
		return true;
	}
	
	
	/**
	 * 
	 * @param schema2
	 * @return true if this contains schema2
	 */
	public boolean containsSchema(ICSchema schema2) {
		if(schema2 == null) 
			return true;
		
		for(int i=0; i<schema2.size(); i++) {
			ICTable table2 = schema2.getTable(i);
			int index = this.indexOf(table2.getTableName());
			if(index < 0)
				return false;
			ICTable table = this.ictables.get(index);
			if(!table.getColumns().containsAll(table2.getColumns()))
				return false;
		}
		
		return true;
	}
	
	
	/**
	 * 
	 * @param schema2
	 * @return true if this contains schema2
	 * @throws SchemaMismatchException 
	 */
	public boolean containsSchemaWithIC(ICSchema schema2, boolean checktype) throws SchemaMismatchException {
		if(schema2 == null) 
			return true;
		
		for(int i=0; i<schema2.ictables.size(); i++) {
			ICTable table2 = schema2.getTable(i);
			
			//avoid fake tables generated by IC management
			if(table2.getTableName().contains("__viol"))
				break;
			
			int index = this.indexOf(table2.getTableName());
			if(index < 0){
				throw new SchemaMismatchException(" Table " + table2.getTableName() + " does not exists in shcema " + this.versionID + ":" + this.minorVersionID);
			}	
			ICTable table = this.ictables.get(index);
			
			//compare ordered columns
			table.orderColumns();
			table2.orderColumns();
			
			if(table.getColumns().size() != table.getColumns().size())
				throw new SchemaMismatchException(" Column count mismatch for table " + table.getTableName());
			
			
			for(int j =0; j<table.getColumns().size(); j++){

				if(!table.getColumn(j).equalsIgnoreCase(table2.getColumn(j)))
					throw new SchemaMismatchException("Table " + table.getTableName() + " has column name mismatch " + table.getColumn(j) + " vs " +  table2.getColumn(j));

				if(checktype)
					if(!table.getColType(j).equalsIgnoreCase(table2.getColType(j)))
						throw new SchemaMismatchException("Table " + table.getTableName() + " has a type mismatch on column \"" +table.getColumn(j)+"\" \"" + table.getColType(j) + "\" vs \"" +  table2.getColType(j)+"\"");

			}
			
			boolean found = false;
			for(IntegrityConstraint c1:table.getConstraints()){
				found = false;
				for(IntegrityConstraint c2:table.getConstraints()){
					if(c1.equals(c2))
						found = true;
				}
				if(!found)
					throw new SchemaMismatchException("Table " + table.getTableName() + " has integrity constraints mismatch");

			}		
		}
			
		return true;
	}
	
	
	
	public ICSchema() {
		ictables = new Vector<ICTable>();
		versionID = -1;
	}

	public ICSchema(int versionID) {
		ictables = new Vector<ICTable>();
		this.versionID = versionID;
	}

	public ICSchema clone() {
		return new ICSchema(this);
	}

	public ICSchema(ICSchema srcSchema) {
		this.schemaName = srcSchema.schemaName;
		this.ictables = new Vector<ICTable>();
		
		for (int i = 0; i < srcSchema.size(); i++)
			try {
				this.addTable(new ICTable(srcSchema.getTable(i)));
			} catch (ColumnsAndTypesMismatch e) {
				e.printStackTrace();
			} catch (NullTableException e) {
				e.printStackTrace();
			}
		this.versionID = srcSchema.versionID;
		this.minorVersionID = srcSchema.minorVersionID;
		this.tstart = srcSchema.tstart;
		this.tend = srcSchema.tend;
	}
	
	public String toStringWithMetadata() {
		return "Schema version: v" + versionID + "m" + minorVersionID + "\n" + 
			ictables.toString();
	}

	
	
	public ICTable getTable(int i) {
		if (ictables != null)
			if (i < ictables.size())
				return ictables.get(i);
		return null;
	}

	public ICTable getTable(String tablename) {
		for (int i = 0; i < ictables.size(); i++)
			if (ictables.get(i).getTableName().equalsIgnoreCase(tablename)) {
				return ictables.get(i);
			}
		return null;
	}
	
	public int size() {
		return ictables.size();
	}
	
	public int indexOf(String tableName) {
		if (ictables != null) {
			for (int i = 0; i < ictables.size(); i++)
				if (ictables.get(i).getTableName().contentEquals(tableName))
					return i;
		}
		return -1;
	}

	// check to see if this schema contains a matching table as table2
	// table name should match, column name should match, and column order should match
	public int indexOf(ICTable table2) {
		for(int i=0; i<ictables.size(); i++) {
			ICTable table = ictables.get(i);

			// test 1: table name match
			if(!table.getTableName().equals(table2.getTableName()))
				continue;
			
			// test 2: table cardinality match
			if(table.getColumns().size() != table2.getColumns().size())
				continue;
			
			// test 3: individual columns
			boolean allColumnsMatch = true;
			for(int j=0; j<table.getColumns().size(); j++) {
				if(!table.getColumns().get(j).equals(table2.getColumns().get(j))) {
					allColumnsMatch = false;
					break;
				}
			}
			if(allColumnsMatch)
				return i;
		}
		return -1;
	}
	
	public void addTable(ICTable table) {
		ictables.add(table);
	}
	
	public boolean containsTable(ICTable table2) {
		int index = indexOf(table2);
		if(index<0) 
			return false;
		else 
			return true;
	}
	
	
	public boolean applyMO(MOImpl mo) throws IllegalSMOException, SchemaErrorException{
		
		if(mo instanceof ICMOImpl)
			return applyICMO((ICMOImpl)mo);
		else
			return applySMO((SMOImpl)mo);
		
	}
	
	public boolean equivalentToSchema(ICSchema testschema,boolean checktype) throws SchemaMismatchException{
		
		//System.err.println("COMPARING: " +this.toStringWithMetadata() + "\n AND " + testschema.toStringWithMetadata());

		if(testschema == null){
			return false;
		}	
		if(!testschema.schemaName.equals(schemaName)){
			System.err.println("schema name differs");
			return false;
		}	
		if(!testschema.containsSchemaWithIC(this,checktype)){
			//System.err.println("SOME TABLE FROM: " +this + "\n ARE NOT FOUND IN " + testschema);
			return false;
		}
		if(!this.containsSchemaWithIC(testschema,checktype)){
			//System.err.println("SOME TABLE FROM: " +testschema + "\n ARE NOT FOUND IN " + this);
			return false;
		}	
		
		return true;
		
	}
	
	
	
	public boolean applyICMO(ICMOImpl operator) throws SchemaErrorException, IllegalSMOException {

		this.versionID = operator.getTargetMajor();
		this.minorVersionID = operator.getTargetMinor();
		
		if(operator instanceof AddPrimaryKey)
			return applyAddPrimaryKey((AddPrimaryKey) operator);
		if(operator instanceof AddForeignKey)
			return applyAddForeignKey((AddForeignKey) operator);
		if(operator instanceof AddValueConstraint)
			return applyAddValueConstraint((AddValueConstraint) operator);
		if(operator instanceof DropPrimaryKey)
			return applyDropPrimaryKey((DropPrimaryKey) operator);
		if(operator instanceof DropForeignKey)
			return applyDropForeignKey((DropForeignKey) operator);
		if(operator instanceof DropValueConstraint)
			return applyDropValueConstraint((DropValueConstraint) operator);
		
		return false;
	}
	
	
	public boolean removeTable(String string) throws TableNotInSchemaException {
		for(ICTable i:this.getICTables()){
			if(i.getTableName().equals(string)){
				//System.out.println("REMOVING TABLE: "+i + " as asked by remove " + string);
				ictables.remove(i);
				return true;
			}	
		}
		throw new TableNotInSchemaException("Table " + string + " is not in the schema");
	}
	
	
	//--------------- APPLY ICMO FUNCTIONS --------------------------
	
	private boolean applyDropValueConstraint(DropValueConstraint operator) throws ValueConstraintNotPresentException, TableNotInSchemaException {
		
		this.getTable(operator.getTableName()).removeValueConstraintById(operator.id);
		removeViolationTable(operator);
		return false;
	}



	private boolean applyDropForeignKey(DropForeignKey operator) throws ForeignKeyNotPresentException, TableNotInSchemaException {
		
		if(operator != null && operator.getTableName()!=null && operator.id != null && this.getTable(operator.getTableName())!=null){
			this.getTable(operator.getTableName()).removeForeignKeyById(operator.id);
			removeViolationTable(operator);
			return true;
		}
		
		return false;
	}

	private boolean applyDropPrimaryKey(DropPrimaryKey operator) throws DropPrimaryKeyMismatchException, TableNotInSchemaException {
		
		ICTable ict = this.getTable(operator.getTableName());
		
		if( ict!= null && 
			ict.getPrimaryKey()!= null && 
			ict.getPrimaryKey().getId() != null && 
			ict.getPrimaryKey().getId().equals(operator.id)){
			
				ict.removePrimaryKey();
				removeViolationTable(operator);
				return true;
		}
		
		throw new DropPrimaryKeyMismatchException("The primary key " +operator.id+ " you are trying to remove is not present in this schema");

	}

	public boolean applyAddPrimaryKey(AddPrimaryKey inputicmo) throws IntegrityConstraintsExistsException {
		
		ICTable ict = this.getTable(inputicmo.getTableName());
		if(ict==null)
			return false;
			
		if(ict.hasPrimaryKey())
			return false;
		
		if(inputicmo.getColumns() == null)
			return false;
		
		
		if(ict.getColumns().containsAll(inputicmo.getColumns())){
			
			// create a table to store the "violating tuples"
			addViolationTable(inputicmo, ict);
	
			//add the constraint to the original table
			ict.addConstraint(new PrimaryKey(inputicmo.id, inputicmo.getColumns()));
			
			return true;
		}
		
		return false;	
	}


	public boolean applyAddForeignKey(AddForeignKey inputicmo) throws IntegrityConstraintsExistsException, TableNotInSchemaException, ForeignKeyLoopException {
		
		// check for loops among the foreign keys.
		if(foreignKeyLoop(inputicmo))
			throw new ForeignKeyLoopException("The foreign key you are trying to add, is creating a dangerous cycle, that might lead to non-termination of the rewriting.");
		
		ICTable tabsource = this.getTable(inputicmo.getTableName());
		ICTable tabtarget = this.getTable(inputicmo.getTableName2());
		
		if(tabsource ==null)
			throw new TableNotInSchemaException("Table " + inputicmo.getTableName() + " is not in the schema");
		if(tabtarget == null)
			throw new TableNotInSchemaException("Table " + inputicmo.getTableName2() + " is not in the schema");
				
		if(inputicmo.getTargetColumns()!=null && inputicmo.getSourceColumns()!=null){
			if(tabsource.getColumns().containsAll(inputicmo.getSourceColumns()) && 
			   tabtarget.getColumns().containsAll(inputicmo.getTargetColumns())){
				
				// create a table to store the "violating tuples"
				addViolationTable(inputicmo, tabsource);
		
				//add the constraint to the original table		
				tabsource.addConstraint(new ForeignKey(inputicmo.id, inputicmo.getSourceColumns(), inputicmo.getTableName2(), inputicmo.getTargetColumns()));
				return true;
			}
		}
		
		return false;	
	}

	
	boolean foreignKeyLoop(AddForeignKey inputicmo) {
		
		//create a directed graph, and load existing FK
		DirectedGraph<ICTable, DefaultEdge> g = new DefaultDirectedGraph<ICTable, DefaultEdge>(DefaultEdge.class);
		
		
		// add all the vertexes
		for(ICTable i:ictables){
			g.addVertex(i);
		}	
		// add all the edges
		for(ICTable i:ictables){
			if(i.getForeignKeys() != null)
			for(ForeignKey fk:i.getForeignKeys())
				if(fk!=null)
					g.addEdge(i,this.getTable(fk.getTargetTable()));
		}
	 	
	 	
	 	// if icmo is null simply check current graph
	 	if(inputicmo!=null){
	 		//load into the graph the  FK being added
	 	
	 		ICTable i1 = this.getTable(inputicmo.getTableName());
	 		ICTable i2 = this.getTable(inputicmo.getTableName2());
	 		//check whether the verteces are in place, before adding an edge
	 		if(i1!=null && i2!=null){
	 			if(!g.containsVertex(i1))
	 				g.addVertex(i1);
	 			if(!g.containsVertex(i2))
	 				g.addVertex(i2);
	 			
	 			g.addEdge(i1, i2);
	 		}	
	 	}
	 	//to avoid concurrent modification while exploring the graph I use a temporary vector to store the edge we want to remove
	 	Vector<DefaultEdge> toberemoved = new Vector<DefaultEdge>();
		
 		//remove those foreign key that cross each others and are specified on the same fields
	 	for(DefaultEdge e:g.edgeSet()){
		
	 		ICTable sourcet = g.getEdgeSource(e);
	 		ICTable targett = g.getEdgeTarget(e);
	 		
	 		//if exists both direction link
	 		if(g.containsEdge(targett, sourcet)){
				
	 			//get the two set of FK
	 			Vector<ForeignKey> sourcefks = sourcet.getForeignKeys();
	 			
	 			//adding the FK we are about to add if any.
	 			if(inputicmo!=null){
	 				
	 				// if sourcefk is null we initialize the vector
	 				if(sourcefks == null)
	 					sourcefks = new Vector<ForeignKey>();
	 				
	 				// we add the FK which is in the graph already but not in the schema yet
					sourcefks.add(new ForeignKey(inputicmo.id, inputicmo.getSourceColumns(), inputicmo.getTableName2(), inputicmo.getTargetColumns()));
	 					
	 			}
	 			
	 			Vector<ForeignKey> targetfks = targett.getForeignKeys();
 			
	 			//System.out.println("checking FKS:" + sourcefks + " AND " + targetfks);
	 			
	 			if(sourcefks!= null && targetfks!=null)
		 			for(ForeignKey sfk:sourcefks){
		 				for(ForeignKey tfk:targetfks){
		 					//we check that:
		 					// 1) the FK is defined between primary keys
		 					// 2) the two FK are defined between the same set of columns
		 				/*	if(sourcet.getPrimaryKeysColumns().containsAll(sfk.getSourceColumns()) &&
		 					  sfk.getSourceColumns().containsAll(sourcet.getPrimaryKeysColumns())  &&
		 					 targett.getPrimaryKeysColumns().containsAll(sfk.getTargetColumns()) &&
		 					  sfk.getTargetColumns().containsAll(targett.getPrimaryKeysColumns())  
		 					){
		 						
		 						//remove the edge for sfk
		 						
		 					}*/
	 						//System.out.println("checking edge between:" + sourcet.getTableName() + " to " + targett.getTableName());

		 					if(sfk.getSourceColumns().containsAll(tfk.getTargetColumns()) && 
		 					   tfk.getTargetColumns().containsAll(sfk.getSourceColumns()) && 
		 					   tfk.getSourceColumns().containsAll(sfk.getTargetColumns()) && 
		 					   sfk.getTargetColumns().containsAll(tfk.getSourceColumns())){
		 						
		 					//remove the one of the two edges for this FK
		 						//System.out.println("removing edge between:" + sourcet.getTableName() + " to " + targett.getTableName());
		 						
		 						//if we are not already removing the complimentary edge, we remove this.
		 						if(!toberemoved.contains(g.getEdge(targett, sourcet)))
		 							toberemoved.add(g.getEdge(sourcet, targett));
		 						
		 					}
		 				}
		 			}	
	 		}
	 	}
	 	
	 	
	 	// actually remove the edges marked for removal.
	 	for(DefaultEdge de:toberemoved){
	 		g.removeEdge(de);
	 	}
	 	
	 	
	 	// detect the cycle and return boolean
		CycleDetector<ICTable, DefaultEdge> cd = new CycleDetector<ICTable, DefaultEdge>(g);
		
		//System.err.println(cd.findCycles());
		
		return cd.detectCycles();

	}
	
	public Set<ICTable> extractLoops() {
		
		//create a directed graph, and load existing FK
		DirectedGraph<ICTable, DefaultEdge> g = new DefaultDirectedGraph<ICTable, DefaultEdge>(DefaultEdge.class);
		
		// add all the vertexes
		for(ICTable i:ictables){
			g.addVertex(i);
		}	
		// add all the edges
		for(ICTable i:ictables){
			if(i.getForeignKeys() != null)
			for(ForeignKey fk:i.getForeignKeys())
				if(fk!=null)
					g.addEdge(i,this.getTable(fk.getTargetTable()));
		//FIXME we should improve this, by removing FK which loops on the same columns, e.g., after a DECOMPOSE	
		}
	 	
	 
	 	// detect the cycle and return boolean
		CycleDetector<ICTable, DefaultEdge> cd = new CycleDetector<ICTable, DefaultEdge>(g);
		return cd.findCycles();

	}
	
	




	public boolean applyAddValueConstraint(AddValueConstraint inputicmo) throws IntegrityConstraintsExistsException, ExpressionParsignErrorException {
		
		ICTable tabsource = this.getTable(inputicmo.getTableName());
		
		// create a table to store the "violating tuples"
		addViolationTable(inputicmo, tabsource);

		tabsource.addConstraint(new ValueConstraint(inputicmo.id, inputicmo.getCondition()));
		
		return true;	
	}

	
	/**
	 * This method add a violation table which is generated from the input table by adding the target version numbers plus the _viol postfix
	 * @param inputicmo
	 * @param ict
	 */
	private void addViolationTable(ICMOImpl inputicmo, ICTable ict) {
		
		ICTable violationtab = ict.clone();
		violationtab.setTableName(inputicmo.getTableName() +"__viol_"+inputicmo.id);
		
		violationtab.setIsViolating(true);
		//System.out.println("PRE:\n" + this);
		
		this.addTable(violationtab);
		//System.out.println("POST:\n" + this);
		
		
	}
	
	/**
	 * This method add a violation table which is generated from the input table by adding the target version numbers plus the _viol postfix
	 * @param inputicmo
	 * @param ict
	 */
	private boolean removeViolationTable(ICMOImpl inputicmo) {
		ICTable toberemoved = null;
		for(ICTable i:this.getICTables()){
			if(i.getTableName().equals(inputicmo.getTableName() +"__viol_"+inputicmo.id) && i.getIsViolating()){
				toberemoved = i;
			}	
		}
		if(toberemoved != null){
			ictables.remove(toberemoved);
			return true;
		}
		return false;
	}
	
	
	
	// ------------------------------ APPLY SMO FUNCTIONS OVERRIDDEN TO CONSIDER IC ----------------------------


	public boolean applyNopSMO(SMOImpl smo) {
		return super.applyNopSMO(smo);
	}

	public boolean applyCreateTableSMO(SMOImpl smo)
			throws TableExistsException, ColumnExistsException,
			TableHasNoColumns {
		if (indexOf(smo.getTableName()) >= 0) {
			//System.err.println("The table to create is already in the schema\n"
			//		+ this.toString() + "\n" + smo);

			throw new TableExistsException("The table " + smo.getTableName()
					+ " is already in the schema\n");

		}
		// Added by Swami - check for table having no columns and throw
		// exception
		if (smo.getColumnNameList() == null
				|| smo.getColumnNameList().size() <= 0) {
			throw new TableHasNoColumns(
					"There are no Columns defined for Table:"
							+ smo.getTableName() + "\n");
		}

		// Added by Swami - check if the columns added to the table are unique
		// need to check if the number of columns is greater than 1 - else its
		// fine
		if (smo.getColumnNameList() != null
				&& smo.getColumnNameList().size() > 1) {
			int i = 0, j = 0, listsize = 0;
			listsize = smo.getColumnNameList().size();
			for (i = 0; i < listsize; i++) {
				if (i + 1 < listsize) {
					for (j = i + 1; j < listsize; j++) {
						if (smo.getColumnNameList().get(i) == smo
								.getColumnNameList().get(j)) {
							// same column name
							// throw same column name exception
							// throw new ColumnExistsException("The column "
							// +smo.getColumnName() +" already exists in table "
							// +smo.getTableName() +"\n");
							throw new ColumnExistsException(
									"Same Column name in the table not allowed in "
											+ smo.getTableName() + "\n");

						}
					}
				}
			}
		}

		ICTable itable = new ICTable(smo.getTableName());
		addTable(itable);
		if (smo.getColumnNameList() != null
				&& smo.getColumnNameList().size() > 0) {
			for (int i = 0; i < smo.getColumnNameList().size(); i++) {
				itable.addColumn(smo.getColumnNameList().get(i), smo
						.getColumnTypeList().get(i));
			}
		}
		this.versionID = smo.getTargetMajor();
		this.minorVersionID = smo.getTargetMinor();
		return true;
	}

	public boolean applyDropColumnSMO(SMOImpl smo)
			throws ColumnIsConstrainedException, TableNotInSchemaException,
			ColumnNotInTableException {

		String tablename = smo.getTableName();
		String columnname = smo.getColumnName();

		if (columnIsPrimaryKey(tablename, columnname))
			throw new ColumnIsConstrainedException(
					"The column "
							+ tablename
							+ "."
							+ columnname
							+ "is primary key. Remove the primary key before proceeding.");
		if (columnIsSourceOfForeignKey(tablename, columnname))
			throw new ColumnIsConstrainedException(
					"The column "
							+ tablename
							+ "."
							+ columnname
							+ " is source of foreign key. Remove the foreign key before proceeding.");
		if (columnIsTargetOfForeignKey(tablename, columnname))
			throw new ColumnIsConstrainedException(
					"The column "
							+ tablename
							+ "."
							+ columnname
							+ " is target of a foreign key. Remove the foreign key before proceeding.");

		int tableIndex = this.indexOf(tablename);
		int columnIndex = -1;
		if (tableIndex < 0) {
			//System.err.println("Dropping column from non-existent table\n"
			//		+ this.toString() + "\n" + smo.getSourceMajor() + "."
			//		+ smo.getSourceMinor() + " " + smo);
			throw new TableNotInSchemaException("Source table"
					+ smo.getTableName() + " is not in the schema\n");
		} 

		columnIndex = ictables.get(tableIndex).indexOf(smo.getColumnName());
		if (columnIndex < 0) {
			//System.err.println("Column to drop is not in the table\n"
			//		+ this.toString() + "\n" + smo.getSourceMajor() + "."
			//			+ smo.getSourceMinor() + " " + smo);
			throw new ColumnNotInTableException("The column "
						+ smo.getColumnName()
						+ " to drop is not in the table\n");
			
		}

		ictables.get(tableIndex).getColumns().remove(columnIndex);
		ictables.get(tableIndex).getColTypes().remove(columnIndex);

		this.versionID = smo.getTargetMajor();
		this.minorVersionID = smo.getTargetMinor();
		return true;

	}

	public boolean applyDropTableSMO(SMOImpl smo)
			throws TableNotInSchemaException, TableConstrainedException {

		if (tableIsTargetOfForeignKey(smo.getTableName()))
			throw new TableConstrainedException("Table " + smo.getTableName()
					+ " is target of at least one Foreign Key: "
					+ getForeignKeysByTarget(smo.getTableName()).firstElement().getId().toString());

		if (indexOf(smo.getTableName()) < 0) {
			//System.err.println("The table to drop is not in the schema\n"
			//		+ this.toString() + "\n" + smo);

			throw new TableNotInSchemaException("The table "
					+ smo.getTableName() + " to be dropped is not in  schema "
					+ this.versionID + ":" + this.minorVersionID + "\n");
		}
		int index = this.indexOf(smo.getTableName());

		ictables.remove(index);
		this.versionID = smo.getTargetMajor();
		this.minorVersionID = smo.getTargetMinor();
		return true;
	}

	
	
	
	
	public boolean applyAddColumnSMO(SMOImpl smo)
			throws TableNotInSchemaException, ColumnExistsException,
			IllegalAddColumnArgumentException {

		int index = this.indexOf(smo.getTableName());
		if (index < 0) {
			//System.err.println("Adding column to non-existent table\n"
			//		+ this.toString() + "\n" + smo);
			throw new TableNotInSchemaException("Source table"
					+ smo.getTableName()
					+ " to be merged is not in the schema\n");

		} else if (getTable(index).indexOf(smo.getColumnName()) >= 0) {
			//System.err.println("The column to add is already in the table\n"
			//		+ this.toString() + "\n" + smo);
			throw new ColumnExistsException("The column " + smo.getColumnName()
					+ " already exists in table " + smo.getTableName() + "\n");

		}
		if (smo.getFunctionArgs() != null) {
			for (int i = 0; i < smo.getFunctionArgs().size(); i++) {
				String functionArg = smo.getFunctionArgs().get(i);
				if (getTable(index).indexOf(functionArg) < 0) {
					//System.err
					//		.println("Add column's function argument is not in the table\n"
					//				+ this.toString() + "\n" + smo);
					throw new IllegalAddColumnArgumentException(
							"Add column's function argument " + functionArg
									+ " is not in the table"
									+ smo.getTableName() + "\n");
				}
			}
		}

		ictables.get(index).addColumn(smo.getColumnName(), smo.getColumnType());
		this.versionID = smo.getTargetMajor();
		this.minorVersionID = smo.getTargetMinor();
		return true;

	}

	public boolean applyRenameTableSMO(SMOImpl smo) throws TableExistsException, TableNotInSchemaException {
		
		//standard checks and operations
		if(indexOf(smo.getTableName()) < 0) {
			//System.err.println("The table to rename is not in the schema\n" +
			//		this.toString() + "\n" + smo);
			throw new TableNotInSchemaException("Renamed table " +smo.getTableName() +" is not in the schema\n"); 
		}
		else if(indexOf(smo.getTableName2()) >= 0) {
			//System.err.println("Table with the new name is already in the schema\n" +
			//		this.toString() + "\n" + smo);
			throw new TableExistsException("Target rename table" +smo.getTableName2() +" is already in the schema\n"); 		}
		int index = this.indexOf(smo.getTableName());

		//RENAME VALUECONSTRAINTS TABLES
		renameTableInValueConstraints(smo.getTableName(),smo.getTableName2());


		ictables.get(index).setTableName(smo.getTableName2());

		this.versionID = smo.getTargetMajor();
		this.minorVersionID = smo.getTargetMinor();
	
		//rename table in foreign keys
		renameTableInExternalForeignKeys(smo.getTableName(),smo.getTableName2());

		return true;
	}
	
	private void renameTableInValueConstraints(String tableName,
			String tableName2) {
			Vector<ValueConstraint> vclist = getTable(tableName).getValueConstraint();
			for (ValueConstraint vc : vclist)
				if(tableName.equals(vc.getTableName()))
					vc.setTableName(tableName2);		
	}

	public boolean applyCopyTableSMO(SMOImpl smo) throws TableExistsException, TableNotInSchemaException {
		if(indexOf(smo.getTableName()) < 0) {
			//System.err.println("Source of copy table is not in the schema\n" +
			//		this.toString() + "\n" + smo);
			throw new TableNotInSchemaException("Source table" +smo.getTableName() +" to be copied is not in the schema\n"); 
		}
		else if(indexOf(smo.getTableName2()) >= 0) {
			//System.err.println("Target of copy table is already in the schema\n" +
			//		this.toString() + "\n" + smo.toMetadataString());
			throw new TableExistsException("Target table" +smo.getTableName2() +" location is already in the schema\n"); 		
		}
		int index = this.indexOf(smo.getTableName());
		ICTable copiedTable = ictables.get(index).clone();
		copiedTable.setTableName(smo.getTableName2());
		ictables.add(copiedTable);
		this.versionID = smo.getTargetMajor();
		this.minorVersionID = smo.getTargetMinor();
		return true;
	}
	
	public boolean applyRenameColumnSMO(SMOImpl smo) throws ColumnNotInTableException, ColumnExistsException, TableNotInSchemaException {
		
		
		// normal checks, and normal column remane
		int index = this.indexOf(smo.getTableName());	
		if(index < 0) {
			//System.err.println("Renaming column in non-existent table\n" +
			//		this.toString() + "\n" + smo);
			throw new TableNotInSchemaException("Source table" +smo.getTableName() +" is not in the schema\n"); 
		}
		else if(getTable(index).indexOf(smo.getColumnName()) < 0) {
			//System.err.println("Column to rename is not in the table\n" +
			//		this.toString() + "\n" + smo);
			throw new ColumnNotInTableException("The source column "+ smo.getColumnName() +" to rename is not in the table\n");
		}
		else if(getTable(index).indexOf(smo.getColumnName2()) >= 0) {
			//System.err.println("Column with new name is already in the table\n" +
			//		this.toString() + "\n" + smo);
			throw new ColumnExistsException("The target column "+ smo.getColumnName2() +" exists\n");
		}
		int tableIndex = this.indexOf(smo.getTableName());
		if(tableIndex<0) 
			throw new TableNotInSchemaException("Source table" +smo.getTableName() +" is not in the schema\n"); 
		int columnIndex = ictables.get(tableIndex).indexOf(smo.getColumnName());
		if(columnIndex<0) 
			throw new ColumnNotInTableException("The source column "+ smo.getColumnName() +" to rename is not in the table\n");

		ictables.get(tableIndex).getColumns().set(columnIndex, smo.getColumnName2());
		this.versionID = smo.getTargetMajor();
		this.minorVersionID = smo.getTargetMinor();
	
		//adjust foreign keys and primary keys	
		renameColumnInPrimaryKeys(smo.getTableName(), smo.getColumnName(), smo.getColumnName2());
		renameColumnInValueConstraints(smo.getTableName(), smo.getColumnName(), smo.getColumnName2());
		renameColumnInInternalForeignKeys(smo.getTableName(), smo.getColumnName(), smo.getColumnName2());
		renameColumnInExternalForeignKeys(smo.getTableName(), smo.getColumnName(), smo.getColumnName2());
		
		return true;
	}

	private void renameColumnInValueConstraints(String tableName,
			String columnName, String columnName2) {

		Vector<ValueConstraint> vclist = getTable(tableName).getValueConstraint();
		for (ValueConstraint vc : vclist)
			if(vc.getColumn().equals(columnName))
				vc.setColumn(columnName2);
	}
		
		
	

	/**
	 * implements the schema modification given a DECOMPOSE smo. Checks and propagates Integrity Cosntraints. Prevent DECOMPOSITION that are not information 
	 * preserving, i.e., invertible (appropraite keys and overllaps are required before performing the operation)
	 * @param smo
	 * @return
	 * @throws IllegalSMOException
	 */
	public boolean applyDecomposeTableSMO(SMOImpl smo) throws IllegalSMOException {
		int inputtableindex = indexOf(smo.getTableName());
		
		if(inputtableindex < 0) {
			//System.err.println("Decomposed table is not in the schema\n" +
			//		this.toString() + "\n" + smo);
			throw new TableNotInSchemaException("Source table" +smo.getTableName() +" is not in the schema\n"); 
		}
		else if(indexOf(smo.getTableName2()) >= 0) {
			// if new table is the same as source table name, then ok
			// since the source table will be deleted
			if(!smo.getTableName().equals(smo.getTableName2())) {
				//System.err.println("Table with new name is already in the schema\n" +
				//		this.toString() + "\n" + smo);
				throw new TableExistsException("Target table" +smo.getTableName2() +" is already in the schema\n"); 
			}
		}
		else if(indexOf(smo.getTableName3()) >= 0) {
			// if new table is the same as source table name, then ok
			// since the source table will be deleted
			if(!smo.getTableName().equals(smo.getTableName3())) {
				//System.err.println("Table with new name is already in the schema\n" +
				//		this.toString() + "\n" + smo);
				throw new TableExistsException("Target table" +smo.getTableName3() +" is already in the schema\n"); 
			}
		}
		
		// check attributes - they should exist in the source schema
		for(int i=0; smo.getAttributeList2() != null &&
						i<smo.getAttributeList2().size(); i++) {
			String attr = smo.getAttributeList2().get(i);
			if(!containsColumn(smo.getTableName(), attr)) {
				//System.err.println("Used attribute is not in the schema\n" +
				//		this.toString() + "\n" + smo);
				throw new ColumnNotInTableException("The column " + attr +" is not present in Table "+ smo.getTableName()+ " in the schema\n"); 
			}
		}
		for(int i=0; smo.getAttributeList3() != null &&
						i<smo.getAttributeList3().size(); i++) {
			String attr = smo.getAttributeList3().get(i);
			if(!containsColumn(smo.getTableName(), attr)) {
				//System.err.println("Used attribute is not in the schema\n" +
				//		this.toString() + "\n" + smo);
				throw new ColumnNotInTableException("The column " + attr +" is not present in Table "+ smo.getTableName()+ " in the schema\n"); 
			}
		}

		
		
		// check information preservation (in the new spirit we do not allow any structural SMO which is not information preserving)
		if(((DecomposeTableSMO)smo).isICInfPreserving(this) == false)
			throw new IllegalSMOException("The DECOMPOSE SMO you are trying to use is not invertible. Please change input integrity constraints and try again");
		
		// checking inbound foreign key
		
		Vector<ForeignKey> fk = getForeignKeysByTarget(smo.getTableName());
		Vector<String> atl2 = smo.getAttributeList2();
		Vector<String> atl3 = smo.getAttributeList3();
		
		if(fk!=null)
			for(ForeignKey f :fk){
				
				// if both tables contain FK target duplicate the FK
				if(atl2.containsAll(f.getTargetColumns()) && atl3.containsAll(f.getTargetColumns())){
					//create new fk to point at table3
					for(ICTable i:ictables){
						if(i.getConstraints().contains(f)){
							ForeignKey newf = f.clone();
							newf.setTargetTable(smo.getTableName3());
							newf.setId(newf.getId() + "_" + smo.getTableName3());
							i.addConstraint(newf);
						}
							
					}
					//point existing fk on table2
					f.setTargetTable(smo.getTableName2());
					
				}
				
				if(atl2.containsAll(f.getTargetColumns()))
					f.setTargetTable(smo.getTableName2());
				
				if(atl3.containsAll(f.getTargetColumns()))
					f.setTargetTable(smo.getTableName3());
				
				
				if(!atl2.containsAll(f.getTargetColumns()) && !atl3.containsAll(f.getTargetColumns())){
					throw new InboundForeignKeySplitException("The DECOMPOSE SMO is splitting an inbound foreign key for table:" + smo.getTableName());
				}
								
			}
	
		// checking outbound foreignKey
		fk = this.getTable(smo.getTableName()).getForeignKeys();
		if(fk!=null)
			for(ForeignKey f :fk){
				if(!atl2.containsAll(f.getSourceColumns()) && !atl3.containsAll(f.getSourceColumns())){
					throw new OutboundForeignKeySplitException("The DECOMPOSE SMO is splitting an internal foreign key for table: " + smo.getTableName());
				}
			}
		
		Vector<String> pk = getTable(smo.getTableName()).getPrimaryKeysColumns();
		
		if(pk!=null && !atl2.containsAll(pk) && !atl3.containsAll(pk))
			throw new PrimaryKeySplitException("The primary key is split by the DECOMPOSE SMO.");
		
		
		ICTable intab = this.getTable(smo.getTableName());		
		
		ICTable table2 = new ICTable(smo.getTableName2());
		ICTable table3 = new ICTable(smo.getTableName3());
		
		table2.setColumns((Vector<String>)smo.getAttributeList2().clone());	
		Iterator<String> col = table2.getColumns().iterator();
		while(col.hasNext()){
			table2.getColTypes().add(intab.getColType(col.next()));
		}
		
		table3.setColumns((Vector<String>)smo.getAttributeList3().clone());
		col = table3.getColumns().iterator();
		while(col.hasNext()){
			table3.getColTypes().add(intab.getColType(col.next()));
		}

		
		// create double foreignkey connecting the two schemas
		Vector<String> fkcols = new Vector<String>();
		for(String col1 : atl2){
			if(atl3.contains(col1)){
				fkcols.add(col1);
			}
		}
		
		try {
			
			table2.addConstraint(new PrimaryKey(intab.getPrimaryKey()));
			table3.addConstraint(new PrimaryKey(intab.getPrimaryKey()));
			ForeignKey fk23 = new ForeignKey("fk_" + smo.getTableName2()+"_"+smo.getTableName3(), fkcols, smo.getTableName3(), fkcols);
			table2.addConstraint(fk23);
			ForeignKey fk32 = new ForeignKey("fk_" + smo.getTableName3()+"_"+smo.getTableName2(), fkcols, smo.getTableName2(), fkcols);
			table3.addConstraint(fk32);
		} catch (IntegrityConstraintsExistsException e) {
			e.printStackTrace();
		}
		
		ictables.remove(inputtableindex);
		ictables.add(inputtableindex, table2);
		ictables.add(inputtableindex+1, table3);
		this.versionID = smo.getTargetMajor();
		this.minorVersionID = smo.getTargetMinor();
		return true;
	}
	
	/**
	 * implements the schema modification given a JOIN smo. Checks and propagates Integrity Cosntraints. Prevent Joins that are not information 
	 * preserving (Foreign Keys, and Primary Keys must be added before issuing the JOIN SMO)
	 * @param smo
	 * @return
	 * @throws IllegalSMOException 
	 * @throws NoTableInSchemaException 
	 * @throws noCommonColumnInfPres 
	 * @throws noColumnsMatchInfPres 
	 * @throws  
	 */
	public boolean applyJoinTableSMO(SMOImpl smo) throws noColumnsMatchInfPres, noCommonColumnInfPres, NoTableInSchemaException, IllegalSMOException {
		
		if(indexOf(smo.getTableName()) < 0) {
			//System.err.println("Joined table is not in the schema\n" +
			//		this.toString() + "\n" + smo);
			throw new TableNotInSchemaException("Joined table" +smo.getTableName() +" is not in the schema\n");
			
		}else if(indexOf(smo.getTableName2()) < 0) {
			//System.err.println("Joined table is not in the schema\n" +
			//		this.toString() + "\n" + smo);
			throw new TableNotInSchemaException("Joined table" +smo.getTableName2() +" is not in the schema\n");
		}
		else if(indexOf(smo.getTableName3()) >= 0) {
			// if new table is the same as source table name, then ok
			// since the source table will be deleted
			if(!smo.getTableName3().equals(smo.getTableName()) &&
					!smo.getTableName3().equals(smo.getTableName2())) {
				//System.err.println("Table with new name is already in the schema\n" +
				//		this.toString() + "\n" + smo);
				throw new TableExistsException("The table " + smo.getTableName3() + " is already in the schema\n");
			}
		}
		
		
		// check information preservation (in the new spirit we do not allow any structural SMO which is not information preserving)
		if(((JoinTableSMO) smo).isICInfPreserving(this) == false){
			throw new IllegalSMOException("The JOIN SMO you are trying to use is not invertible. Please change input integrity constraints and try again");
		}
		
		int index = this.indexOf(smo.getTableName());
		int index2 = this.indexOf(smo.getTableName2());
		
		
		// create a union of columns from two source tables (table2, table3)
		ICTable table = this.getTable(index);
		ICTable table2 = this.getTable(index2);
		ICTable newTable = new ICTable(smo.getTableName3());
		newTable.setColumns((Vector<String>)table.getColumns().clone());
		newTable.setColTypes((Vector<String>)table.getColTypes().clone());
		
	
		
		
		
		for(int i=0; i<table2.getColumns().size(); i++) {
			if(!newTable.getColumns().contains(table2.getColumns().get(i))){
				newTable.getColumns().add(table2.getColumns().get(i));
				newTable.getColTypes().add(table2.getColTypes().get(i));	
			}			
			// check whether the column is part of the join condition if not complains, otherwise add only one of the two columns
			else if(!((JoinTableSMO)smo).joinOnColumn(table2.getColumns().get(i))){			
				//System.err.println("Joined tables have two columns with a same name\n" + this.toString() + "\n" + smo);
				//System.err.println(table2.getColumns().get(i));
				
				throw new SameColumnNameException("Joined tables have two columns with a same name " + table2.getColumns().get(i) +"\n");
			}
				
		}
		
		
		//migrate outbound foreign keys 
		newTable.addForeignKeys(table.getForeignKeys());
		newTable.addForeignKeys(table2.getForeignKeys());
		
		//fix inbound foreignkeys renaming them to target the joined table
		Vector<ForeignKey> tempfklist = this.getForeignKeysByTarget(table.getTableName());
		if(tempfklist!=null)
			for(ForeignKey f : tempfklist){
				f.setTargetTable(newTable.getTableName());
			}

		tempfklist = this.getForeignKeysByTarget(table2.getTableName());
		if(tempfklist!=null)
		
			for(ForeignKey f : tempfklist){
				f.setTargetTable(newTable.getTableName());
			}
		
		//fix primary keys
		if(table.getPrimaryKey()!=null)
				newTable.addConstraint(table.getPrimaryKey());	
		
		//FIXME note, there are more things we might want to do here. E.g., declare the redundancy, or remove self-foreignkeys etc. 
		

		
		// drop source tables, create target table
		
		index = this.indexOf(smo.getTableName());
		index2 = this.indexOf(smo.getTableName2());
		
		ictables.set(index, newTable);			
		ictables.remove(index2);
		
		this.versionID = smo.getTargetMajor();
		this.minorVersionID = smo.getTargetMinor();
		return true;
	}
	
	
	
	
	public ICSchema applySMOSequence(SMOSequenceInterface smos) throws IllegalSMOException {
		ICSchema newSchema = new ICSchema(this);
		newSchema.versionID = -1; // nullify the new version's ID - we don't know it
		for(int i=0; i<smos.size(); i++)
			try {
				newSchema.applyOperator(smos.get(i));
			} catch (ExpressionParsignErrorException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (SchemaErrorException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		return newSchema;
	}
	
	public ICSchema applySMOSequenceFirstN(SMOSequenceInterface smos, int n) throws IllegalSMOException {
		ICSchema newSchema = new ICSchema(this);
		newSchema.versionID = -1; // nullify the new version's ID - we don't know it
		
		if(n<0) return null;
		if(n>smos.size()) n=smos.size();
		
		for(int i=0; i<n; i++)
			try {
				newSchema.applyOperator(smos.get(i));
			} catch (ExpressionParsignErrorException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (SchemaErrorException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		return newSchema;
	}
		
	
	
	/**
	 * @param smo
	 * @return
	 * @throws MergingDifferentTablesException 
	 * @throws InputTableReplicatedException 
	 * @throws TableNotInSchemaException 
	 * @throws TableExistsException 
	 * @throws notInfPresException 
	 * @throws InboundForeignKeyOnMergeException 
	 * @throws OutboundForeignKeyOnMergeException 
	 * @throws IntegrityConstraintsExistsException 
	 */
	public boolean applyMergeTableSMO(SMOImpl smo) throws MergingDifferentTablesException, InputTableReplicatedException, TableNotInSchemaException, TableExistsException, notInfPresException, InboundForeignKeyOnMergeException, OutboundForeignKeyOnMergeException, IntegrityConstraintsExistsException {

		//boolean sameName2 = false; // to manage merge into an existing table
		if(indexOf(smo.getTableName()) < 0) {
			//System.err.println("Input table is not in the schema\n" +
			//		this.toStringWithMetadata() + "\n" + smo.toMetadataString());
			throw new TableNotInSchemaException("Source table" +smo.getTableName() +" to be merged is not in the schema\n"); 
		}
		else if(indexOf(smo.getTableName2()) < 0) {
			//System.err.println("Input table is not in the schema\n" +
			//		this.toStringWithMetadata() + "\n" + smo.toMetadataString());
			throw new TableNotInSchemaException("Source table" +smo.getTableName2() +" to be merged is not in the schema\n"); 
		}
		else if(smo.getTableName2().equals(smo.getTableName())) {
			//System.err.println("Two input tables have same name! \n" +
			//		this.toString() + "\n" + smo);
			throw new InputTableReplicatedException("The two source table have the same name\n");
		}
		
		if(indexOf(smo.getTableName3()) >= 0 &&
				!smo.getTableName3().equals(smo.getTableName()) &&
				!smo.getTableName3().equals(smo.getTableName2())) {
			//System.err.println("New tablename is already used in the schema (reusing one of two input table name is allowed)\n" +
			//		this.toStringWithMetadata() + "\n" + smo.toMetadataString());
			throw new TableExistsException("Target partition table" +smo.getTableName3() +" is already in the schema\n"); 		

		}
		
		int ind = this.indexOf(smo.getTableName());
		int ind2 = this.indexOf(smo.getTableName2());
		
		// check to see if two tables have same columns
		ICTable tab1 = ictables.get(ind);
		ICTable tab2 = ictables.get(ind2);
		if(tab1.getNumColumns() != tab2.getNumColumns()) {
			//System.err.println("Merging two tables with different signatures 1\n" +
			//		this.toString() + "\n" + smo);
			throw new MergingDifferentTablesException("The table being merged have different signatures\n");
		}
		
		for(int i=0; i<tab1.getColumns().size(); i++) {
			if(!tab2.getColumns().contains(tab1.getColumns().get(i))) {
				//System.err.println("Merging two tables with different signatures 2\n" +
				//		this.toString() + "\n" + smo);
				throw new MergingDifferentTablesException("The table being merged have different signatures\n");
			}
		}
		for(int i=0; i<tab2.getColumns().size(); i++) {
			if(!tab1.getColumns().contains(tab2.getColumns().get(i))) {
				//System.err.println("Merging two tables with different signatures 3\n" +
				//		this.toString() + "\n" + smo);
				throw new MergingDifferentTablesException("The table being merged have different signatures\n");
			}
		}

		
		// check information preservation with IC
		if(!((MergeTableSMO)smo).isICInfPreserving(this))
			throw new notInfPresException("The MERGE operator is not information preserving. Fix input integrity constraints before proceeding.");
		
		Vector<ForeignKey> tempfk = this.getForeignKeysByTarget(smo.getTableName());
		//check inbound FK
		if(tempfk!= null)
			throw new InboundForeignKeyOnMergeException("The source table " + smo.getTableName()+ " is target of a foreign key " + tempfk +" and cannot be merged");
		
		tempfk = this.getForeignKeysByTarget(smo.getTableName2());
		if(tempfk!= null)
			throw new InboundForeignKeyOnMergeException("The source table " + smo.getTableName2()+ " is target of a foreign key " + tempfk+ " and cannot be merged");

		//check outbound fk
		Vector<ForeignKey> fk1list = this.getTable(smo.getTableName()).getForeignKeys();
		Vector<ForeignKey> fk2list = this.getTable(smo.getTableName2()).getForeignKeys();

		if(fk1list!= null && fk2list== null)
			throw new OutboundForeignKeyOnMergeException("The source table " + smo.getTableName()+ " is source of a foreign key, and cannot be merged");
		
		if(fk1list== null && fk2list!= null){
			throw new OutboundForeignKeyOnMergeException("The source table " + smo.getTableName2()+ " is source of a foreign key, and cannot be merged");
		} else{
			if(fk2list!=null)
			for(IntegrityConstraint i:fk2list)
				ictables.get(ind).addConstraint(i);
		}
		
		if(fk1list!= null && fk2list!= null)
			if(!ForeignKey.equivalentForeignKeys(fk1list,fk2list))
				throw new OutboundForeignKeyOnMergeException("The two source table have different (incompatible) outbound foreign keys");
		
		//if the two primary keys have different columns remove them from the output (which is the first of the two tables renamed)
		if(!this.getTable(smo.getTableName()).getPrimaryKey().sameFields(this.getTable(smo.getTableName2()).getPrimaryKey()))
			ictables.get(ind).removePrimaryKey();
		
		
		
		// we check that the primary keys contains two value constraints that guarantee disjointness!
		ICTable pktab1 = this.getTable(smo.getTableName());
		ICTable pktab2 = this.getTable(smo.getTableName2());
		Vector<ValueConstraint> vc1list = pktab1.getValueConstraint();
		Vector<ValueConstraint> vc2list = pktab2.getValueConstraint();
		boolean keypartition = false;
		for(ValueConstraint vc1 : vc1list){
			for(ValueConstraint vc2 : vc2list)
				if(vc1.isComplementary(vc2) && pktab1.isInPrimaryKey(vc1.getColumn()) && pktab2.isInPrimaryKey(vc2.getColumn()) )
					keypartition = true;
		}
		
		if(!keypartition)
			ictables.get(ind).removePrimaryKey();
		
		ictables.get(ind).setTableName(smo.getTableName3());
		ictables.remove(ind2);
		this.versionID = smo.getTargetMajor();
		this.minorVersionID = smo.getTargetMinor();
		return true;
	}
	
	
	/**
	 * @param smo
	 * @return
	 * @throws TableExistsException 
	 * @throws TableNotInSchemaException 
	 * @throws ColumnsAndTypesMismatch 
	 * @throws notInfPresException 
	 * @throws NotImplementedException 
	 * @throws IntegrityConstraintsExistsException 
	 * @throws InboundForeignKeyOnMergeException 
	 */
	public boolean applyPartitionTableSMO(SMOImpl smo) throws TableExistsException, TableNotInSchemaException, ColumnsAndTypesMismatch, notInfPresException, InboundForeignKeyOnPartitionException, NotImplementedException, IntegrityConstraintsExistsException {
		if(indexOf(smo.getTableName()) < 0) {
			//System.err.println("Partitioned table is not in the schema\n" +
			//		this.toString() + "\n" + smo);
			throw new TableNotInSchemaException("Source table" +smo.getTableName() +" to be partitioned is not in the schema\n"); 
		}
		else if(indexOf(smo.getTableName2()) >= 0) {
			// if new table is the same as source table name, then ok
			// since the source table will be deleted
			if(!smo.getTableName().equals(smo.getTableName2())) {
				//System.err.println("Table with new name is already in the schema\n" +
				//		this.toString() + "\n" + smo);
				throw new TableExistsException("Target partition table" +smo.getTableName2() +" is already in the schema\n"); 		
			}
		}
		else if(indexOf(smo.getTableName3()) >= 0) {
			// if new table is the same as source table name, then ok
			// since the source table will be deleted
			if(!smo.getTableName().equals(smo.getTableName3())) {
				//System.err.println("Table with new name is already in the schema\n" +
				//		this.toString() + "\n" + smo);
				throw new TableExistsException("Target partition table" +smo.getTableName3() +" is already in the schema\n"); 		
			}
		}
		
		// check information preservation with IC
		if(!((PartitionTableSMO)smo).isICInfPreserving(this))
			throw new notInfPresException("The PARTITION operator is not information preserving. Fix input integrity constraints before proceeding.");
		
		Vector<ForeignKey> tempfk = this.getForeignKeysByTarget(smo.getTableName());
		//check inbound FK
		if(tempfk!= null)
			throw new InboundForeignKeyOnPartitionException("The source table " + smo.getTableName()+ " is target of a foreign key " + tempfk +" and cannot be partitioned");
				

		//no actions required for outbound fk which are just duplicated in both tables
		//no actions required for pks which keep to hold in both tables
		
		
		//FIXME add complementary ValueConsrtaints based on the PARTITION conditions
		
		Predicate p2 = smo.getTargetConditions2().conjunctives.get(0).predicates.get(0);
		Predicate p3 = smo.getTargetConditions3().conjunctives.get(0).predicates.get(0);
		p2.term1.tableName = smo.getTableName2();
		p3.term1.tableName = smo.getTableName3();
		
		int index = this.indexOf(smo.getTableName());
	
		ICTable table2 = ictables.get(index);
		ICTable table3 = null;
		try {
			table3 = new ICTable(table2);
		} catch (NullTableException e) {
			e.printStackTrace();
		}
		table2.setTableName(smo.getTableName2());
		table3.setTableName(smo.getTableName3());
		
		table2.addConstraint(new ValueConstraint(smo.getTableName2() + "_part_vc",p2));
		table3.addConstraint(new ValueConstraint(smo.getTableName3() + "_part_vc", p3));
		
		
		ictables.add(index+1, table3);
		this.versionID = smo.getTargetMajor();
		this.minorVersionID = smo.getTargetMinor();
		return true;
	}
	

	//---------------------------- SUPPORT METHODS -----------------------
	
	

	public boolean containsColumn(String tablename, String columnname){
		ICTable table = null;
				
		for(int i=0; i<ictables.size(); i++) {
			if(ictables.get(i).getTableName().contentEquals(tablename)) {
				table = ictables.get(i);
				break;
			}
		}
		
		if(table == null)			
			return false;
		else if(table.indexOf(columnname) > -1)
			return true;
		else 
			return false;
	}
	
	private void renameTableInExternalForeignKeys(String tableName, String tableName2) {

		Vector<ForeignKey> fk = getForeignKeysByTarget(tableName);
		if(fk!=null)
			for(int i=0; i<fk.size();i++){
				fk.elementAt(i).setTargetTable(tableName2);
			}
		
	}
	
	private void renameColumnInExternalForeignKeys(String tableName,
			String columnName, String columnName2) {
		  Vector<ForeignKey> fk = getForeignKeysByTarget(tableName);
		  if(fk!=null)
			  for(int i = 0; i<fk.size();i++){
				if(fk.elementAt(i).isTargetColumn(columnName))
					fk.elementAt(i).renameTargetColumn(columnName, columnName2);
				
			}
		
	}

	private void renameColumnInInternalForeignKeys(String tableName,
			String columnName, String columnName2) {
		Vector<ForeignKey> foreignKeys = getTable(tableName).getForeignKeys();
		if(foreignKeys!=null)
			for(int i = 0; i<foreignKeys.size();i++){
				if(foreignKeys.elementAt(i).isSourceColumn(columnName))
					foreignKeys.elementAt(i).renameSourceColumn(columnName,columnName2);
			}
		
	}

	private void renameColumnInPrimaryKeys(String tableName, String columnName,	String columnName2) {

		Vector<String> primaryKeys = getTable(tableName).getPrimaryKeysColumns();
		if(primaryKeys!= null)
		for(int i=0;i<primaryKeys.size();i++){
			if(primaryKeys.get(i).compareTo(columnName)==0)
				primaryKeys.setElementAt(columnName2, i);
		}
		
	}

	private boolean columnIsTargetOfForeignKey(String tablename,
			String columnname) {

		Vector<ForeignKey> tabs = getForeignKeysByTarget(tablename);
		if (tabs != null) {
			Iterator<ForeignKey> it = tabs.iterator();
			while (it.hasNext()) {
				ForeignKey fk = it.next();
				if (fk.getTargetColumns().contains(columnname))
					return true;
			}
		}
		return false;
	}

	private boolean columnIsPrimaryKey(String tablename, String columnname) {
		if (ictables != null) {
			Iterator<ICTable> it = ictables.iterator();
			while (it.hasNext()) {
				ICTable t = it.next();
				if (t.getTableName().equals(tablename))
					if (t.isKey(columnname))
						return true;
			}
		}
		return false;
	}

	
	private Vector<ForeignKey> getForeignKeysByTarget(String tableName) {
		Iterator<ICTable> it = ictables.iterator();
		Vector<ForeignKey> retval = null;
		while (it.hasNext()) {
			ICTable t = it.next();
			Vector<ForeignKey> fk = t.getForeignKeyTargetting(tableName);
			if (retval == null)
				retval = fk;
			else{
				if(fk!=null)
					retval.addAll(fk);
			}
		}
		return retval;
	}

	
	private boolean tableIsTargetOfForeignKey(String tableName) {

		Iterator<ICTable> it = ictables.iterator();
		while (it.hasNext()) {
			ICTable t = it.next();
			if (t.getForeignKeyTargetting(tableName) != null)
				return true;
		}
		return false;
	}

	private boolean columnIsSourceOfForeignKey(String tableName, String column) {

		Iterator<ICTable> it = ictables.iterator();
		while (it.hasNext()) {
			ICTable t = it.next();
			if (t.getTableName().equals(tableName)) {
				if (t.isSourceofForeignKey(column))
					return true;
			}
		}
		return false;
	}

	public String toString() {
		return ictables.toString();
	}

	
	public Vector<ICTable> getICTables(){
		return ictables;
	}

	public DEDSet getICasDEDs() {
		DEDSet ds = new DEDSet();
		for(ICTable i: ictables){
			ds.addAll(i.getConstraintsAsDED("v" + this.versionID + "_m" + this.minorVersionID + "_"));
		}
		return ds;
	}



}
