package edu.ucla.cs.db.prism.backend.schema.integrityconstraints;

import java.util.Iterator;
import java.util.Vector;

import edu.ucla.cs.db.prism.backend.ded.DED;
import edu.ucla.cs.db.prism.backend.ded.DEDSet;
import edu.ucla.cs.db.prism.backend.ded.QuantifiedDisjunctive;
import edu.ucla.cs.db.prism.backend.ded.Quantifier;
import edu.ucla.cs.db.prism.backend.schema.Table;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.exceptions.ColumnsAndTypesMismatch;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.exceptions.IntegrityConstraintsExistsException;
import edu.ucla.cs.db.prism.backend.smo.SMOImpl;
import edu.ucla.cs.db.prism.backend.smo.util.Conjunctive;
import edu.ucla.cs.db.prism.backend.smo.util.Predicate;
import edu.ucla.cs.db.prism.backend.smo.util.Term;

/**
 * This class represent a Table WITH integrity constraints, extend its supertype, by adding the constraints management.
 * @author Carlo A. Curino (carlo@curino.us)
 *
 */
public class ICTable extends Table {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	private Vector<IntegrityConstraint> constraints = null;
	public boolean violating = false;
	
	public ICTable(String tableName) {
		this.setTableName(tableName);
		this.setConstraints(new Vector<IntegrityConstraint>());
	}

	public ICTable(ICTable srcTable) throws ColumnsAndTypesMismatch, NullTableException {
		if(srcTable == null)
			throw new NullTableException("The input table is null");
		
		this.setTableName(srcTable.getTableName());
		this.setColumns(new Vector<String>());
		this.setColTypes(new Vector<String>());
		
		if(srcTable.getColumns().size()!= srcTable.getColTypes().size()){
			throw new ColumnsAndTypesMismatch("The number of columns and type in table:" + srcTable.getTableName() + " mismatch");
		}
		
		for(int i=0; i<srcTable.getColumns().size(); i++){ 
			this.getColumns().add(new String(srcTable.getColumns().get(i)));
			if(srcTable.getColTypes().get(i)==null){
				System.err.println("NULL TYPE IN TABLE -> SET DEFAULT:" + srcTable);
				this.getColTypes().add(new String("varchar(255)"));
			}
			else	
				this.getColTypes().add(new String(srcTable.getColTypes().get(i)));
		}
	
		constraints = new Vector<IntegrityConstraint>();
		for(IntegrityConstraint ic :srcTable.getConstraints()){
			this.constraints.add(ic.clone());
		}
	
	}
	
	@Override
	public ICTable clone() {
		try {
			return new ICTable(this);
		} catch (ColumnsAndTypesMismatch e) {
			e.printStackTrace();
		} catch (NullTableException e) {
			e.printStackTrace();
		}
		return null;
	}
	
	
	/**
	 * @param aints the constraints to set
	 */
	private void setConstraints(Vector<IntegrityConstraint> aints) {
		this.constraints = aints;
	}


	/**
	 * @return the constraints
	 */
	Vector<IntegrityConstraint> getConstraints() {
		return constraints;
	}

	/**
	 * Return all the foreign keys targetting tableName 
	 * @param tableName
	 * @return
	 */
	public Vector<ForeignKey> getForeignKeyTargetting(String tableName) {
		
		Vector<ForeignKey> fk = new Vector<ForeignKey>();
		Iterator<IntegrityConstraint> it = constraints.iterator();
		while(it.hasNext()){
			IntegrityConstraint ic = it.next();
			if(ic instanceof ForeignKey){
				if(((ForeignKey)ic).isTargetTable(tableName))
					fk.add((ForeignKey)ic);
			}
		}
		if(fk.size()==0)
			fk=null;
		return fk;
	}
	
	public void addConstraint(Vector<IntegrityConstraint> iclist) throws IntegrityConstraintsExistsException{
		for(IntegrityConstraint ic : iclist){
			addConstraint(ic);
		}
	}
	
	public void addConstraint(IntegrityConstraint ic) throws IntegrityConstraintsExistsException{
		
		for(IntegrityConstraint c : constraints){
			if(c!=null && c.getId().equals(ic.getId()))
				throw new IntegrityConstraintsExistsException("A constraint " + ic.getId() + " already exists in this table");
		}
		constraints.add(ic);
	}

	public boolean isSourceofForeignKey(String columnName) {
		
		Vector<ForeignKey> fk = new Vector<ForeignKey>();
		Iterator<IntegrityConstraint> it = constraints.iterator();
		while(it.hasNext()){
			IntegrityConstraint ic = it.next();
			if(ic instanceof ForeignKey){
				if(((ForeignKey)ic).isSourceColumn(columnName))
					return true;
			}
		}
		return false;
	}
	
	
	/**
	 * returns whether a columnname is key for the table
	 * @param column
	 * @return
	 */
	public boolean isKey(String column){
		
		Iterator<IntegrityConstraint> it = constraints.iterator();
		while(it.hasNext()){
			IntegrityConstraint ic = it.next();
			if(ic instanceof PrimaryKey){
				// CHECK THIS LINE.. is the contains working?
				if(((PrimaryKey)ic).getFields().contains(column)){
					return true;
				}
			}
		}
		return false;
		
	}

	public Vector<ForeignKey> getForeignKeys() {
		Vector<ForeignKey> fk = new Vector<ForeignKey>();
		Iterator<IntegrityConstraint> it = constraints.iterator();
		while(it.hasNext()){
			IntegrityConstraint ic = it.next();
			if(ic instanceof ForeignKey){
					fk.add((ForeignKey)ic);
			}
		}
		if(fk.size()==0)
			fk=null;
		return fk;
	}

	public boolean hasSameConstraints(ICTable table) {
		if(getConstraints() == null || table.getConstraints() == null )
			return false;
		if(getConstraints().size() != table.getConstraints().size() )
			return false;
		for(IntegrityConstraint ic : constraints) {
			if(!table.hasConstraint(ic))
				return false;
		}
		
		for(IntegrityConstraint ic : table.constraints) {
			if(!this.hasConstraint(ic))
				return false;
		}
		
		return true;
	}
	
	public boolean equals(ICTable table2) {
		if(table2 == null)
			return false;
		else if(this.getColumns().size() != table2.getColumns().size())
			return false;
		else {
			for(int i=0; i<this.getColumns().size(); i++) {
				if(!table2.hasColumn(this.getColumns().get(i)))
					return false;
			}
		}
		//FIXME we should also compare the set of IC
		
		// passed all test
		return true;
	}

	private boolean hasConstraint(IntegrityConstraint ic) {

		for(IntegrityConstraint c : constraints)
			if(c.equals(ic))
				return true;
		return false;
	}

	@Override
	public String toString(){
		String str = getTableName() + " (";
		int i;
		
		for(i=0; i<getNumColumns()-1; i++)
			str += getColumn(i) + ", ";
		if(getNumColumns() > 0)
			str += getColumn(i);

		str += ")" + constraints.toString();

		return str + "\n";
	}
	
	
	@Override
	public Vector<String> getPrimaryKeysColumns() {
		if(getPrimaryKey()!=null)
			return getPrimaryKey().getFields();
		return null;
	}

	public PrimaryKey getPrimaryKey(){	
		for(IntegrityConstraint ic : constraints)
			if(ic instanceof PrimaryKey)
				return (PrimaryKey) ic;	
		return null;
	}
	
	@Override
	public boolean isInPrimaryKey(String column){
		if(getPrimaryKey()!=null)
			return getPrimaryKey().isKey(column);
		return false;
	}

	public void addForeignKeys(Vector<ForeignKey> foreignKeys) throws IntegrityConstraintsExistsException {
		if(foreignKeys!=null)
			for(ForeignKey f : foreignKeys){
				addConstraint(f);
			}
		
	}

	public void removePrimaryKey() {
		Vector<IntegrityConstraint> v = getConstraints();
		
		for( IntegrityConstraint i:v){
			if(i instanceof PrimaryKey){
				v.remove(i);
				return;
			}
		}
	}
	
	public boolean hasPrimaryKey(){
		
		Vector<IntegrityConstraint> v = getConstraints();
		
		for( IntegrityConstraint i:v){
			if(i instanceof PrimaryKey){
				return true;
			}
		}
		
		return false;
		
	}

	public Vector<ValueConstraint> getValueConstraint() {
		
		Vector<ValueConstraint> ret = new Vector<ValueConstraint>();
		for(IntegrityConstraint ic : constraints)
			if(ic instanceof ValueConstraint)
				ret.add((ValueConstraint) ic);	

		return ret;
	}

	public void removeForeignKeysByTarget(String string) {
		
		Vector<IntegrityConstraint> c = new Vector<IntegrityConstraint>();
		for(IntegrityConstraint i:getConstraints()){
			if(i instanceof ForeignKey)
				if(((ForeignKey)i).getTargetTable().equals(string)){
					c.add(i);
				}
		}
		constraints.removeAll(c);
	}

	public String getConstraintsAsDEDString(String tableprefix) {
		return getConstraintsAsDED(tableprefix).toStringNoMetadata();
	}

	/**
	 * @param tableprefix
	 * @return
	 */
	public DEDSet getConstraintsAsDED(String tableprefix) {
	
		
		DEDSet ds = new DEDSet();
		
		//set source tables to this table
		Vector<String> sourceTables = new Vector<String>();
		sourceTables.add(getTableName());
		
		//Target tables depend on the SMO
		Vector<String> targetTables = new Vector<String>();
		
		
		for(IntegrityConstraint ic : constraints){
			
			String dedName = SMOImpl.removeSpaces(getTableName() + "_" + ic.getId());	
			DED fwdded = null;
				
			if(ic instanceof ValueConstraint){
				fwdded = new DED(null, dedName, true, sourceTables, null, null);
				fwdded.islocal = true;
				//forward direction
				Quantifier q1 = new Quantifier("all", "R0", tableprefix + getTableName());
				fwdded.globalQuantifiers.quantifiers.add(q1);
	
				QuantifiedDisjunctive tc1 = new QuantifiedDisjunctive(); 
				fwdded.thenClauses.add(tc1);
				
				Conjunctive c1 = new Conjunctive();
				tc1.disjunctive.add(c1);
				c1.predicates = new Vector<Predicate>();
				

				
				Predicate p = ((ValueConstraint)ic).getAsPredicate();
				p.term1.tableName = "R0";
				p.term1.columnName = ((ValueConstraint)ic).getColumn();
				
				c1.predicates.add(p);

				ds.add(fwdded);
				
			}
			
			if(ic instanceof ForeignKey){
				ForeignKey fk = (ForeignKey) ic;
				targetTables.add(fk.getTargetTable());
				
				fwdded = new DED(null, dedName, true, sourceTables, targetTables, null);
				fwdded.islocal = true;
				//forward direction
				Quantifier q1 = new Quantifier("all", "R0", tableprefix + getTableName());
				fwdded.globalQuantifiers.quantifiers.add(q1);
				
				Quantifier q2 = new Quantifier("some", "R1", tableprefix + fk.getTargetTable());
				fwdded.globalQuantifiers.quantifiers.add(q2);				
	
				QuantifiedDisjunctive tc1 = new QuantifiedDisjunctive(); 
				fwdded.thenClauses.add(tc1);
				
				Conjunctive c1 = new Conjunctive();
				tc1.disjunctive.add(c1);
				c1.predicates = new Vector<Predicate>();
				
				for(int i = 0; i<fk.getSourceColumns().size();i++){
					Predicate p = new Predicate();
					p.term1 = new Term();
					p.term2 = new Term();
					
					p.op = "=";
					p.term1.tableName = "R0";
					p.term2.tableName  = "R1";
					p.term1.columnName = fk.getSourceColumns().elementAt(i);
					p.term2.columnName = fk.getTargetColumns().elementAt(i);
					c1.predicates.add(p);
				}
				ds.add(fwdded);
				
			}
			
			if(ic instanceof PrimaryKey){
				
				PrimaryKey pk = (PrimaryKey) ic;
				
				fwdded = new DED(null, dedName, true, sourceTables, null, null);
				fwdded.islocal = true;
				//forward direction
				Quantifier q1 = new Quantifier("all", "R0", tableprefix + getTableName());
				fwdded.globalQuantifiers.quantifiers.add(q1);
				
				Quantifier q2 = new Quantifier("all", "R1", tableprefix + getTableName());
				fwdded.globalQuantifiers.quantifiers.add(q2);				
	
				Conjunctive c0 = new Conjunctive();
			
				fwdded.ifClause.disjunctive.conjunctives.add(c0);
				c0.predicates = new Vector<Predicate>();
				
				
				
				for(int i = 0; i<pk.getFields().size();i++){
					Predicate p = new Predicate();
					p.term1 = new Term();
					p.term2 = new Term();
					p.op = "=";
					p.term1.tableName = "R0";
					p.term2.tableName  = "R1";
					p.term1.columnName = pk.getFields().elementAt(i);
					p.term2.columnName = pk.getFields().elementAt(i);
					c0.predicates.add(p);
				}
				
				QuantifiedDisjunctive tc1 = new QuantifiedDisjunctive(); 
				fwdded.thenClauses.add(tc1);
				
				Conjunctive c1 = new Conjunctive();
				tc1.disjunctive.add(c1);
				c1.predicates = new Vector<Predicate>();
				
				// TRY to say that the two tuple-variables are the same
				Predicate p = new Predicate();
				p.term1 = new Term();
				p.term2 = new Term();
				p.op = "=";
				p.term1.tableName = "R0";
				p.term2.tableName  = "R1";
				c1.predicates.add(p);
				
				
				/*// 
				for(String col : this.getColumns()){
					Predicate p = new Predicate();
					p.term1 = new Term();
					p.term2 = new Term();
					
					p.op = "=";
					p.term1.tableName = "R0";
					p.term2.tableName  = "R1";
					p.term1.columnName = col;
					p.term2.columnName = col;
					c1.predicates.add(p);
				}
				*/
				
				
				ds.add(fwdded);
			}
			
			
		}
		return ds;
	}

	public Vector<String> getConstraintsAsStringVector() {
		Vector<String> ret = new Vector<String>();
		
		for(IntegrityConstraint con:constraints){
			ret.add(con.toString());
		}		
		return ret;
	}

	
	/**
	 * @param tableName the tableName to set
	 */
	@Override
	public void setTableName(String tableName) {
		this.tableName = tableName;
	}

	public void removeForeignKeyById(String id) throws ForeignKeyNotPresentException {

		for(IntegrityConstraint ic: constraints){
			if(ic instanceof ForeignKey)
				if(ic.getId().equals(id)){
					constraints.remove(ic);
					return;
				} 	
		}
		throw new ForeignKeyNotPresentException("Table " + this.getTableName()+" does not contain ForeignKey " + id);
	}

	public void removeValueConstraintById(String id) throws ValueConstraintNotPresentException {
		for(IntegrityConstraint ic: constraints){
			if(ic instanceof ValueConstraint)
				if(ic.getId().equals(id)){
					constraints.remove(ic);
					return;
				} 	
		}
		throw new ValueConstraintNotPresentException("Table " + this.getTableName()+" does not contain ValueConstraint " + id);
		
	}

	public void setIsViolating(boolean b) {
		violating = true;		
	}
	
	public boolean getIsViolating() {
		return violating;		
	}
}