/**
 * 
 */
package edu.ucla.cs.db.prism.backend.smo;

import java.util.Iterator;
import java.util.Vector;

import edu.ucla.cs.db.prism.backend.ded.DED;
import edu.ucla.cs.db.prism.backend.ded.DEDSet;
import edu.ucla.cs.db.prism.backend.ded.QuantifiedDisjunctive;
import edu.ucla.cs.db.prism.backend.ded.Quantifier;
import edu.ucla.cs.db.prism.backend.icmo.ICMOSMOSequence;
import edu.ucla.cs.db.prism.backend.schema.Schema;
import edu.ucla.cs.db.prism.backend.schema.Table;
import edu.ucla.cs.db.prism.backend.schema.exception.IllegalSMOException;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.ForeignKey;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.ICSchema;
import edu.ucla.cs.db.prism.backend.schema.integrityconstraints.ICTable;
import edu.ucla.cs.db.prism.backend.smo.exception.DisjunctionInJoinClauseNotImplementedException;
import edu.ucla.cs.db.prism.backend.smo.exception.NoForeignKeysDefinedException;
import edu.ucla.cs.db.prism.backend.smo.exception.NoPrimaryKeysDefinedException;
import edu.ucla.cs.db.prism.backend.smo.util.Conjunctive;
import edu.ucla.cs.db.prism.backend.smo.util.Disjunctive;
import edu.ucla.cs.db.prism.backend.smo.util.Predicate;
import edu.ucla.cs.db.prism.backend.smo.util.SMOFactory;
import edu.ucla.cs.db.prism.backend.smo.util.SMOSequence;
import edu.ucla.cs.db.prism.backend.smo.util.SMOTypeUnknownException;
import edu.ucla.cs.db.prism.backend.smo.util.Term;

/**
 * @author krl
 *
 */
public class JoinTableSMO extends SMOImpl {

	/**
	 * 
	 */
	public JoinTableSMO() {
		super(SMOInterface.JOIN);
		 setHasPerfectInverse(false);

	}

	public String toString() {
		String str = getOperatorType() + " ";	

		str += getTableName() + ", " + getTableName2() + " INTO " + 
			   getTableName3();
		if(getTargetConditions() != null)
			str += " WHERE " + getTargetConditions().toString();
		
		return str + "\n";
	}
	
	/**
	 * @deprecated
	 */
	public boolean isInfPreserving() {
		return false;
	}
	
	public boolean isICInfPreserving(ICSchema curSchema) throws IllegalSMOException {
		
		ICTable ictab1 = curSchema.getTable(this.getTableName());
		ICTable ictab2 = curSchema.getTable(this.getTableName2());
		Vector<ForeignKey> fklist12 = ictab1.getForeignKeyTargetting(ictab2.getTableName());
		Vector<ForeignKey> fklist21 = ictab2.getForeignKeyTargetting(ictab1.getTableName());
		Disjunctive dis = this.getTargetConditions();
		
		// if the JOIN degenerates into a cartesian products is information preserving (duplicates might arise)
		if(dis == null || dis.conjunctives == null)
			return true;
		
		if(dis.size()>1){
			//FIXME we should improve this, but for now if there are disjunctions in the WHERE clause we assume no inf.pres.
			throw new DisjunctionInJoinClauseNotImplementedException("The condition of the join uses a disjoint that cannot be checked for information preservation yet. ");
		}
		
		// if there are no cross foreign key, the JOIN cannot be guaranteed to be inf-pres
		if(fklist12 == null || fklist21 == null)
			throw new NoForeignKeysDefinedException("A Foreign Key needed to guarantee information preservation is missing.");
		
		// check that for each conjunct in the WHERE clause there is a portion of a foreignkey guaranteeing that the join is lossless
		for(Conjunctive c : dis.conjunctives){
			if(c.disjunctives.size()>0){
				//FIXME we don't deal with nested disjunctions either
				throw new DisjunctionInJoinClauseNotImplementedException("The condition of the join uses a disjoint that cannot be checked for information preservation yet. ");
			}
			
			for(Predicate p : c.predicates){		
				boolean covered = false;
				
				
				
				if(p.op == null || 
				   p.term1.tableName == null ||
				   p.term1.columnName == null ||
				   p.term2.tableName== null ||
				   p.term2.columnName== null){
					
					throw new IllegalSMOException("A predicate in the condition is not well defined");
					
				}
				
				
				//check fk from tab1 to tab2
				for(ForeignKey fk12: fklist12){
					for(int i=0; i<fk12.getSourceColumns().size();i++){
						String sourceCol = fk12.getSourceColumns().elementAt(i);
						String targetCol = fk12.getTargetColumns().elementAt(i);
						String targetTab = fk12.getTargetTable();
					
						if(p.op.equals("=") && 
						   p.term1.tableName.equals(ictab1.getTableName()) &&
						   p.term1.columnName.equals(sourceCol) &&
						   p.term2.tableName.equals(targetTab) &&
						   p.term2.columnName.equals(targetCol)){
							covered = true;
						}	
						
						if(p.op.equals("=") && 
						   p.term2.tableName.equals(ictab1.getTableName()) &&
						   p.term2.columnName.equals(sourceCol) &&
						   p.term1.tableName.equals(targetTab) &&
						   p.term1.columnName.equals(targetCol)){
							covered = true;
						}	
						
					
					}
				}
				if(!covered)
					throw new NoForeignKeysDefinedException("A ForeignKey needed to guarantee information preservation for the condition: " + p + " is missing.");
				
				covered=false;
				//check fk from tab2 to tab1
				for(ForeignKey fk21: fklist21){
					for(int i=0; i<fk21.getSourceColumns().size();i++){
						String sourceCol = fk21.getSourceColumns().elementAt(i);
						String targetCol = fk21.getTargetColumns().elementAt(i);
						String targetTab = fk21.getTargetTable();
					
						//check both fwd and bkwd match
						if(p.op.equals("=") && 
						   p.term1.tableName.equals(ictab2.getTableName()) &&
						   p.term1.columnName.equals(sourceCol) &&
						   p.term2.tableName.equals(targetTab) &&
						   p.term2.columnName.equals(targetCol)){
							covered = true;
						}		
						if(p.op.equals("=") && 
						   p.term2.tableName.equals(ictab2.getTableName()) &&
						   p.term2.columnName.equals(sourceCol) &&
						   p.term1.tableName.equals(targetTab) &&
						   p.term1.columnName.equals(targetCol)){
						   covered = true;
						}		
					}
				}
				if(!covered)
					throw new NoForeignKeysDefinedException("A ForeignKey needed to guarantee information preservation for the condition: " + p + " is missing.");
				
				if(!p.term1.tableName.equals(ictab1.getTableName()) || !ictab1.isInPrimaryKey(p.term1.columnName))
					throw new NoPrimaryKeysDefinedException("A PrimaryKey needed to guarantee information preservation for the condition: " + p + " is missing.");
				
				if(!p.term2.tableName.equals(ictab2.getTableName()) || !ictab2.isInPrimaryKey(p.term2.columnName))
					throw new NoPrimaryKeysDefinedException("A PrimaryKey needed to guarantee information preservation for the condition: " + p + " is missing.");
				
			}
		}
		return true;
	}
	
	public Vector<String> getModifiedInputSchemaParts() {
		Vector<String> schemaParts = new Vector<String>();
		schemaParts.add(getTableName());
		schemaParts.add(getTableName2());
		return schemaParts;	
	}	
	
	public Vector<String> getModifiedExistingTables(Schema queryFootprintTargetSchema) {
		Vector<String> modifiedExistingTables = new Vector<String>();
		modifiedExistingTables.add(getTableName());
		modifiedExistingTables.add(getTableName2());
		return modifiedExistingTables;
	}

	public Vector<SMOSequence> getCandidateInverses(Schema inputSchema) {
		Vector<SMOSequence> v = new Vector<SMOSequence>();
		SMOSequence s = new SMOSequence();
		SMOImpl newSMO = null;
		try {
			newSMO = SMOFactory.createSMO("DECOMPOSE TABLE");
		} catch (SMOTypeUnknownException e) {
			e.printStackTrace();
		}
		newSMO.setTableName(this.getTableName3());
		newSMO.setTableName2(this.getTableName());
		newSMO.setTableName3(this.getTableName2());
		
		newSMO.setAssignedInverse(new ICMOSMOSequence(this));
		
		//add the columns from the inputSchema
		Iterator<Table> tab = inputSchema.getTables().iterator();
		while(tab.hasNext()){
			Table t = tab.next();
			if(t.getTableName().equalsIgnoreCase(this.getTableName())){
				newSMO.setAttributeList2(t.getColumns());
			}
			if(t.getTableName().equalsIgnoreCase(this.getTableName2())){
				newSMO.setAttributeList3(t.getColumns());
			}
		}
		
		// switch source and target and set them in the answer
		s.sourceVersion=this.getTargetMajor();
		s.targetVersion=this.getSourceMajor();
		newSMO.setSourceMajor(this.getTargetMajor());
		newSMO.setSourceMinor(this.getTargetMinor());
		newSMO.setTargetMajor(this.getSourceMajor());
		newSMO.setTargetMinor(this.getSourceMinor());

		s.add(newSMO);
		v.add(s);
		return v;
	}
	
	public DEDSet getAllDEDs(Schema inputSchema){
		
		Vector<String> excludeList = new Vector<String>();
		excludeList.add(getTableName());
		excludeList.add(getTableName2());
		excludeList.add(getTableName3());
			
		DEDSet ds = getDEDs(inputSchema.clone());
		ds.addAll(getIdentityDEDs(inputSchema,excludeList));
		
		return ds;
	}
	
	
	@Override
	public Vector<String> getIdentitySQLViews(Schema inputSchema, String dbname) {
	
		Vector<String> excludeList = new Vector<String>();
		excludeList.add(getTableName());
		excludeList.add(getTableName2());
		excludeList.add(getTableName3());
		
		return this.getIdentitySQLView(inputSchema, dbname, excludeList);
	}
	
	
	public DEDSet getDEDs(Schema inputSchema) {
		DEDSet ds = new DEDSet(getSourceMajor(),getSourceMinor(),getTargetMajor(), getTargetMinor());
		
		String dedId = null;
		String dedName = removeSpaces(getOperatorType()) + "_" + getTableName();	
		Vector<String> sourceTables = new Vector<String>();
		sourceTables.add(getTableName3());
		
		Vector<String> targetTables = new Vector<String>();
		targetTables.add(getTableName());
		targetTables.add(getTableName2());
	
		genDed(inputSchema.clone(), ds, dedId, dedName, false, sourceTables,targetTables);
		genDed(inputSchema.clone(), ds, dedId, dedName, true,targetTables,sourceTables);
		
		return ds;
	}

	private void genDed(Schema inputSchema, DEDSet ds, String dedId,
			String dedName, boolean chaseFlag, Vector<String> sourceTables,
			Vector<String> targetTables) {
		
		dedName += "_" + getTableName2() + "_" + getTableName3();
		
		if(chaseFlag)
			dedName += "_dm"; 
		else
			dedName += "_qr"; 
		
		DED d = new DED(dedId, dedName, chaseFlag, sourceTables, targetTables, null);
				
		Quantifier q1,q2 = null;
		
		if(chaseFlag){			 
			 q1 = new Quantifier("all", "R0", "v" + getSourceMajor() + "m" +getSourceMinor() + "_" + getTableName());
			 q2 = new Quantifier("all", "R1", "v" + getSourceMajor() + "m" +getSourceMinor() + "_" + getTableName2());
			 d.globalQuantifiers.quantifiers.add(q1);
			 d.globalQuantifiers.quantifiers.add(q2);

		}
		 else{
			 q1 = new Quantifier("all", "R0", "v" + getTargetMajor() + "m" +getTargetMinor() + "_" + getTableName3());
  			 d.globalQuantifiers.quantifiers.add(q1);
		}		
		
		QuantifiedDisjunctive tc1 = new QuantifiedDisjunctive(); 
		d.thenClauses.add(tc1);

		if(chaseFlag){

			q1 = new Quantifier("some", "R2", "v" + getTargetMajor() + "m" + getTargetMinor() + "_" + getTableName3());
			tc1.quantifiers.add(q1);
		
		}
		else{	
			q1 = new Quantifier("some", "R1", "v" + getSourceMajor() + "m" + getSourceMinor() + "_" + getTableName());
			q2 = new Quantifier("some", "R2", "v" + getSourceMajor() + "m" + getSourceMinor() + "_" + getTableName2());
			tc1.quantifiers.add(q1);
			tc1.quantifiers.add(q2);			
		}
		
		Conjunctive c1 = new Conjunctive();
		tc1.disjunctive.add(c1);
		c1.predicates = new Vector<Predicate>();
		
		Table in1 = inputSchema.getTable(getTableName());
		Table in2 = inputSchema.getTable(getTableName2());
		try {
			inputSchema.applySMO(this);
		} catch (IllegalSMOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		Table outTab = inputSchema.getTable(getTableName3());
				
		Conjunctive c3 = new Conjunctive();
		
		if(this.getTargetConditions()!=null){
			if(chaseFlag){
				Iterator<Conjunctive> cond = this.getTargetConditions().conjunctives.iterator();
				while(cond.hasNext()){
					Conjunctive c = cond.next();
					Iterator<Predicate> predicates = c.predicates.iterator();
				
					while(predicates.hasNext()){
						Predicate pr = predicates.next();
						c3.predicates.add(new Predicate(new Term("R0",pr.term1.columnName),pr.op,new Term("R1",pr.term2.columnName)));
					}
					
				}
	
				d.ifClause.disjunctive.add(c3);
			}else{
				Iterator<Conjunctive> cond = this.getTargetConditions().conjunctives.iterator();
				while(cond.hasNext()){
					Conjunctive c = cond.next();
					Iterator<Predicate> predicates = c.predicates.iterator();
				
					while(predicates.hasNext()){
						Predicate pr = predicates.next();
						
						c1.predicates.add(new Predicate(new Term("R1",pr.term1.columnName),pr.op,new Term("R2",pr.term2.columnName)));
					}
					
				}
			}
		}
		
		
		
		Iterator<String> inColumns1 = in1.getColumns().iterator();
		
		while(inColumns1.hasNext()){
			String inCol = inColumns1.next();
			Iterator<String> outColumns = outTab.getColumns().iterator();
			while(outColumns.hasNext()){
				String outCol = outColumns.next();
				if(outCol.equalsIgnoreCase(inCol)){
					Predicate p1 = new Predicate();
					c1.predicates.add(p1);
					if(chaseFlag){
						p1.term1 = new Term("R0", inCol);
						p1.term2 = new Term("R2", outCol);
					}else{
						p1.term1 = new Term("R0", inCol);
						p1.term2 = new Term("R1", outCol);			
					}
					
					
					p1.op = "=";
				}
			}
		}	
	
		Iterator<String> inColumns2 = in2.getColumns().iterator();
		while(inColumns2.hasNext()){
			String inCol = inColumns2.next();
			Iterator<String> outColumns = outTab.getColumns().iterator();
			while(outColumns.hasNext()){
				String outCol = outColumns.next();
				if(outCol.equalsIgnoreCase(inCol)){
					Predicate p1 = new Predicate();
					c1.predicates.add(p1);
					if(chaseFlag){
						p1.term1 = new Term("R1", inCol);
						p1.term2 = new Term("R2", outCol);
					}else{
						p1.term1 = new Term("R0", inCol);
						p1.term2 = new Term("R2", outCol);						
					}
					
					p1.op = "=";
				}
			}
		}	
				
		ds.add(d);
	}

	private boolean inList(String type, String inCol) {
		Vector<String> v = null;
		if(type.equalsIgnoreCase("t2")){
			v= getAttributeList2();
		}
		else{
			v = getAttributeList3();
		}
		
		Iterator<String> it = v.iterator();
		while(it.hasNext())
			if(it.next().equalsIgnoreCase(inCol))
				return true;
		return false;
	}
	

	public String getSQLView(Schema schema, String inputDBname, String outputDBName){
		String sql = "";
		Vector<String> collist = new Vector();
		String colstring = "";
		
		/* 
		 * this is required to ensure correctness of multiple SMOS between 2 major 
		 * versions. Major Versions are represented by name convention in SCHEMA NAMEs, 
		 * while Minor Versions are represeted by name conventions in TABLE NAMES
		*/
		String sourceTabName1 =  this.getTableName(); 
		String targetTabName1 =  this.getTableName(); 
		String sourceTabName2 =  this.getTableName2(); 
		String targetTabName2 =  this.getTableName2(); 
		String sourceTabName3 =  this.getTableName3(); 
		String targetTabName3 =  this.getTableName3(); 
		
		if(this.getSourceMinor()>0){
			sourceTabName1 += "_" + this.getSourceMinor();
			sourceTabName2 += "_" + this.getSourceMinor();
			sourceTabName3 += "_" + this.getSourceMinor();
		}
		if(this.getTargetMinor()>0){
			targetTabName1 += "_" + this.getTargetMinor();	
			targetTabName2 += "_" + this.getTargetMinor();
			targetTabName3 += "_" + this.getTargetMinor();
		}
		
		Iterator<String> c1  = null;
		Iterator<String> c2 =null;
		Iterator<Table> tlist = schema.getTables().iterator();
		while(tlist.hasNext()){
			Table t = tlist.next();
			if(t.getTableName().equalsIgnoreCase(getTableName())){
				c1 = t.getColumns().iterator();
			}else if(t.getTableName().equalsIgnoreCase(getTableName2())){
				c2 = t.getColumns().iterator();	
			}
		}
		
		while(c1.hasNext()){
			String c1name = c1.next();
			collist.add("t1." + c1name);
			while(c2.hasNext()){
				String c2name = c2.next();						
				if(c1name.equalsIgnoreCase(c2name)){
					collist.add("t2." + c2name + " AS t2" + c1name);
				}
				else{
					collist.add("t2." + c2name);
				}
			}
		}	
		
		Iterator<String> colit = collist.iterator();
		
		while(colit.hasNext()){
			colstring +=colit.next();
			if(colit.hasNext())	
				colstring+=",";
		}
		
		String conditions = "";
		Iterator<Conjunctive> cond = this.getTargetConditions().conjunctives.iterator();
		while(cond.hasNext()){
			Conjunctive c = cond.next();
			Iterator<Predicate> predicates = c.predicates.iterator();
		
			while(predicates.hasNext()){
				Predicate pr = predicates.next();
				conditions+= "t1." + pr.term1.columnName + pr.op + "t2." + pr.term2.columnName;
				if(predicates.hasNext())
					conditions+=" AND ";
			}
		}
		
		sql += "CREATE VIEW " + outputDBName + "." + targetTabName3 + " AS ";
		sql += "\n( SELECT " + colstring + " FROM " + inputDBname + "." + sourceTabName1 + " t1, " + inputDBname + "." + sourceTabName2 +" t2";
		sql += " WHERE " + conditions;
		sql += " );\n"; 
			
		return sql;
	}
	
	public String getSQLDDL(Schema schema, String outputDBname) {
		String sqlddl = "";	
		String collist3 = "";
		
		//CARLO: we only need the target table here, thus i apply the SMO to the schema and use that
		Schema tempsch = null;
		try {
			tempsch = schema.cloneApplySMO(this);
		} catch (IllegalSMOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		Vector<String> veccol3 = tempsch.getTable(getTableName3()).getColumns();
		Iterator<String> cols3 = veccol3.iterator();
		String temp = "";
		//String colAndType = "";
		//String colAndType2 = "";
		/*
		String collist = "";
		String collist2 = "";
		
		//int i=0;	
	System.out.println("schema : "+schema.toString()+"\n");
	System.out.println("getTableName() : "+getTableName()+"\n");
	System.out.println("getTableName2() : "+getTableName2()+"\n");
	//System.out.println("getTableName2() : "+getTableName3()+"\n");
	//System.out.println("getColType() : "+schema.getTable(getTableName()).getColType(0));
	
	System.out.println("schema.getTable(getTableName()) : "+schema.getTable(getTableName())+"\n");
	System.out.println("schema.getTable(getTableName2()) : "+schema.getTable(getTableName2())+"\n");
	//System.out.println("schema.getTable(getTableName3()) : "+schema.getTable(getTableName3())+"\n");
	//System.out.println("getTableName() : "+schema.getTable(getTableName()).toString() +"\n"); // null
	//System.out.println("getTableName()2 : "+schema.getTable(getTableName2()).toString() +"\n"); // null
		
		Vector<String> veccol = schema.getTable(getTableName()).getColumns();
		Vector<String> veccol2 = schema.getTable(getTableName2()).getColumns();
		Iterator<String> cols = veccol.iterator();
		Iterator<String> cols2 = veccol2.iterator();
		//Iterator<String> cols = getAttributeList2().iterator();
		//Iterator<String> cols2 = getAttributeList3().iterator();
		
		while (cols.hasNext()) {
			temp = cols.next();
			if(veccol2.contains(temp)==true) {
				//colAndType += getTableName()+"_";
				collist += getTableName()+".";

			}
			//colAndType += temp + " " + schema.getTable(getTableName3()).getColType(temp);
			
			collist += temp;
			if (cols.hasNext()) {
				collist += ",";
				//colAndType += ",";
			}
			//i++;
		}
		//i=0;
		while (cols2.hasNext()) {
			temp = cols2.next();
			if(veccol.contains(temp)==true) {
				//colAndType2 += getTableName2()+"_";
				collist2 += getTableName2()+".";
			}
			//colAndType2 += temp + " " + schema.getTable(getTableName3()).getColType(temp);
			
			collist2 += temp;
			if (cols2.hasNext()) {
				collist2 += ",";
				//colAndType2 += ",";
			}
			//i++;
		}
		*/
		while (cols3.hasNext()) {
			temp = cols3.next();
			
			// CARLO: REMOVED TO FIX TARGET LIST TABLENAME MISMATCH!!
			//if(veccol3.contains(temp)==true) {
			//	collist3 += getTableName3()+".";
			//}
			collist3 += temp;
			if(cols3.hasNext()) {
				collist3 += ",";				
			}
		}
		Disjunctive conditions = getTargetConditions();
		
	//System.out.println("collist:"+collist+"\n");
	System.out.println("collist3:"+collist3+"\n");
	//System.out.println("colAndType:"+colAndType+"\n");
	//System.out.println("colAndType2:"+colAndType2+"\n");
		
		if(typeofDBMS.equalsIgnoreCase("MYSQL")) {
			System.out.println("getTableName1:"+getTableName());
			System.out.println("getTableName2:"+getTableName2());
			System.out.println("getTableName3:"+getTableName3());
			System.out.println("outputDBname :"+outputDBname);
			
			if(getTableName().equalsIgnoreCase(getTableName3()) | getTableName2().equalsIgnoreCase(getTableName3())) {
				sqlddl += "CREATE TABLE "+ outputDBname + "." + "TEMP_"+getTableName3()+"_TEMP" + " SELECT " + collist3 + " FROM " + outputDBname + "." + getTableName() + " JOIN "+ outputDBname + "." 
				+ getTableName2();
				if(conditions != null)
					sqlddl += " ON " + conditions.toString() + ";\n";
				else
					sqlddl += " ;\n";
			}
			else {
			sqlddl += "CREATE TABLE "+ outputDBname + "." + getTableName3() + " SELECT " + collist3 + " FROM " + outputDBname + "." + getTableName() + " JOIN "+ outputDBname + "."
			+ getTableName2();
			if(conditions != null)
				sqlddl += " ON " + conditions.toString() + ";\n";
			else
				sqlddl += " ;\n";
			}
			sqlddl += "DROP TABLE " + outputDBname + "." + getTableName() + ";\n";
			sqlddl += "DROP TABLE " + outputDBname + "." + getTableName2() + ";\n";
			if(getTableName().equalsIgnoreCase(getTableName3())) {
				sqlddl += "RENAME TABLE " + outputDBname+"."+"TEMP_"+getTableName3()+"_TEMP TO "+outputDBname+"."+getTableName() + ";\n";
			}
			else if(getTableName2().equalsIgnoreCase(getTableName3())) {
				sqlddl += "RENAME TABLE " + outputDBname+"."+"TEMP_"+getTableName3()+"_TEMP TO "+outputDBname+"."+getTableName2() + ";\n";
			}
		}
		else if(typeofDBMS.equalsIgnoreCase("DB2")) {
			// not implemented
			/*
			sqlddl += "CREATE TABLE " + outputDBname + "." + getTableName3()
					+ " ("+colAndType+","+colAndType2+");\n";
			sqlddl += "INSERT INTO " + outputDBname + "." + getTableName3()
					+ " (SELECT " + collist + ","+collist2 + " FROM " 
					+ outputDBname + "." + getTableName() + " JOIN "+ outputDBname + "." 
					+ getTableName2() + " ON " + conditions.toString() + ");\n";
			sqlddl += "DROP TABLE " + outputDBname + "." + getTableName() + ";\n";
			sqlddl += "DROP TABLE " + outputDBname + "." + getTableName2() + ";\n";
			return "not implemented";
			*/
		}
		else {
			System.out.println("Wrong typeofDBMS\n");
		}
		return sqlddl;
		
	}

	/**
	 * check whether the JOIN is performed on a given column, used to check whether a 
	 * join of two table sharing a column is ok or not (if the column is also part of the join condition ok)
	 * @param column
	 * @return
	 */
	public boolean joinOnColumn(String column) {			
		Disjunctive conditions = getTargetConditions();
		if(conditions == null)
			return false;
		for(int i = 0; i<conditions.conjunctives.size();i++){
			for(int j=0;j<conditions.conjunctives.get(i).predicates.size();j++){
				Predicate p = conditions.conjunctives.get(i).predicates.get(i);
				if(p.op.equalsIgnoreCase("=") && 
				   (p.term2.columnName.equalsIgnoreCase(column) && p.term1.columnName.equalsIgnoreCase(column))){
					return true;
				}
					
			}
		}
		return false;
	}


	
}

