/**
 * Copyright (C) 2009 
 * Developer: Fabrizio Moroni fabrizio.moroni@yahoo.it
 * Contributor: Carlo Curino	carlo@curino.us
 * 
 *  * This file is part of SchemEvolutionToolsuite
 * 
 * SchemEvolutionToolsuite is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SchemaEvolutionToolsuite is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with SchemEvolutionToolsuite.  If not, see <http://www.gnu.org/licenses/>.
 */

package edu.ucla.cs.dbevolution.DbBackend;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Vector;

import edu.ucla.cs.dbevolution.ApplicationParameters;
import edu.ucla.cs.statistics.MonthRevision;
import edu.ucla.cs.statistics.StatisticValue;

/**
 * This class is the mapping of the database evolution tables
 * @author fangar
 *
 */
public class DbVersion {

	private String catalogName;
	private String schemaUniqueName;
	private String schemaName;
	private int majorOrdinalVersion;
	private int minorOrdinalVersion;
	private String svnVersion;
	private String comment;
	private Date tStart;
	private Date tEnd;
	private int state;
	private String errorCode;
	private String errorMsg;


	/**
	 * Retrieve the databases installed from the evolution database
	 * 
	 * @return Vector<String> list of the installed databases
	 * @throws SQLException
	 */
	public static Vector<DbVersion> getDbInstalled() throws SQLException {

		String sql = "	SELECT schema_unique_name " +
				"		FROM " + ApplicationParameters.getEvolutionTable() + " " +
				"		WHERE schema_name like ? " +
				"		ORDER BY tstart";
		
		Vector<DbVersion> dbs = new Vector<DbVersion>();
		DbVersion db = null;
		
		Connection conn = null;
		PreparedStatement pStmt = null;
		ResultSet rs = null;

		
		
		try {
			conn = DbUtil.connect(ApplicationParameters.getDbevolution());
			pStmt = conn.prepareStatement(sql);
			pStmt.setString(1, ApplicationParameters.getDbbasename() + "%");
			rs = pStmt.executeQuery();

			if (rs != null) {
				while (rs.next()) {
					db = new DbVersion();
					db.setSchemaUniqueName(rs.getString(1));
					dbs.add(db);
				}
				rs.close();
			}
			conn.close();
		} catch (SQLException sEx) {
			System.out
			.println("An error occured during retrieving the databases installed from the evolution database: "
					+ sEx.getMessage());
			throw sEx;
		}
		return dbs;
	}
	
	public static int getMaxRevisionLifetime(){

		//sql statement for the max lifetime
		String sql = "	SELECT MAX(date_diff) " +
				"		FROM (" +
				"				SELECT schema_unique_name, DATEDIFF(tend,tstart) AS date_diff " +
				"				FROM " + ApplicationParameters.getEvolutionTable() + " " +
				"				WHERE 	schema_unique_name like ? AND " +
				"						tend < (" +
				"								SELECT MAX(tend) " +
				"								FROM " + ApplicationParameters.getEvolutionTable() + " " +
				"								WHERE schema_unique_name like ?)" +
				"				ORDER BY tstart) AS x";
		
		PreparedStatement pStmt = null;
		Connection conn = null;
		ResultSet rs = null;
		
		
		int lifetime = 0;
		
		try{
			conn = DbUtil.connect(ApplicationParameters.getDbevolution());
			pStmt = conn.prepareStatement(sql);
			pStmt.setString(1,ApplicationParameters.getDbbasename() + "%");
			pStmt.setString(2,ApplicationParameters.getDbbasename() + "%");
			rs = pStmt.executeQuery();
			if(rs!=null){
				if(rs.next()){
					lifetime = rs.getInt(1);
				}
			}
			conn.close();
		}
		catch (SQLException e) {
			e.printStackTrace();
		}
		return lifetime;
	}
	
	
	/**
	 * Return the date of the first relase
	 * @return a date in a string format
	 * @throws SQLException
	 */
	public static DbVersion getDatabaseFirstRelease() throws SQLException {
		
		String sql = "	SELECT DATE(tstart) " +
				"		FROM " +  ApplicationParameters.getEvolutionTable() + " " +
				"		WHERE schema_name LIKE ? AND tstart = (" +
				"												SELECT MIN(tstart) " +
				"												FROM " +  ApplicationParameters.getEvolutionTable() + " " +
						"										WHERE schema_name LIKE ? )";
		DbVersion db = null;
		
		Connection conn = null;
		PreparedStatement pStmt = null;
		ResultSet rs = null;

		
		try {
			conn = DbUtil.connect(ApplicationParameters.getDbevolution());
			pStmt = conn.prepareStatement(sql);
			pStmt.setString(1, ApplicationParameters.getDbbasename());
			pStmt.setString(2, ApplicationParameters.getDbbasename());
			rs = pStmt.executeQuery();
			
			if(rs!= null){
				if(rs.next()){
					db = new DbVersion();
					db.setTStart(rs.getDate(1));
				}
			}
			conn.close();
		}
		catch (SQLException sEx) {
			System.out.println("An error occured during retrieving the databases installed from the evolution database:");
			System.out.println(sEx.getMessage());
			throw sEx;
		}
		return db;
	}
	
	
	
	/**
	 * Return the list of the databases installed (for the system in use)
	 * @return
	 */
	public static List<DbVersion> getDownloadedDatabases(){

		PreparedStatement pStmt=null;
		Connection conn=null;
		String sqlGetSchemas = "SELECT schema_unique_name,svn_version, state " +
				"				FROM " + ApplicationParameters.getEvolutionTable() + " " +
				"				WHERE schema_name = ?";
		
		ResultSet rs=null;
		DbVersion db=null;
		List<DbVersion> dbs = new ArrayList<DbVersion>();

		try{
			conn = DbUtil.connect(ApplicationParameters.getDbevolution());
			pStmt = conn.prepareStatement(sqlGetSchemas);
			pStmt.setString(1, ApplicationParameters.getDbbasename());
			rs = pStmt.executeQuery();

			while (rs.next()) {
				//retrieve the values from the result set
				db = new DbVersion();
				db.setSchemaUniqueName(rs.getString(1));
				db.setSvnVersion(rs.getString(2));
				db.setState(rs.getInt(3));
				dbs.add(db);
			}
			conn.close();
		}
		catch (SQLException e) {
			System.out.println("An error occurred during rteading the evolution table: " + e.getMessage());
			e.printStackTrace();
			System.exit(1);
		}
		return dbs;
	}
	
	
	/**
	 * Return the name of the database installed with the minimum date with respect to the parameters passed
	 * The first database where the table occured
	 * @param tableSchema
	 * @param tableName
	 * @return
	 * @throws SQLException
	 */
	public static DbVersion getSchemaByDate(String tableSchema, String tableName,String date) throws SQLException{

		String sql = "	SELECT schema_unique_name " +
				"		FROM " + ApplicationParameters.getEvolutionTable() + " " +
				"		WHERE schema_name = ? and tstart = ";
		
		Connection conn = null;
		PreparedStatement pStmt = null;
		ResultSet rs = null;
		DbVersion db = null;

		if(date.equalsIgnoreCase("min"))
			sql+="(	SELECT MIN(tstart) " +
				"	FROM " +  ApplicationParameters.getEvolutionTable() + " d, " +  ApplicationParameters.getHeaderEvolutionTables() + "TABLES t " +
				"	WHERE d.schema_unique_name = t.table_schema and schema_name=? and t.table_name = ?)";
		else
			sql+=	"(	SELECT MAX(tstart) " +
					"	FROM " +  ApplicationParameters.getEvolutionTable() + " d, " +  ApplicationParameters.getHeaderEvolutionTables() + "TABLES t " +
					"	WHERE d.schema_unique_name = t.table_schema and schema_name=? and t.table_name = ?)";

		try {
			conn = DbUtil.connect(ApplicationParameters.getDbevolution());
			pStmt = conn.prepareStatement(sql);

			pStmt.setString(1, tableSchema);
			pStmt.setString(2, tableSchema);
			pStmt.setString(3, tableName);

			rs = pStmt.executeQuery();
			if (rs != null) {
				rs.next();
				db.setSchemaUniqueName(rs.getString(1));
			}
			conn.close();
		} 
		catch (SQLException sEx) {
			System.out.println("An error is occured during executing statistic: "+ sEx.getMessage());
			throw sEx;
		}
		return db;
	}
	
	
	/**
	 * Return the date of the last relase
	 * @return a date in a string format
	 * @throws SQLException
	 */
	public static DbVersion getDatabaseLastRelease() throws SQLException {
		
		String sql = "	SELECT schema_unique_name, DATE(tstart) " +
				"		FROM " +  ApplicationParameters.getEvolutionTable() + " " +
				"		WHERE schema_name LIKE ? AND tstart = (" +
				"												SELECT MAX(tstart) " +
				"												FROM " +  ApplicationParameters.getEvolutionTable() + " " +
						"										WHERE schema_name LIKE ? )"; 
		
		DbVersion db = null;
		
		Connection conn = null;
		PreparedStatement pStmt = null;
		ResultSet rs = null;

		
		try {
			conn = DbUtil.connect(ApplicationParameters.getDbevolution());
			pStmt = conn.prepareStatement(sql);
			pStmt.setString(1, ApplicationParameters.getDbbasename());
			pStmt.setString(2, ApplicationParameters.getDbbasename());
			rs = pStmt.executeQuery();
			
			if(rs!= null){
				if(rs.next()){
					db = new DbVersion();
					db.setSchemaUniqueName(rs.getString(1));
					db.setTEnd(rs.getDate(2));
				}
			}
			conn.close();
		}
		catch (SQLException sEx) {
			System.out.println("An error occured during retrieving the databases installed from the evolution database:");
			System.out.println(sEx.getMessage());
			throw sEx;
		}
		return db;
		
	}
	
	
	/**
	 * Get the count of the columns for every database
	 * @return a vector containing the count of the table for the selected schema
	 */
	public static Vector<StatisticValue> getRevisionLifetime()throws SQLException {

		Vector<StatisticValue> chartValue = new Vector<StatisticValue>();

		Connection conn = null;
		PreparedStatement pStmt = null;
		ResultSet rs = null;
		
		String sql = "	SELECT schema_unique_name, DATEDIFF(tend,tstart) " +
				"		FROM " + ApplicationParameters.getEvolutionTable() + " " +
				"		WHERE schema_unique_name like ? AND tend < (" +
				"													SELECT MAX(tend) " +
				"													FROM " + ApplicationParameters.getEvolutionTable() + " " +
				"													WHERE schema_unique_name like ?) " +
				"		ORDER BY tstart";

		try {
			conn = DbUtil.connect(ApplicationParameters.getDbevolution());
			pStmt = conn.prepareStatement(sql);
			pStmt.setString(1, ApplicationParameters.getDbbasename() + "%");
			pStmt.setString(2, ApplicationParameters.getDbbasename() + "%");

			rs = pStmt.executeQuery();
			if (rs != null) {
				while (rs.next()) {
					StatisticValue statColCount = new StatisticValue();
					statColCount.setX(rs.getString(1));
					statColCount.setY(rs.getInt(2));
					chartValue.add(statColCount);
				}
				rs.close();
				rs = null;
			}
			conn.close();
			conn = null;

			pStmt.close();
			pStmt = null;
		} catch (SQLException sEx) {
			System.out.println("An error is occured during executing statistic: "+ sEx.getMessage());
			throw sEx;
		}
		return chartValue;
	}

	
	
	/**
	 * Get the number of month each revision was live
	 * @param schema 
	 * @param sql
	 * @return
	 * @throws SQLException
	 */
	public static Vector<MonthRevision> getMonthRevisions(String schema, String sql)throws SQLException {

		Vector<MonthRevision> chartValue = new Vector<MonthRevision>();

		Connection conn = null;
		PreparedStatement pStmt = null;
		ResultSet rs = null;

		try {
			conn = DbUtil.connect(ApplicationParameters.getDbevolution());
			pStmt = conn.prepareStatement(sql);
			pStmt.setString(1, schema + "%");

			rs = pStmt.executeQuery();
			if (rs != null) {
				while (rs.next()) {
					MonthRevision month = new MonthRevision();
					month.setMonth(rs.getString(2)); 
					month.setYear(rs.getString(1));
					month.setRevisions(rs.getInt(3));
					chartValue.add(month);
				}
				rs.close();
				rs = null;
			}
			conn.close();
			conn = null;

			pStmt.close();
			pStmt = null;
		} catch (SQLException sEx) {
			System.out.println("An error is occured during executing statistic: "+ sEx.getMessage());
			throw sEx;
		}
		return chartValue;
	}
	
	public String getCatalogName() {
		return catalogName;
	}
	public void setCatalogName(String catalogName) {
		this.catalogName = catalogName;
	}
	public String getSchemaUniqueName() {
		return schemaUniqueName;
	}
	public void setSchemaUniqueName(String schemaUniqueName) {
		this.schemaUniqueName = schemaUniqueName;
	}
	public String getSchemaName() {
		return schemaName;
	}
	public void setSchemaName(String schemaName) {
		this.schemaName = schemaName;
	}
	public int getMajorOrdinalVersion() {
		return majorOrdinalVersion;
	}
	public void setMajorOrdinalVersion(int majorOrdinalVersion) {
		this.majorOrdinalVersion = majorOrdinalVersion;
	}
	public int getMinorOrdinalVersion() {
		return minorOrdinalVersion;
	}
	public void setMinorOrdinalVersion(int minorOrdinalVersion) {
		this.minorOrdinalVersion = minorOrdinalVersion;
	}
	public String getSvnVersion() {
		return svnVersion;
	}
	public void setSvnVersion(String svnVersion) {
		this.svnVersion = svnVersion;
	}
	public String getComment() {
		return comment;
	}
	public void setComment(String comment) {
		this.comment = comment;
	}
	
	public int getState() {
		return state;
	}
	public void setState(int state) {
		this.state = state;
	}
	public String getErrorCode() {
		return errorCode;
	}
	public void setErrorCode(String errorCode) {
		this.errorCode = errorCode;
	}
	public String getErrorMsg() {
		return errorMsg;
	}
	public void setErrorMsg(String errorMsg) {
		this.errorMsg = errorMsg;
	}

	public Date getTStart() {
		return tStart;
	}

	public void setTStart(Date start) {
		tStart = start;
	}

	public Date getTEnd() {
		return tEnd;
	}

	public void setTEnd(Date end) {
		tEnd = end;
	}
	

	
}
