package edu.ucla.cs.dbevolution.persistence.sqlDAO;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;

import edu.ucla.cs.dbevolution.ApplicationParameters;
import edu.ucla.cs.dbevolution.DbBackend.DbColumn;
import edu.ucla.cs.dbevolution.DbBackend.DbTable;
import edu.ucla.cs.dbevolution.DbBackend.DbTableExtended;
import edu.ucla.cs.dbevolution.DbBackend.DbUtil;
import edu.ucla.cs.dbevolution.DbBackend.DbVersion;
import edu.ucla.cs.dbevolution.DbBackend.TablesLifetime;

public class DbTableDAO {

	

	/**
	 * Return the number of tables in the database at the first revision
	 * @return
	 * @throws SQLException
	 */
	public static int getDatabaseInitialSize() throws SQLException {
		
		Integer size = null;
		
		PreparedStatement ps = null;
		ResultSet rs = null;
		Connection conn = null;
		
		HashMap<String, Object> params = new HashMap<String, Object>();
		params.put("SCHEMA_NAME",ApplicationParameters.getDatabaseQueryLikeName());
		
		String sql = "	SELECT count( * )" +
		"		FROM PRISM_VERSIONS v, PRISM_TABLES t" +
		"		WHERE tstart = (" +
		"						SELECT min( tstart )" +
		"						FROM PRISM_VERSIONS" +
		"						WHERE schema_unique_name LIKE  )" +
		"						AND v.schema_unique_name = t.table_schema";
		
		try{
			conn = DbUtil.connect(ApplicationParameters.getDbevolution());
			ps = conn.prepareStatement(sql);
			ps.setString(1,(String)params.get("SCHEMA_NAME"));
			rs = ps.executeQuery();
			while(rs.next()){
				size = rs.getInt(1);
			}
		}
		catch(SQLException sEx){
			throw sEx;
		}
		finally{
			conn.close();
			rs.close();
			ps.close();
			rs = null;
			ps = null;
		}
		return size;
	}
	
	
	
	
	/**
	 * Return the name of the tables for the specified database 
	 * @param db name
	 * @return Vector of tables name
	 * @throws SQLException
	 */  
	public static ArrayList<DbTable> getTables(String db) throws SQLException {

		ArrayList<DbTable> tables = null;
		
		PreparedStatement ps = null;
		ResultSet rs = null;
		Connection conn = null;
		
		HashMap<String, Object> params = new HashMap<String, Object>();
		params.put("SCHEMA_NAME",ApplicationParameters.getDbbasename());
		
		String sql = "	SELECT	DISTINCT TABLE_NAME" +
		"		FROM 	PRISM_TABLES" +
		"		WHERE 	TABLE_SCHEMA LIKE ?";
		
		try{
			conn = DbUtil.connect(ApplicationParameters.getDbevolution());
			ps = conn.prepareStatement(sql);
			ps.setString(1,(String)params.get("SCHEMA_NAME"));
			rs = ps.executeQuery();
			
			tables = new ArrayList<DbTable>();
			
			while(rs.next()){
				DbTable table = new DbTable();
				table.setTABLE_NAME(rs.getString(1));
				tables.add(table);
			}
		}
		catch(SQLException sEx){
			throw sEx;
		}
		finally{
			conn.close();
			rs.close();
			ps.close();
			rs = null;
			ps = null;
		}
		return tables;
	}
	
	
	/**
	 * Return the number of tables in the database at the last revision
	 * @return
	 * @throws SQLException
	 */
	public static int getDatabaseEndSize() throws SQLException {
		
		int size = 0;
		
		PreparedStatement ps = null;
		ResultSet rs = null;
		Connection conn = null;
		
		HashMap<String, Object> params = new HashMap<String, Object>();
		params.put("SCHEMA_NAME",ApplicationParameters.getDatabaseQueryLikeName());
		
		String sql = "	SELECT count( * )" +
				"		FROM PRISM_VERSIONS v, PRISM_TABLES t" +
				"		WHERE tstart = (" +
				"						SELECT max( tstart )" +
				"						FROM PRISM_VERSIONS" +
				"						WHERE schema_unique_name LIKE ? )" +
				"						AND v.schema_unique_name = t.table_schema";
		
		try{
			conn = DbUtil.connect(ApplicationParameters.getDbevolution());
			ps = conn.prepareStatement(sql);
			ps.setString(1,(String)params.get("SCHEMA_NAME"));
			rs = ps.executeQuery();
			while(rs.next()){
				size = rs.getInt(1);
			}
		}
		catch(SQLException sEx){
			throw sEx;
		}
		finally{
			conn.close();
			rs.close();
			ps.close();
			rs = null;
			ps = null;
		}
		return size;
	}
	
	
	/**
	 * 
	 * @return
	 * @throws SQLException
	 */
	public static List<TablesLifetime> getTablesLifetime() throws SQLException{
		return (List<TablesLifetime>)SqlMapDAOManager.getSqlClient().queryForList("getTablesLifetime", ApplicationParameters.getDatabaseQueryLikeName());
	}
	
	
	public static int getMaxTablesLifetime() throws SQLException{
		
		int lifetime = 0;
		
		PreparedStatement ps = null;
		ResultSet rs = null;
		Connection conn = null;
		
		HashMap<String, Object> params = new HashMap<String, Object>();
		params.put("SCHEMA_NAME",ApplicationParameters.getDatabaseQueryLikeName());
		
		String sql = "	SELECT max(count)" +
		"		FROM (" +
		"				SELECT TABLE_NAME AS t_name, COUNT(*) AS count" +
		"				FROM PRISM_TABLES" +
		"				WHERE TABLE_SCHEMA LIKE ? GROUP BY t_name) AS x";
		
		try{
			conn = DbUtil.connect(ApplicationParameters.getDbevolution());
			ps = conn.prepareStatement(sql);
			ps.setString(1,(String)params.get("SCHEMA_NAME"));
			rs = ps.executeQuery();
			while(rs.next()){
				lifetime = rs.getInt(1);
			}
		}
		catch(SQLException sEx){
			throw sEx;
		}
		finally{
			conn.close();
			rs.close();
			ps.close();
			rs = null;
			ps = null;
		}
		return lifetime;
	}
	
	/**
	 * Get the name of the tables with index;
	 * use the table table_privileges to perform a comparison and discard those tuples corresponding to primary keys and foreign keys
	 * @param db
	 * @return 
	 * @return
	 * @throws SQLException
	 */
	public static List<DbTableExtended> getTablesWithIndexes(String dbName) throws SQLException {
		return (List<DbTableExtended>)SqlMapDAOManager.getSqlClient().queryForList("getTablesWithIndexes", ApplicationParameters.getDatabaseQueryLikeName());
	}
	
	
	/**
	 * TO BE REPLACED BY THE IBATIS VERSION
	 * Get the name of the columns
	 * @param dbName the database name
	 * @param tableName the table name
	 * @return column's name for each revision of the table
	 * @throws SQLException
	 */
	public static LinkedHashMap<DbVersion, List<DbColumn>>  getColumnsInTable(String dbName,String tableName) throws SQLException {
		return null;
	}
	
	
}
