/**
 * Copyright (C) 2009 
 * Developer: Fabrizio Moroni fabrizio.moroni@yahoo.it
 * Contributor: Carlo Curino	carlo@curino.us
 * 
 *  * This file is part of SchemEvolutionToolsuite
 * 
 * SchemEvolutionToolsuite is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SchemaEvolutionToolsuite is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with SchemEvolutionToolsuite.  If not, see <http://www.gnu.org/licenses/>.
 */

package edu.ucla.cs.statistics;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Set;

import org.jfree.data.category.DefaultCategoryDataset;

import edu.ucla.cs.dbevolution.ApplicationParameters;
import edu.ucla.cs.dbevolution.DbBackend.DbColumn;
import edu.ucla.cs.dbevolution.DbBackend.DbStatistics;
import edu.ucla.cs.dbevolution.DbBackend.DbTable;
import edu.ucla.cs.dbevolution.DbBackend.DbTableConstraints;
import edu.ucla.cs.dbevolution.DbBackend.DbVersion;
import edu.ucla.cs.dbevolution.DbBackend.IndexesInTable;
import edu.ucla.cs.dbevolution.persistence.sqlDAO.DbTableDAO;


/**
 * Create the statistic that inspect every single table, looking for a change in the key
 * The statistic analyze a table, checking if the keys are changed during the different revisions
 * @throws SQLException
 */
public class IndexChanges extends Statistic implements StatisticRisk{


	/**
	 * TODO
	 * Set the default filed for the statistic
	 */
	public IndexChanges(){

		setTitle("Index Changes");
		setStatDescription("The statistic show the number of primary keys modified during the lifetime of a table\n The aggregation level is the table name. " +
		"This can be used as an index of database stability");
		setSqlStmt("The sql analyze the number of columns belonging to the key in each revision of the databas during the table lifetime");
		setRange("Operations");
		setDomain("Tables");
		setStatName("statIndexChanges");
		setChartOrientation("h");
		setChartType("bar");
		setTitlePage("Statistic: Index Changes");
		setStatPage(ApplicationParameters.getDbbasename() + getStatName() + ".html");
		setStatGroup("indexes");
	}


	private HashMap<String, Integer> tablesIndexChanges;


	/**
	 * TODO
	 * @throws SQLException 
	 * 
	 */
	public void execute() throws SQLException{
		setDs(indexChangesDataset());
		executeStatistic(this);
	}


	/**
	 * Set the dataset of the statistic
	 * @param ds the DefaultCategoryDataset
	 */
	public void setDataset(DefaultCategoryDataset ds){
		setDs(ds);
	}


	/**
	 * TODO
	 * @return
	 * @throws SQLException
	 */
	public DefaultCategoryDataset indexChangesDataset() throws SQLException{

		//the analysis is performed for single index and for single table,
		//but the result are aggregated over the table to make compact the structure of the chart

		//counter for the changes in the columns name for the index and the entire table

		String dbName = ApplicationParameters.getDbbasename() + "%";

		int columnNameChanges = 0;
		int columnSizeChanges = 0;

		int tableNameChanges = 0;
		int tableSizeChanges = 0;

		//risk mapping
		HashMap<String,Integer> tablesIndexChanges = new HashMap<String, Integer>();

		//create the dataset
		DefaultCategoryDataset ds = new DefaultCategoryDataset();
		List<IndexesInTable> tables = DbTableDAO.getTablesWithIndexes(dbName);


		//create the CSV file and add the header to the file
		createCSVFile(this);
		addValueToCsv(this, new String[]{"TableName","IndexChanges - Count","IndexChange - Name"});

		for(IndexesInTable idxTable: tables){

			//for each table
			//for each index in the table
			//compare the number of columns in the index
			//compare the name of the columns in the index

			DbTable table = (DbTable)idxTable.getTable();
			String tableName = table.getTABLE_NAME();
			ArrayList<DbTableConstraints> indexes= (ArrayList<DbTableConstraints>)idxTable.getIndexes();
			//DbStatistics index = indexes.get(0);

			for(int i=0;i<indexes.size(); i++){

				DbTableConstraints index = indexes.get(i);

				System.out.println("INDEX_CHANGES - PERFORMING TABLE: " + tableName + ", INDEX: " + index.getCONSTRAINT_NAME());

				//for each index in the table
				//retrieve the revision where the table was present
				//get the columns belonging to the index and check if there've been changes

				LinkedHashMap<DbVersion, List<DbColumn>> result = DbStatistics.getColumnsInIndex(dbName, tableName, index.getCONSTRAINT_NAME());

				Set<DbVersion> set = result.keySet();

				Object[] dbs = set.toArray();

				for(int j=0;j<result.size()-1;j++){

					ArrayList<DbColumn> column1 = (ArrayList<DbColumn>)result.get(dbs[j]);
					ArrayList<DbColumn> column2 = (ArrayList<DbColumn>)result.get(dbs[j+1]);
					columnNameChanges += DbColumn.compareColumnsName(column1, column2);
					columnSizeChanges += DbColumn.compareColumnsSize(column1, column2);
				}

				tableNameChanges += columnNameChanges;
				tableSizeChanges += columnSizeChanges;

				columnNameChanges = 0;
				columnSizeChanges = 0;

			}

			//add values to the risk mapping
			//add the value to the mapping only if there's a column added in the primary key
			tablesIndexChanges.put(tableName, tableNameChanges);

			//add the value to the csv
			addValueToCsv(this, new String[]{tableName,String.valueOf(tableNameChanges), String.valueOf(tableSizeChanges)});

			//add values to the dataset
			addValueToDataset(ds, tableNameChanges, "colname", tableName);
			addValueToDataset(ds, tableSizeChanges, "colSize", tableName);

			tableNameChanges = 0;
			tableSizeChanges = 0;
		}

		setTablesPrimaryChanges(tablesIndexChanges);

		return ds;
	}




	/**
	 * TODO
	 * @param tableName
	 * @return
	 */
	public int getRiskMapping(String tableName) {
		try{
			return tablesIndexChanges.get(tableName);
		}
		catch(NullPointerException nEx){
			return 0;
		}
	}


	public void setTablesPrimaryChanges(
			HashMap<String, Integer> tablesPrimaryChanges) {
		this.tablesIndexChanges = tablesPrimaryChanges;
	}


}
