/**
 * Copyright (C) 2009 
 * Developer: Fabrizio Moroni fabrizio.moroni@yahoo.it
 * Contributor: Carlo Curino	carlo@curino.us
 * 
 *  * This file is part of SchemEvolutionToolsuite
 * 
 * SchemEvolutionToolsuite is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SchemaEvolutionToolsuite is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with SchemEvolutionToolsuite.  If not, see <http://www.gnu.org/licenses/>.
 */

package edu.ucla.cs.statistics;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;

import org.jfree.data.category.DefaultCategoryDataset;

import edu.ucla.cs.dbevolution.ApplicationParameters;
import edu.ucla.cs.dbevolution.DbBackend.DbColumn;
import edu.ucla.cs.dbevolution.DbBackend.DbTable;
import edu.ucla.cs.dbevolution.DbBackend.DbVersion;
import edu.ucla.cs.dbevolution.persistence.sqlDAO.DbTableDAO;

/**
 * Create and execute the statistic that performs an analysis over the database, retrieving if there are tables that over
 * their lifetime are evolved and have been rolled back to a previous version 
 *	@author fangar
 */
public class RolledBackTables extends Statistic implements StatisticRisk{

	private HashMap<String, Integer> tableRollbacks;

	/**
	 * TODO
	 * Set the default filed for the statistic
	 */
	public RolledBackTables() {
		setTitle("Rolledback tables");
		setStatDescription("The statistic analyze the evolution's story of the table, looking into possible situations of rollbacks; " +
				"a \"rollback\" occurs when a situation like the one described here occurs: at the time n-1 the table has a structure x; then at the time n, the table is modified by the database administrator" +
				"changing his structure; the new structure is called y; the database evolvs and the db admin decide, at the time n+1, to change again the structure of the table into x;" +
		"the table has undergone a rollback changing its structure from x to y and then back to x again");
		setSqlStmt("The sql analyze the structure, in terms of columns, of the table during each revision comparing it with the structure of the revision + 1; creates a checkpoint every change of structure and then analyze each checkpoint" +
		"looking for possible rollback situations");
		setRange("rollbacks");
		setDomain("Tables");
		setStatName("statRolledbackTables");
		setChartOrientation("h");
		setChartType("bar");
		setTitlePage("Statistic: Rolledback Tables");
		setStatPage(ApplicationParameters.getDbbasename() + getStatName() + ".html");
		setStatGroup("tables");
	}


	/**
	 * TODO
	 * @throws SQLException 
	 * 
	 */
	public void execute() throws SQLException{
		setDs(getTablesRolledBack());
		executeStatistic(this);
	}


	/**
	 * @param table
	 * @return
	 * @throws SQLException 
	 */
	private DefaultCategoryDataset getTablesRolledBack() throws SQLException{

		int rollbackCount;
		DbVersion dbVersion = null;
		ArrayList<Savepoint> points = null;
		Savepoint sPrev = null;

		//dataset
		DefaultCategoryDataset ds = new DefaultCategoryDataset();

		//mapping of the table with its number of rollback
		HashMap<String, Integer> tableRollbacks = new HashMap<String, Integer>();

		ArrayList<DbTable> tables = DbTableDAO.getTables(ApplicationParameters.getDbbasename()+"%");

		if(tables!=null){

			//create the CSV file and add the header to the file
			createCSVFile(this);
			addValueToCsv(this, new String[]{"TableName","Rollbacks"});

			for(DbTable table:tables){
				rollbackCount = 0;


				System.out.println("ROLLBACK - PERFORMING TABLE: " + table.getTABLE_NAME());

				points = new ArrayList<Savepoint>();
				sPrev = null;

				LinkedHashMap<DbVersion, List<DbColumn>> result = DbTable.getColumnsInTable(ApplicationParameters.getDbbasename()+"%", table.getTABLE_NAME());


				if(result!=null){

					Iterator<DbVersion> itResult = result.keySet().iterator();

					//for(int i=1;i<result.size(); i++){
					while(itResult.hasNext()){

						dbVersion = itResult.next();

						Savepoint s = new Savepoint();
						s.setColumnsCount(result.get(dbVersion).size());
						s.setColumnsName((ArrayList<DbColumn>)result.get(dbVersion));
						//points.add(s);

						if(sPrev!=null){
							if(sPrev.getColumnsCount()!=s.getColumnsCount()){
								sPrev = s;
								points.add(s);
							}
						}
						else{
							sPrev = s;
						}
					}
				}


				//search for a rolled back situation
				rollbackCount += checkTableRolledBack(points);



				//the analysis of the table ended, write the number of rollback into the csv file

				//add the values to the hashmap
				tableRollbacks.put(table.getTABLE_NAME(), new Integer(rollbackCount));

				if(rollbackCount!=0){
					//add values to the dataset only if different from zero
					//avoid filling the table with lots of zero values
					addValueToDataset(ds, rollbackCount, "rollbacks", table.getTABLE_NAME());
				}

				//add the value to the csv
				//the csv file is complet, contains also the zero rolled back tables
				addValueToCsv(this, new String[]{table.getTABLE_NAME(),String.valueOf(rollbackCount)});	

			}

			//set the mapping to the local one
			setTableRollbacks(tableRollbacks);
		}
		return ds;
	}






	/**
	 * @param points
	 */
	private int checkTableRolledBack(ArrayList<Savepoint> points) {

		int rollbackCount = 0;
		for(int countPoints = points.size()-1;countPoints>=0;countPoints--){
			Savepoint s = (Savepoint)points.get(countPoints);
			//compare savepoint with previous ones looking for rolledback situation
			for(int countPointsPrev =  0 + countPoints-1;countPointsPrev>=0;countPointsPrev--){
				if(s.compareSavepoints(s, points.get(countPointsPrev))){
					rollbackCount++;
				}
			}
		}
		return rollbackCount;

	}

	//				
	//				Set tableNames = result.keySet();
	//
	//				Object[] tableNamesArray = tableNames.toArray();
	//
	//				for(int countPoints = tableNamesArray.length-1;countPoints>=0;countPoints--){
	//					ArrayList<DbColumn> s = (ArrayList<DbColumn>)result.get(tableNamesArray[countPoints]);
	//					//compare savepoint with previous ones looking for rolledback situation
	//					for(int countPointsPrev =  0 + countPoints-1;countPointsPrev>=0;countPointsPrev--){
	//						ArrayList<DbColumn> s1 = (ArrayList<DbColumn>)result.get(tableNamesArray[countPointsPrev]);
	//						if(DbTable.compareTableStructure(s, s1)){
	//							rollbackCount++;
	//						}
	//					}
	//				}
	//
	//				//the analysis of the table ended, write the number of rollback into the csv file
	//
	//				//add the values to the hashmap
	//				tableRollbacks.put(table.getTABLE_NAME(), new Integer(rollbackCount));
	//
	//				if(rollbackCount!=0){
	//					//add values to the dataset only if different from zero
	//					//avoid filling the table with lots of zero values
	//					addValueToDataset(ds, rollbackCount, "rollbacks", table.getTABLE_NAME());
	//				}
	//
	//				//add the value to the csv
	//				//the csv file is complet, contains also the zero rolled back tables
	//				addValueToCsv(this, new String[]{table.getTABLE_NAME(),String.valueOf(rollbackCount)});	
	//				
	//				
	//				rollbackCount = 0;
	//			}
	//		}
	//		//set the mapping to the local one
	//		setTableRollbacks(tableRollbacks);
	//
	//		return ds;





	/**
	 * TODO
	 * @param tableName
	 * @return
	 */
	public int getRiskMapping(String tableName) {
		try{
			return tableRollbacks.get(tableName);
		}
		catch(NullPointerException nEx){
			return 0;
		}
	}



	public void setTableRollbacks(HashMap<String, Integer> tableRollbacks) {
		this.tableRollbacks = tableRollbacks;
	}





}
