/**
 * Copyright (C) 2009 
 * Developer: Fabrizio Moroni fabrizio.moroni@yahoo.it
 * Contributor: Carlo Curino	carlo@curino.us
 * 
 *  * This file is part of SchemEvolutionToolsuite
 * 
 * SchemEvolutionToolsuite is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SchemaEvolutionToolsuite is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with SchemEvolutionToolsuite.  If not, see <http://www.gnu.org/licenses/>.
 */

package edu.ucla.cs.statistics;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.io.File;
import java.io.IOException;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartUtilities;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.CategoryAxis;
import org.jfree.chart.axis.DateAxis;
import org.jfree.chart.axis.DateTickUnit;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.NumberTickUnit;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.plot.CategoryPlot;
import org.jfree.chart.plot.Marker;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.ValueMarker;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.category.BarRenderer;
import org.jfree.chart.renderer.category.LineAndShapeRenderer;
import org.jfree.chart.renderer.xy.XYBarRenderer;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.data.category.DefaultCategoryDataset;
import org.jfree.data.xy.IntervalXYDataset;
import org.jfree.data.xy.XYSeriesCollection;

import edu.ucla.cs.dbevolution.ApplicationParameters;
import edu.ucla.cs.dbevolution.Utils;


/**
 * The class receive the type of the statistic and the schema over which perform it, create the chart with the option 
 * selected by the user and return the image of the graph
 * @author fangar
 *
 */

public class StatisticChart {

	/**
	 * The class create a chart for the given statistic
	 * The field represent the option of the chart that can be customized
	 */
	
	private String imageName;
	private DefaultCategoryDataset dataset;
	private IntervalXYDataset intervalDataset;
	private XYSeriesCollection xyDataset;
	private String chartTitle;
	private String domain;
	private String range;
	
	private PlotOrientation orientation;
	
	private boolean tooltips;
	private boolean legend;
	private boolean urls;

	private String sqlStat;
	private String schema;

	private String pathToChart;
	
	private int width;
	private int height;
	
	private final int factorWidth = 15;
	private final int factorHeight = 23;
	
	//defines the minimum pixel width for a bar in a bar chart
	private final int minBarWidth = 13;
	//defines the margin to be added to a bar chart
	//takes into account the space needed for the range/domain label and the space for the axis
	private final int marginXBarChart = 40;
	private final int marginYBarChart = 50;
	
	private int columnsCount;
	
	private final String imageExtension = ".png"; 
	
	/** 
	 * Constructor
	 * @param schema the name of the schema over which execute the statistic
	 * @param stat the type of statistic to execute
	 */
	/*public StatisticChart(String stat){
		this.schema = ApplicationParameters.getDbbasename();
		this.imageName = stat;
	}*/
	
	public StatisticChart(){

		this.schema = ApplicationParameters.getDbbasename();
		//set the common chart variable
		//the user can anyway override this settings by calling the corresponding methods
		this.orientation = PlotOrientation.VERTICAL;
		this.legend = false;
		this.tooltips = false;
		this.urls = false;
		
		//if not set, the chart dimension are the default set into the configuration file
		this.width = 1024; 
		this.height = 800;
	}
	
	/**
	 * Set some of the chart data
	 * @param title chart's title
	 * @param range chart's range
	 * @param domain chart's domain
	 */
	public void setChartData(String title, String range, String domain){
		/*this.chartTitle = title;*/
		this.domain = domain;
		this.range = range;
	}
	
	/**
	 * Set others chart data
	 * @param legend chart's legend
	 * @param tooltips chart's tooltips
	 * @param url chart's url
	 * @param or chart's orientation
	 */
	/*public void fillCommonData(boolean legend, boolean tooltips, boolean url, PlotOrientation or){
		this.orientation = or;
		this.legend = legend;
		this.tooltips = tooltips;
		this.urls = url;
	}
	*/
		
		
	
	/**
	 * Create a Line chart 
	 * @throws IOException 
	 */
	public void createLineChart() throws IOException{
			
		//create the chart...
		JFreeChart chart = ChartFactory.createLineChart(
		    this.chartTitle, // chart title
		    this.domain, // domain axis label
		    this.range, // range axis label
		    this.dataset, // data
		    this.orientation, // orientation
		    this.legend, // include legend
		    this.tooltips, // tooltips
		    this.urls // urls
		);
		
		//customize the chart
		customizeChart(chart);
		
		saveChartImage(chart);
	}
	
	
	public void createXYLineChart() throws IOException{
		
		//create the chart...
		JFreeChart chart = ChartFactory.createXYLineChart(
		    this.chartTitle, // chart title
		    this.domain, // domain axis label
		    this.range, // range axis label
		    this.xyDataset, // data
		    this.orientation, // orientation
		    this.legend, // include legend
		    this.tooltips, // tooltips
		    this.urls // urls
		);
		
		//customize the chart
		customizeChart(chart);
		
		saveChartImage(chart);
	}
	
	
	
	/**
	 * Create a XyBar chart
	 * @throws IOException 
	 */
	public void createXyBarChart() throws IOException{
		

		//create the chart...
		JFreeChart chart = ChartFactory.createXYBarChart(
		    this.chartTitle, // chart title
		    this.domain, // domain axis label
		    true,
		    this.range, // range axis label
		    this.intervalDataset, // data
		    this.orientation, // orientation
		    this.legend, // include legend
		    this.tooltips, // tooltips
		    this.urls // urls
		);
		
		//customize the chart
		customizeChart(chart);
				
		saveChartImage(chart);
	}
	

	

	
	/**
	 * Create a bar chart
	 * @throws IOException 
	 */
	public void createBarChart() throws IOException{
		
		// create the chart...
		JFreeChart chart = ChartFactory.createBarChart(
		    this.chartTitle, 
		    this.domain,
		    this.range,
		    this.dataset,
		    this.orientation,
		    this.legend,
		    this.tooltips,
		    this.urls
		);

		//customize the chart
		customizeChart(chart);
		setDimension(chart);
		saveChartImage(chart);
		
		//this.chart = chart;
	}
	
	
	/**
	 * Save the chart as png image on the user selected directory
	 * @param chart the chart to save
	 * @throws IOException
	 */
	private void saveChartImage(JFreeChart chart) throws IOException{
		
		File fImage = null;
		try{
		if((Utils.checkCreateDir(StatisticsParameters.getStatPagePath()) && Utils.checkCreateDir(StatisticsParameters.getStatImagePath()))){
			//System.out.println("CHECKING DIR");
			fImage = new File(StatisticsParameters.getStatImagePath() + this.schema + this.imageName.toLowerCase() + imageExtension);
			//ChartUtilities.saveChartAsPNG(fImage, chart, chartWidthDimension(this.dataset.getColumnCount()),chartHeightDimension(this.dataset.getColumnCount()));
			//ChartUtilities.saveChartAsPNG(fImage, chart, 1024,800);
			ChartUtilities.saveChartAsPNG(fImage, chart, this.width,this.height);
		}
		}
		catch(IOException iEx){
			System.out.println("An error occured during the creation of the image of the chart:" + iEx.getMessage());
			throw iEx;
		}
		this.pathToChart =  fImage.getPath();
	}
	
	
	/**
	 * Customize chart's property like axis rendering and tick unit.
	 * Handle different type of chart calling the appropriate customization method depending on the type of chart submitted
	 * @param chart the chart to customize
	 */
	private void customizeChart(JFreeChart chart){
		chartProperties(chart);
		
		//cutomize the axis properties
		if (chart.getPlot() instanceof CategoryPlot){
			CategoryPlot plot = (CategoryPlot)chart.getPlot();
			categoryAxisProperties(plot);
			//customise the rendering properties
			if (plot.getRenderer()  instanceof BarRenderer) {
				BarRenderer barRdr = (BarRenderer)plot.getRenderer();
				renderBarProperties(barRdr);
			}
			else if(plot.getRenderer() instanceof LineAndShapeRenderer){
				LineAndShapeRenderer lineRdr = (LineAndShapeRenderer)plot.getRenderer();
				renderLineProperties(lineRdr);
			}
		}
		if(chart.getPlot() instanceof XYPlot){
			XYPlot plot = (XYPlot)chart.getPlot();
			xyAxisProperties(plot);
			if (plot.getRenderer()  instanceof XYBarRenderer) {
				XYBarRenderer barRdr = (XYBarRenderer)plot.getRenderer();
				renderXYBarProperties(barRdr);
			}
			if (plot.getRenderer()  instanceof XYLineAndShapeRenderer) {
				XYLineAndShapeRenderer lineRdr = (XYLineAndShapeRenderer)plot.getRenderer();
				renderXYLineProperties(lineRdr);
			}
		}
	}
	
	
	/**
	 * Set the chart dimension based on the type of the chart
	 * @param chart 
	 */
	public void setDimension(JFreeChart chart) {
	
		if (chart.getPlot() instanceof CategoryPlot){
			CategoryPlot plot = (CategoryPlot)chart.getPlot();
			//customise the rendering properties
			if (plot.getRenderer()  instanceof BarRenderer) {
				//BarRenderer barRdr = (BarRenderer)plot.getRenderer();
				if(this.orientation == PlotOrientation.HORIZONTAL){
					this.width = 1024;
					this.height = minBarWidth * columnsCount + marginXBarChart;
					if(this.height<150){
						this.height = 150;
					}
				}
			}
			/*else if(plot.getRenderer() instanceof LineAndShapeRenderer){
				LineAndShapeRenderer lineRdr = (LineAndShapeRenderer)plot.getRenderer();
				renderLineProperties(lineRdr);
			}*/
		}
		/*if(chart.getPlot() instanceof XYPlot){
			XYPlot plot = (XYPlot)chart.getPlot();
			xyAxisProperties(plot);
			if (plot.getRenderer()  instanceof XYBarRenderer) {
				XYBarRenderer barRdr = (XYBarRenderer)plot.getRenderer();
				renderXYBarProperties(barRdr);
			}
			if (plot.getRenderer()  instanceof XYLineAndShapeRenderer) {
				XYLineAndShapeRenderer lineRdr = (XYLineAndShapeRenderer)plot.getRenderer();
				renderXYLineProperties(lineRdr);
			}
		}*/
	}
	
	
	/**
	 * Customize the rendering property of a bar chart
	 * @param barRdr the renderer
	 */
	private void renderBarProperties(BarRenderer barRdr){
		barRdr.setDrawBarOutline(false);
		barRdr.setItemMargin(0.0);
		barRdr.setMaximumBarWidth(0.05);
		barRdr.setSeriesPaint(0, new Color(104,173,254));
	}
	
	/**
	 * Customize the rendering property of an XY bar chart
	 * @param barRdr the renderer
	 */
	private void renderXYBarProperties(XYBarRenderer barRdr){
		barRdr.setDrawBarOutline(false);
		barRdr.setMargin(0.01);
		barRdr.setSeriesPaint(0, new Color(104,173,254));
		//set the base of the bar, the point to start drawing the bar
		barRdr.setBase(0.0);
	}
	
	/**
	 * Customize the rendering property of a line chart
	 * @param lineRdr the renderer
	 */
	private void renderLineProperties(LineAndShapeRenderer lineRdr){
		
		//get the render for a line chart
		//hide the square of the point x/y
		lineRdr.setShapesVisible(false);
		lineRdr.setDrawOutlines(false);
		lineRdr.setUseFillPaint(false);
		lineRdr.setSeriesPaint(0, new Color(104,173,254));
		lineRdr.setStroke(new BasicStroke(2.0f));
	}
	
	
	/**
	 * Customize the rendering property of an XY line chart
	 * @param the renderer
	 */
	private void renderXYLineProperties(XYLineAndShapeRenderer lineRdr){
		
		//get the render for a line chart
		//hide the square of the point x/y
		lineRdr.setShapesVisible(false);
		lineRdr.setDrawOutlines(false);
		lineRdr.setUseFillPaint(false);
		lineRdr.setSeriesPaint(0, new Color(104,173,254));
		lineRdr.setStroke(new BasicStroke(2.0f));
	}
	
	/**
	 * Customize the axis of a category chart
	 * @param plot the chart's plot
	 */
	private void categoryAxisProperties(CategoryPlot plot){

		//cutomize the range axis
	    valueAxisProperty(plot.getRangeAxis());
	    
		//domani axis
		CategoryAxis domainAxis = plot.getDomainAxis();
		domainAxis.setLabelFont(new Font("Arial",Font.BOLD,14));
		domainAxis.setTickLabelFont(new Font("Arial",Font.PLAIN,12));
		
		//set the lower and upper margin to 1% of the chart dimension
		domainAxis.setLowerMargin(0.01);
		domainAxis.setUpperMargin(0.01);
	}
	
	
	/**
	 * Customize the axis of an XY chart
	 * Set the properties for range axis and domain axis of a Xy plot
	 * Set font label , font tick and tick unit
	 * @param plot
	 */
	private void xyAxisProperties(XYPlot plot){
		
		//cutomize the range axis
	    valueAxisProperty(plot.getRangeAxis());
	    
		//cutomize the domain axis
		ValueAxis domainAxis = plot.getDomainAxis();
		domainAxis.setLabelFont(new Font("Arial",Font.BOLD,14));
		domainAxis.setTickLabelFont(new Font("Arial",Font.PLAIN,12));
		if(domainAxis instanceof DateAxis){
			
			DateAxis dateAxis = (DateAxis)domainAxis;
			
			Date dmin = dateAxis.getMinimumDate();
			Date dmax = dateAxis.getMaximumDate();
			
			Calendar cmin = new GregorianCalendar(dmin.getYear()+1900,dmin.getMonth(),dmin.getDay());
			Calendar cmax = new GregorianCalendar(dmax.getYear()+1900,dmax.getMonth(),dmax.getDay());
			
			//customize the date axis of the chart...
			// set the tick size to one week, with formatting...
			DateTickUnit unit = new DateTickUnit(DateTickUnit.YEAR, 1);
			dateAxis.setTickUnit(unit);
			
			//change the grid color and style
			plot.setDomainGridlineStroke(new BasicStroke(1.0f));
			
			//add month marker to the chart
			//Calendar d = new GregorianCalendar();
			//d.set(2001, 1, 01);

			double millis = cmin.getTimeInMillis();
		    Marker start = new ValueMarker(millis);
	        start.setPaint(Color.GRAY);
	        start.setStroke(new BasicStroke(0.1f));
	        plot.addDomainMarker(start);
			
			//while(d.before(new GregorianCalendar(2006,12,01))){
	        //while(d.before(new GregorianCalendar(dmax.getYear(),dmax.getMonth(),01))){
	        while(cmin.before(cmax)){
				cmin.add(Calendar.MONTH,1);
				
			    millis = cmin.getTimeInMillis();
			    start = new ValueMarker(millis);
		        start.setPaint(Color.GRAY);
		        start.setStroke(new BasicStroke(0.1f,BasicStroke.CAP_BUTT,BasicStroke.JOIN_MITER, 10.0f, new float[]{2.0f}, 0.0f));
		        plot.addDomainMarker(start);
			}
		}
	}
	
	/**
	 * Customize the range axis
	 * Set the range axis font and tick unit
	 * @param rangeAxis
	 */
	private void valueAxisProperty(ValueAxis rangeAxis){
		
		rangeAxis.setTickLabelFont(new Font("Arial",Font.PLAIN,12));
	    rangeAxis.setLabelFont(new Font("Arial",Font.BOLD,14));
	    rangeAxis.setVerticalTickLabels(false);
	    rangeAxis.setAutoRange(false);
	    
	    rangeAxis.setStandardTickUnits(NumberAxis.createIntegerTickUnits());
	    if(rangeAxis instanceof NumberAxis){
	    	NumberAxis axis = (NumberAxis) rangeAxis;
	    	axis.setAutoRangeIncludesZero(true);
	    	
	    	double tick = 1.0;
	    	double increment = 1.0;
	    	double rangeDiff = rangeAxis.getUpperBound() - rangeAxis.getLowerBound();
			//System.out.println("RANGE DIFF:" + rangeDiff);
	    	if(this.orientation == PlotOrientation.VERTICAL){
		    	while((rangeDiff/tick)*factorHeight > this.height/2){
		    		//System.out.println("END:" + (rangeDiff/tick)*factorHeight);
		    		tick = tick+increment;
		    	}
		    }
		    if(this.orientation == PlotOrientation.HORIZONTAL){
		    	while((rangeDiff/tick)*factorWidth > this.width/3){
		    		//System.out.println("END:" + (rangeDiff/tick)*factorWidth);
		    		tick = tick+increment;
		    	}
		    }	
		    //System.out.println("TICK: " + tick);
		    NumberTickUnit unit = new NumberTickUnit(tick);		    	
    		
		    axis.setTickUnit(unit);
	    }
	}
	
	
	/**
	 * Customize the chart
	 * Set the background color
	 * @param chart
	 */
	private void chartProperties(JFreeChart chart){
		chart.setBackgroundPaint(Color.WHITE);
	}
	
	/**
	 * Calculate the width dimension of the chart
	 * The maximum width is set by the user in the configuration file, the real dimension is calculated as product of the columns and a fixed factor (22)
	 * The factor 22 is evaluated as the minimum space needed for a label for avoid overlapping 
	 * @param columnCount the number of columns in the chart (the number of categories)
	 * @return
	 */
	public void setWidthDimension(int columnCount){
		//this.width = columnCount*factorWidth<1024?1024:columnCount*factorWidth;
		//this.width = columnCount*factorWidth;//<1024?1024:columnCount*factorWidth;
		this.width = 1024;
	}
	
	/**
	 * Calculate the height dimension of the chart
	 * The maximum height is set by the user in the configuration file, the real dimension is calculated as product of the columns and a fixed factor (15)
	 * The factor 15 is evaluated as the minimum space needed for a label for avoid overlapping
	 * @param columnCount the number of columns in the chart (the number of categories)
	 * @return
	 */
	public void setHeightDimension(int columnCount){
		//this.height = columnCount*factorHeight<800?800:columnCount*factorHeight;
		this.height = 800;//<800?800:columnCount*factorHeight;
	}
	

	//**********************************************//
	//		GETTERS AND SETTERS						//
	//**********************************************//
	
	
	
	public String getChartName() {
		return imageName;
	}



	public void setChartName(String chartName) {
		this.imageName = chartName;
	}



	public String getChartTitle() {
		return chartTitle;
	}



	public void setChartTitle(String chartTitle) {
		this.chartTitle = chartTitle;
	}



	public DefaultCategoryDataset getDataset() {
		return dataset;
	}



	public void setDataset(DefaultCategoryDataset dataset) {
		this.dataset = dataset;
	}



	public String getDomain() {
		return domain;
	}



	public void setDomain(String domain) {
		this.domain = domain;
	}



	public boolean isLegend() {
		return legend;
	}



	public void setLegend(boolean legend) {
		this.legend = legend;
	}



	public PlotOrientation getOrientation() {
		return orientation;
	}



	public void setOrientation(PlotOrientation orientation) {
		this.orientation = orientation;
	}



	public String getRange() {
		return range;
	}



	public void setRange(String range) {
		this.range = range;
	}



	public boolean isTolltips() {
		return tooltips;
	}



	public void setTolltips(boolean tolltips) {
		this.tooltips = tolltips;
	}



	public boolean isUrls() {
		return urls;
	}



	public void setUrls(boolean urls) {
		this.urls = urls;
	}


	public String getSchema() {
		return schema;
	}


	public void setSchema(String schema) {
		this.schema = schema;
	}


	public String getSqlStat() {
		return sqlStat;
	}


	public void setSqlStat(String sqlStat) {
		this.sqlStat = sqlStat;
	}


	public String getImageName() {
		return imageName;
	}


	public void setImageName(String imageName) {
		this.imageName = imageName;
	}

	public String getPathToChart() {
		return pathToChart;
	}

	public void setPathToChart(String pathToChart) {
		this.pathToChart = pathToChart;
	}

	public IntervalXYDataset getIntervalDataset() {
		return intervalDataset;
	}

	public void setIntervalDataset(IntervalXYDataset dataset2) {
		intervalDataset = dataset2;
	}

	public XYSeriesCollection getXyDataset() {
		return xyDataset;
	}

	public void setXyDataset(XYSeriesCollection xyDataset) {
		this.xyDataset = xyDataset;
	}

	public void setColumnsCount(int columnCount) {
		this.columnsCount = columnCount;
	}

	public int getColumnsCount() {
		return columnsCount;
	}


}
